PROGRAM Clique (Infile, Treefile, OUTPUT);
  (* version 3.3. (c) Copyright 1986, 1990 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein,
     Jerry Shurman, and Hisashi Horino.  Permission is granted to copy
     and use this program provided no fee is charged for it and provided
     that this copyright notice is not removed. *)

   CONST NmLngth = 10; (* number of characters in a name *)
      MaxGrp = 2000;   (* max. no. of groups in all tree found *)
      MaxSpp = 50;     (* max. number of species *)
      MaxChr = 190;    (* max. number of characters *)
      FormWide = 80;   (* width of output page *)

   TYPE Vector = ARRAY [1..MaxChr] OF BOOLEAN;
      NameArray = ARRAY [1..MaxSpp, 1..NmLngth] OF CHAR;
      DataArray = ARRAY [1..MaxSpp, 1..MaxChr] OF BOOLEAN;
      Matrix = ARRAY [1..MaxChr] OF Vector;
      spset = SET OF 1..MaxSpp;   
      nodeptr = ^node;   (* nodes will form a binary tree *)
      node = RECORD      (* describes a tip species or an ancestor *)
         next, back: nodeptr;  (* pointers to nodes *)
         index : INTEGER;      (* number of the node *)
         maxpos: INTEGER;       
         tip : BOOLEAN;        (* present species are tips of tree *)
         nodeset : spset;      (* used by accumulate *)
         xcoord, ycoord, ymin, ymax : INTEGER   (* used by printree *)
         END;
      pointarray = ARRAY[1..MaxSpp] OF nodeptr;

   VAR NumSpp, NumChars, ActualChars, Cliqmin, outgrno : INTEGER;
      ancvar, Clmin, Factors, outgropt, trout,
         weights, Anerror, noroot : BOOLEAN;
      Nayme : NameArray;
      ancone : Vector;
      Factor : ARRAY[1..MaxChr] OF CHAR;
      ActChar, oldweight, weight : ARRAY[1..MaxChr] OF INTEGER;
      Data : DataArray;
      Comp : Matrix; (* the character compatibility matrix *)
      Infile, treefile : TEXT;
      root: nodeptr;
      grouping: ARRAY[1..MaxGrp] OF spset;
      treenode : pointarray;      (* pointers to all nodes in tree *)

   PROCEDURE gnu (VAR p : nodeptr);
      (* replacement for NEW *)
   BEGIN (* gnu *)
      NEW (p);
      p^.next := nil;
      p^.tip := FALSE;
   END; (* gnu *)

   PROCEDURE setuptree;
      (* initialization of tree pointers, variables *)
      VAR i : INTEGER;
   BEGIN (* setuptree *)
      FOR i := 1 TO MaxSpp DO BEGIN
         gnu (treenode[i]);
         treenode[i]^.next := NIL;
         treenode[i]^.back := NIL;
         treenode[i]^.index := i;
         treenode[i]^.tip := FALSE;
         END;
   END; (* setuptree *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a digit *)
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   PROCEDURE NewLine (i, j : INTEGER);
      (* goes to new line if i MOD j is zero *)
      VAR k : INTEGER;
   BEGIN (* NewLine *)
      IF ((i MOD j) = 0)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR k := 1 TO NmLngth+1 DO
            WRITE(OUTPUT, ' ');
         END;
   END; (* NewLine *)

   PROCEDURE ReadData;
      (* reads the species names and character data *)
      VAR i, j, Extranum : INTEGER;
         ch : CHAR;

      FUNCTION letter (ch : CHAR) : BOOLEAN;
         (* test whether ch is a letter
             -- works for both ASCII and EBCDIC codes *)
      BEGIN (* letter  *)
         letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
      END; (* letter *)

      PROCEDURE uppercase (VAR ch : CHAR);
         (* convert a character to upper case -- either ASCII or EBCDIC *)
      BEGIN (* uppercase *)
         IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                        OR  ('a' < 'A') AND (ch < 'A'))
         THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
      END; (* uppercase *)
   
      PROCEDURE inputancestors;
         (* reads the ancestral states for each character *)
         VAR i : INTEGER;
            ch : CHAR;
      BEGIN (* inputancestors *)
         FOR i := 1 TO NmLngth-1 DO READ(Infile, ch);
         ancvar := TRUE;
         FOR i := 1 TO NumChars DO BEGIN
            REPEAT
               IF EOLN(Infile) THEN READLN(Infile);
               READ(Infile, ch);
            UNTIL ch <> ' ';
            IF (ch = '0') OR (ch = '1')
            THEN BEGIN
               CASE ch OF
                  '1' : ancone[i] := TRUE;
                  '0' : ancone[i] := FALSE
                  END;
               END
            ELSE BEGIN
               WRITELN(OUTPUT, 'BAD ANCESTOR STATE: ', ch,
                          ' AT CHARACTER ', i:4);
               anerror := TRUE;
               END;
            END;
         READLN(Infile);
      END; (* inputancestors *)

      PROCEDURE printancestors;
         (* print out list of ancestral states *)
         VAR i : INTEGER;
      BEGIN (* printancestors *)
         WRITELN(OUTPUT, 'Ancestral states:');
         FOR i := 1 TO NmLngth+2 DO WRITE(OUTPUT, ' ');
         FOR i := 1 TO NumChars DO BEGIN
            Newline (i, 55);
            IF ancone[i]
            THEN WRITE(OUTPUT, '1')
            ELSE WRITE(OUTPUT, '0');
            IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* printancestor *)

      PROCEDURE inputfactors;
         (* reads the factor symbols *)
         VAR i : INTEGER;
            ch : CHAR;
      BEGIN (* inputfactors *)
         ActualChars := 1;
         FOR i := 1 TO NmLngth-1 DO READ(infile, ch);
         FOR i := 1 TO NumChars DO BEGIN
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, Factor[i]);
            IF (i>1)
            THEN IF Factor[i] <> Factor[i-1]
               THEN ActualChars := ActualChars + 1;
            ActChar[i] := ActualChars;
            END;
         READLN(infile);
      END; (* inputfactors *)

      PROCEDURE printfactors;
         (* print out list of factor symbols *)
         VAR i : INTEGER;
      BEGIN (* printfactors *)
         WRITE(OUTPUT, 'Factors:');
         FOR i := 1 TO NmLngth-6 DO WRITE(OUTPUT, ' ');
         FOR i := 1 TO NumChars DO BEGIN
            NewLine (i, 50);
            WRITE(OUTPUT, Factor[i]);
            IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* printfactors *)

      PROCEDURE inputoutgroup;
         (* input the number of the outgroup *)
      BEGIN (* inputoutgroup *)
         READLN(Infile, outgrno);
         IF (outgrno < 1) OR (outgrno > NumSpp)
         THEN BEGIN
            WRITELN(OUTPUT, 'BAD OUTGROUP NUMBER: ', outgrno:4);
            anerror := TRUE;
            END;
      END; (* inputoutgroup *)

      PROCEDURE inputweights;
         (* input the character weights, 0-9 and A-Z for weights 0 - 35 *)
         VAR ch : CHAR;
            i : INTEGER;
      BEGIN (* inputweights *)
         FOR i := 1 TO NmLngth-1 DO READ(infile, ch);
         FOR i := 1 TO NumChars DO BEGIN
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            oldweight[i] := 1;
            IF digit(ch)
            THEN oldweight[i] := ORD(ch) - ORD('0')
            ELSE IF letter(ch)
               THEN BEGIN
                  uppercase (ch);
                  IF (ch >= 'A') AND (ch <= 'I')
                  THEN oldweight[i] := ORD(ch) - ORD('A') + 10
                  ELSE IF (ch >= 'J') AND (ch <= 'R')
                     THEN oldweight[i] := ORD(ch) - ORD('J') + 19
                     ELSE oldweight[i] := ORD(ch) - ORD('S') + 28;
                  END
               ELSE BEGIN
                  WRITELN(OUTPUT, 'BAD WEIGHT CHARACTER: ',ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         weights := TRUE;
      END; (* inputweights *)

      PROCEDURE printweights;
         (* print out the weights of characters *)
         VAR i, j, k : INTEGER;
      BEGIN (* printweights *)
         WRITELN(OUTPUT, '   Characters are weighted as follows:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT,'     *---------------------------------');
         FOR j := 0 TO (ActualChars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*j:5, '! ');
            FOR i := 0 TO 9 DO BEGIN
               k := 10*j + i;
               IF (k>0) AND (k<=ActualChars)
               THEN WRITE(OUTPUT, oldweight[k]:3)
               ELSE WRITE(OUTPUT, '   ');
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
      END; (* printweights *)

   BEGIN (* ReadData *)
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, 'Largest clique program, version 3.3');
      WRITELN(OUTPUT);
      READ (Infile, NumSpp, NumChars);
      IF NumSpp > MaxSpp
      THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: increase CONSTants');
      IF NumChars > MaxChr
      THEN WRITELN(OUTPUT, 'TOO MANY CHARACTERS: increase CONSTants');
      Anerror := (NumSpp > MaxSpp) OR (NumChars > MaxChr);
      WRITELN(OUTPUT, NumSpp:2, ' species, ', NumChars:3, ' character states');
      WRITELN(OUTPUT);
      ancvar := FALSE;
      Clmin := FALSE;
      Factors := FALSE;
      outgropt := FALSE;
      weights := FALSE;
      trout := FALSE;
      ActualChars := NumChars;
      FOR i := 1 TO NumChars DO ActChar[i] := i;
      FOR i := 1 TO NumChars DO oldweight[i] := 1;
      outgrno := 1;
      Extranum := 0;
      WHILE NOT EOLN(Infile) DO BEGIN
         READ(infile, ch);
         uppercase (ch);
         IF (ch = 'A') OR (ch = 'C') OR (ch = 'F')
           OR (ch = 'O') OR (ch = 'W') OR (ch = 'Y')
         THEN BEGIN
            Extranum := Extranum + 1;
            CASE ch OF
               'A' : ancvar := TRUE;
               'C' : Clmin := TRUE;
               'F' : Factors := TRUE;
               'O' : outgropt := TRUE;
               'W' : weights := TRUE;
               'Y' : BEGIN
                     trout := TRUE;
                     Extranum := Extranum - 1;
                     END
               END;
            END
         ELSE IF ch <> ' '
            THEN BEGIN
               WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
               WRITELN(OUTPUT);
               Anerror := TRUE;
               END;
         END;
      READLN (Infile);
      FOR i := 1 TO Extranum DO
         IF NOT Anerror
         THEN BEGIN
            READ(infile, ch);
            uppercase (ch);
            Anerror := NOT ((ch = 'A') OR (ch = 'C') OR (ch = 'F')
                         OR (ch = 'O') OR (ch = 'W'));
            IF Anerror
            THEN WRITELN(OUTPUT, 'ERROR: INCORRECT AUXILIARY',
                  ' OPTIONS LINE WHICH STARTS WITH ', ch);
            IF ch = 'A' THEN inputancestors;
            IF ch = 'C' THEN READLN (Infile, Cliqmin);
            IF ch = 'F' THEN inputfactors;
            IF ch = 'O' THEN inputoutgroup;
            IF ch = 'W' THEN inputweights;
            END;
      IF weights THEN printweights;
      IF Factors THEN printfactors;
      IF ancvar THEN printancestors;
      noroot := NOT (outgropt OR ancvar);
      j := (NumChars DIV 2) + (((NumChars DIV 5) - 1) DIV 2) - 5;
      IF j < 0 THEN j := 0;
      IF j > 27 THEN j := 27;
      WRITE(OUTPUT, 'Species  ');
      FOR i := 1 To j DO WRITE(OUTPUT, ' ');
      WRITELN(OUTPUT, 'Character states');
      WRITE(OUTPUT, '-------  ');
      FOR i := 1 To j DO WRITE(OUTPUT, ' ');
      WRITELN(OUTPUT, '--------- ------');
      WRITELN(OUTPUT);
      FOR i := 1 TO NumSpp DO
         IF NOT Anerror
         THEN BEGIN
            FOR j := 1 TO NmLngth DO
               IF NOT Anerror
               THEN BEGIN
                  Anerror := Anerror OR EOF(infile) OR EOLN(infile);
                  IF Anerror
                  THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                     'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                  READ(Infile, Nayme[i, j]);
                  WRITE(OUTPUT, Nayme[i, j]);
                  END;
            WRITE(OUTPUT, '  ');
            FOR j := 1 TO NumChars DO BEGIN
               REPEAT
                  IF EOLN (Infile) THEN READLN (Infile);
                  READ (Infile, ch);
               UNTIL ch <> ' ';
               WRITE(OUTPUT, ch);
               NewLine(j, 55);
               IF NOT ((ch = '0') OR (ch = '1'))
               THEN BEGIN
                  WRITE(OUTPUT, 'BAD CHARACTER STATE: ', ch, ' (NOT 0 OR 1)',
                     ' AT CHARACTER', j:3, ' OF SPECIES', i:3);
                  WRITELN(OUTPUT);
                  Anerror := TRUE;
                  END;
               IF (j MOD 5 = 0)
               THEN WRITE(OUTPUT, ' ');
               Data [i, j] := (ch = '1');
               END;
            READLN (Infile);
            WRITELN(OUTPUT);
            END;
      WRITELN(OUTPUT);
      For i := 1 TO NumChars DO BEGIN
         IF (i = 1) OR (NOT Factors)
         THEN weight[i] := oldweight[i]
         ELSE IF Factor[i] <> Factor[i-1]
            THEN weight[ActChar[i]] := oldweight[i];
         END;
   END; (* ReadData *)

   PROCEDURE SetUp (VAR Comp : Matrix);
      (* sets up the compatibility matrix *)
      VAR i, j : INTEGER;

      FUNCTION Compatible (ch1, ch2 : INTEGER) : BOOLEAN;
         (* TRUE if two characters ch1 < ch2 are compatible *)
         VAR i, j, k : INTEGER;
            Compt, Done1, Done2 : BOOLEAN;
            Info : ARRAY [1..4] OF BOOLEAN;
      BEGIN (* Compatible *)
         Compt := TRUE;
         j := 1;
         WHILE ch1 > ActChar[j] DO j := j + 1;
         Done1 := NOT (ch1 = ActChar[j]);
         WHILE (NOT Done1) DO BEGIN
            k := j;
            WHILE ch2 > ActChar[k] DO k := k + 1;
            Done2 := NOT (ch2 = ActChar[k]);
            WHILE (NOT Done2) DO BEGIN
               FOR i := 1 TO 4 DO Info [i] := FALSE;
               IF ancvar
               THEN BEGIN
                  IF ancone[j] AND ancone[k]
                  THEN Info [1] := TRUE
                  ELSE IF ancone[j] AND NOT ancone[k]
                     THEN Info [2] := TRUE
                     ELSE IF (NOT ancone[j]) AND ancone[k]
                        THEN Info [3] := TRUE
                        ELSE Info [4] := TRUE;
                  END;
               FOR i := 1 TO NumSpp DO
                  IF Data[i, j] AND Data[i, k]
                  THEN Info [1] := TRUE
                  ELSE IF Data [i, j] AND NOT Data [i, k]
                     THEN Info [2] := TRUE
                     ELSE IF (NOT Data [i, j]) AND Data [i, k]
                        THEN Info [3] := TRUE
                        ELSE Info [4] := TRUE;
               Compt := Compt AND (NOT (Info [1] AND Info [2] AND
                                  Info[3] AND Info [4]));
               k := k + 1;
               Done2 := k > NumChars;
               IF NOT Done2
               THEN Done2 := NOT (ch2 = ActChar[k]);
               END;
            j := j + 1;
            Done1 := j > NumChars;
            IF NOT Done1
            THEN Done1 := NOT (ch1 = ActChar[j]);
            END;
         Compatible := Compt;
      END; (* Compatible *)

   BEGIN (* SetUp *)
      IF Factors
      THEN WRITELN(OUTPUT, '      (For original multistate characters)');
      WRITELN(OUTPUT, 'Character Compatibility Matrix (1 if compatible)');
      WRITELN(OUTPUT, '--------- ------------- ------ -- -- -----------');
      WRITELN(OUTPUT);
      FOR i:= 1 TO ActualChars DO BEGIN
         FOR j := 1 TO ((48-ActualChars) DIV 2) DO WRITE(OUTPUT, ' ');
         FOR j := 1 TO i-1 DO BEGIN
            IF Comp [i][j]
            THEN WRITE(OUTPUT, '1')
            ELSE WRITE(OUTPUT, '.');
            NewLine (j, 70);
            END;
         Comp [i][i] := TRUE;
         WRITE(OUTPUT, '1');
         FOR j := i+1 TO ActualChars DO BEGIN
            Comp [i][j] := Compatible (i, j);
            IF Comp [i][j]
            THEN WRITE(OUTPUT, '1')
            ELSE WRITE(OUTPUT, '.');
            Comp [j][i] := Comp [i][j];
            END;
         WRITELN(OUTPUT);
         END;
      WRITELN(OUTPUT);
   END; (* Setup *)

   PROCEDURE GetMaxCliques (VAR Comp : Matrix);
      (* recursively generates the largest cliques *)
   VAR i : INTEGER;
       Chars, Temp, Processed, Rarer, Candidates, Excluded : Vector;
   
      PROCEDURE Intersect (V1, V2 : Vector; VAR V3 : Vector);
         (* takes the logical intersection V1 AND V2 *)
      VAR i : INTEGER;
      BEGIN (* Intersect *)
         FOR i := 1 TO ActualChars DO
            V3 [i] := V1 [i] AND V2 [i];
      END; (* Intersect *)
   
      FUNCTION CountStates (V : Vector) : INTEGER;
         (* counts the 1's in V *)
         VAR i, TempCount : INTEGER;
      BEGIN (* CountStates *)
         TempCount := 0;
         FOR i := 1 TO ActualChars DO
            IF V[i] THEN TempCount := TempCount + weight[i];
         CountStates := TempCount;
      END; (* CountStates *)

      PROCEDURE Gen1 (i, CurSize : INTEGER; Chars, Candidates,
                       Excluded : Vector);
         (* finds largest size cliques and prints them out *)
         VAR CurSize2, j, k, Actual, Possible : INTEGER;
            Futile : BOOLEAN;
            Chars2, Cands2, Excl2, Cprime, Exprime : Vector;
      BEGIN (* Gen1 *)
         CurSize2 := CurSize;
         Chars2 := Chars;
         Cands2 := Candidates;
         Excl2 := Excluded;
         j := i;
         WHILE (j <= ActualChars) DO BEGIN
            IF Cands2[j]
            THEN BEGIN
               Chars2[j] := TRUE;
               Cands2[j] := FALSE;
               CurSize2 := CurSize2 + weight[j];
               Possible := CountStates (Cands2);
               Intersect (Cands2, Comp[j], Cprime);
               Actual := CountStates (Cprime);
               Intersect (Excl2, Comp[j], Exprime);
               Futile := FALSE;
               FOR k := 1 TO j-1 DO BEGIN
                  IF (Exprime[k] AND (NOT Futile))
                  THEN BEGIN
                     Intersect (Cprime, Comp[k], Temp);
                     Futile := (CountStates(Temp) = Actual);
                     END;
                  END;
               IF (CurSize2 + Actual >= Cliqmin) AND (NOT Futile)
               THEN IF Actual > 0
                  THEN Gen1 (j+1, CurSize2, Chars2, Cprime, Exprime)
                  ELSE IF CurSize2 > Cliqmin
                     THEN Cliqmin := CurSize2;
               IF (Possible > Actual)
               THEN BEGIN
                  Chars2[j] := FALSE;
                  Excl2[j] := TRUE;
                  CurSize2 := CurSize2 - weight[j];
                  END
               ELSE j := ActualChars;
               END;
            j := j + 1;
            END;
      END; (* Gen1 *)
      
      PROCEDURE DoAll (Chars : Vector; VAR Processed, Rarer : Vector);
         (* print out a clique and its tree *)
      TYPE SpVector = ARRAY [1..MaxSpp] OF INTEGER;
         ChVector = ARRAY [1..MaxChr] OF INTEGER;
      VAR i, n, MaxChars : INTEGER;
         SpOrder : SpVector;
         ChOrder, Count : ChVector;

         FUNCTION Ingroupstate (i : INTEGER) : BOOLEAN;
            (* the ingroup state for the i-th character *)
            VAR outstate : BOOLEAN;
         BEGIN (* Ingroupstate *)
            IF noroot
            THEN outstate := Data[1, i]
            ELSE IF ancvar
               THEN outstate := ancone[i]
               ELSE outstate := Data[outgrno, i];
            Ingroupstate := NOT outstate;
         END; (* Ingroupstate *)
   
         PROCEDURE makeset;
            (* make up set of species for given set of characters *)
         VAR i, j : INTEGER;
             instate : BOOLEAN;
             st: spset;
         BEGIN (* makeset *)
            n := 1;
            FOR i := 1 TO MaxChars DO BEGIN
               st := [];
               instate := Ingroupstate (ChOrder[i]);
               FOR j := 1 TO NumSpp DO
                  IF Data[SpOrder[j], ChOrder[i]] = instate
                  THEN st := st + [SpOrder[j]];
               grouping[n] := st;
               n := n + 1;
               END;
            FOR i := 1 TO NumSpp DO BEGIN
               grouping[n] := [SpOrder[i]];
               n := n + 1;
               END;
         END; (* makeset *)
   
         PROCEDURE Init (VAR ChOrder, Count : ChVector;
                          VAR MaxChars : INTEGER);
            (* initialize vectors and character count *)
            VAR i, j, temp : INTEGER;
               instate : BOOLEAN;
         BEGIN (* Init *)
            MaxChars := 0;
            FOR i := 1 TO NumChars DO
               IF Chars [ActChar[i]]
               THEN BEGIN
                  MaxChars := MaxChars + 1;
                  ChOrder [MaxChars] := i;
                  instate := Ingroupstate (i);
                  temp := 0;
                  FOR j := 1 TO NumSpp DO BEGIN
                     IF Data [j, i] = instate
                     THEN temp := temp + 1;
                     END;
                  Count [i] := temp;
                  END;
         END; (* Init *)
   
         PROCEDURE ChSort (VAR ChOrder : ChVector; Count : ChVector);
            (* sorts the characters by number of ingroup states *)
            VAR j, temp : INTEGER;
               ordered : BOOLEAN;
         BEGIN (* ChSort *)
            ordered := FALSE;
            WHILE NOT ordered DO BEGIN
               ordered := TRUE;
               FOR j := 1 TO MaxChars-1 DO
                  IF Count[ChOrder[j]] < Count[ChOrder[j+1]]
                  THEN BEGIN
                     ordered := FALSE;
                     temp := ChOrder [j];
                     ChOrder [j] := ChOrder [j+1];
                     ChOrder [j+1] := temp;
                     END;
               END;
         END; (* ChSort *)
   
         PROCEDURE PrintClique (Chars : Vector);
            (* prints the characters in a clique *)
            VAR i, j : INTEGER;
         BEGIN (* PrintClique *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            IF Factors
            THEN BEGIN
               WRITE(OUTPUT, 'Actual Characters: (');
               j := 0;
               FOR i := 1 TO ActualChars DO BEGIN
                  IF Chars [i]
                  THEN BEGIN
                     WRITE(OUTPUT, i:3);
                     j := j + 1;
                     NewLine (j , (FormWide-22) DIV 3)
                     END;
                  END;
               WRITELN(OUTPUT, ')');
               END;
            IF Factors
            THEN WRITE(OUTPUT, 'Binary ');
            WRITE(OUTPUT, 'Characters: (');
            j := 0;
            FOR i := 1 TO NumChars DO BEGIN
               IF Chars [ActChar[i]]
               THEN BEGIN
                  WRITE(OUTPUT, i:3);
                  j := j + 1;
                  IF Factors
                  THEN NewLine (j , (FormWide-22) DIV 3)
                  ELSE NewLine (j , (FormWide-15) DIV 3);
                  END;
               END;
            WRITELN(OUTPUT, ')');
            WRITELN(OUTPUT);
         END; (* PrintClique *)
   
         PROCEDURE reconstruct;
            (* reconstruct tree from the subsets *)
            VAR s : spset;
   
            PROCEDURE bigsubset (VAR st : spset);
               (* find a maximal subset of st among the groupings *)
               VAR i : INTEGER;
                  su : spset;
            BEGIN (* bigsubset *)
               su := [];
               FOR i := 1 TO n DO
                 IF (st >= grouping[i]) AND (grouping[i] <> st)
                 THEN IF (grouping[i] >= su) AND (grouping[i] <> su)
                    THEN su := grouping[i];
               st := su;
            END; (* bigsubset *)
   
            PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
               (* traverse to reconstruct the tree from the characters *)
               VAR i, j, k, maxpos: INTEGER;
                  found : BOOLEAN;
                  tempset : spset;
                  q: nodeptr;
            BEGIN (* recontraverse *)
               k := 0;
               FOR i := 1 TO NumSpp DO
                  IF i IN st
                  THEN BEGIN
                     k := k + 1;
                     j := i;
                     END;
               IF k = 1 
               THEN BEGIN
                  p := treenode[j];
                  p^.tip := TRUE;
                  p^.index := j;
                  END
               ELSE BEGIN
                  gnu(p);
                  p^.index := 0;
                  tempset := st;
                  q := p;
                  IF tempset <> [] THEN bigsubset(tempset);
                  WHILE (st <> []) AND (tempset <> []) DO BEGIN
                     gnu (q^.next);
                     q := q^.next;
                     recontraverse(q^.back, tempset);
                     i := 1;
                     maxpos := 0;
                     WHILE i <= MaxChars DO BEGIN
                        IF grouping[i] = tempset THEN
                           maxpos := i;
                        i := i + 1;   
                        END;    
                     q^.back^.maxpos := maxpos;    
                     q^.back^.back := q;
                     st := st - tempset;
                     tempset := st;
                     found := FALSE;
                     i := 1;
                     WHILE (NOT found) AND (i <= n) DO BEGIN
                        IF grouping[i] = tempset THEN found := TRUE
                        ELSE i := i + 1;
                        END;
                     IF (tempset <> []) AND (NOT found)
                     THEN bigsubset(tempset); 
                     END;
                  q^.next := p;
                  END; 
            END; (* recontraverse *)
   
         BEGIN (* reconstruct *)
            s := [1..NumSpp];
            recontraverse (root, s);
         END; (* reconstruct *)
   
         PROCEDURE reroot (outgroup : nodeptr);
            (* reorients tree, putting outgroup in desired position. *)
            VAR i: INTEGER;
               nroot: BOOLEAN;
               p, q: nodeptr;
         BEGIN (* reroot *)
            nroot := FALSE;
            p := root^.next;
            WHILE(p <> root) DO
               IF (outgroup^.back = p)
               THEN BEGIN
                  nroot := TRUE;
                  p := root;
                  END
               ELSE p := p^.next;
            IF NOT nroot
            THEN BEGIN
               p := root;
               i:=0;
               WHILE ( p^.next <> root) DO BEGIN
                  p := p^.next;
                  i:=i+1;
                  END;
               IF (i = 2)
               THEN BEGIN
                  root^.next^.back^.back := p^.back;
                  p^.back^.back := root^.next^.back;
                  q := root^.next;
                  END
               ELSE BEGIN
                  p^.next := root^.next;
                  gnu(root^.next); 
                  q := root^.next;
                  gnu(q^.next); 
                  p := q^.next;
                  p^.next := root;
                  q^.tip := false;
                  p^.tip := false; 
                  END;
               q^.back := outgroup;
               p^.back := outgroup^.back;
               outgroup^.back^.back :=p;
               outgroup^.back :=q;
               END;
         END; (* reroot *)
      
         PROCEDURE printree;
            (* prints out diagram of the tree *)
            CONST down = 2;
            VAR tipy, i : INTEGER;
   
            PROCEDURE coordinates (p : nodeptr);
               (* establishes coordinates of nodes *)
              VAR q, first, last: nodeptr;         
                 maxx : INTEGER;
            BEGIN (* coordinates *)
               IF p^.tip
               THEN BEGIN
                  p^.xcoord := 0;
                  p^.ycoord := tipy;
                  p^.ymin := tipy;
                  p^.ymax := tipy;
                  tipy := tipy + down;
                  END
               ELSE BEGIN
                  q := p^.next; 
                  maxx := 0;
                  WHILE(q <> p) DO BEGIN
                     coordinates (q^.back);
                     IF NOT q^.back^.tip
                     THEN IF q^.back^.xcoord > maxx
                        THEN  maxx := q^.back^.xcoord;
                     q := q^.next;
                     END;
                  first := p^.next^.back; 
                  q := p;
                  WHILE q^.next <> p DO 
                     q := q^.next;
                  last := q^.back;
                  p^.xcoord := 3 * (MaxChars - p^.maxpos) - 2;
                  IF p = root THEN p^.xcoord := p^.xcoord + 2;     
                  p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
                  p^.ymin := first^.ymin;
                  p^.ymax := last^.ymax;
                  END;
            END; (* coordinates *)
   
            PROCEDURE drawline(i : INTEGER);
               (* draws one row of the tree diagram by moving up tree *)
               VAR p, q : nodeptr;
                  n, m, j, k, sumlocpos, size, locpos, branchpos: INTEGER;
                  poslist : ARRAY [1..MaxGrp] OF INTEGER;
                  extra, done, plus, found: BOOLEAN;
                  r, first, last : nodeptr;
            BEGIN (* drawline *)
               branchpos := 0;
               p := root;
               q := root;
               WRITE(OUTPUT, '  ');
               extra := FALSE;
               plus := FALSE;
               REPEAT
                  IF NOT p^.tip
                  THEN BEGIN
                     found := FALSE;
                     r := p^.next;
                     WHILE (r <> p) AND NOT found DO 
                        IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                        THEN BEGIN 
                           q := r^.back;
                           found := TRUE;
                           END
                        ELSE r := r^.next;
                     first := p^.next^.back; 
                     r := p;
                     WHILE r^.next <> p DO 
                        r := r^.next;
                     last := r^.back;
                     END;
                  done := (p^.tip) OR (p = q);
                  n := p^.xcoord - q^.xcoord;
                  m := n;
                  IF extra
                  THEN BEGIN
                     n := n - 1;
                     extra := FALSE;
                     END;
                  IF (q^.ycoord = i) AND (NOT DONE)
                  THEN BEGIN
                     IF NOT (q^.tip)
                     THEN BEGIN
                        WRITE(OUTPUT, '+');
                        plus := TRUE;
                        j := 1;       
                        FOR k := 1 TO q^.maxpos DO
                           IF grouping[k] = grouping[q^.maxpos] 
                           THEN BEGIN
                              poslist[j] := k;
                              j := j + 1;
                              END;
                        size := j-1;
                        IF size = 0 
                        THEN BEGIN
                           FOR k := 1 TO n-1 DO WRITE(OUTPUT, '-');
                           sumlocpos := n;
                           END
                        ELSE BEGIN
                           sumlocpos := 0;
                           j := 1;   
                           WHILE  j <= size DO BEGIN
                              locpos := 3*poslist[j];
                              IF j <> 1
                              THEN locpos := locpos - 3 * poslist[j-1] 
                              ELSE locpos := locpos - branchpos;
                              FOR k := 1 TO locpos-1 DO WRITE(OUTPUT, '-');
                              IF Rarer[ChOrder[poslist[j]]]
                              THEN WRITE(OUTPUT, '1')
                              ELSE WRITE(OUTPUT, '0'); 
                              sumlocpos := sumlocpos + locpos;
                              j := j + 1;
                              END; 
                           FOR j := sumlocpos+1 TO n-1 DO WRITE(OUTPUT, '-');
                           WRITE(OUTPUT, '+');
                           IF m > 0
                           THEN branchpos := branchpos + m;
                           extra := TRUE;
                           END;
                        END
                     ELSE BEGIN 
                        IF NOT plus 
                        THEN BEGIN
                           WRITE(OUTPUT, '+');
                           plus := FALSE;
                           END 
                        ELSE n := n +1;
                        j := 1;
                        FOR k := 1 TO q^.maxpos DO
                           IF grouping[k] = grouping[q^.maxpos] 
                           THEN BEGIN
                              poslist[j]  := k;
                              j := j + 1;
                              END;
                        size := j-1;
                        IF size = 0 
                        THEN BEGIN
                           FOR k := 1 TO n DO WRITE(OUTPUT, '-');
                           sumlocpos := n;
                           END
                        ELSE BEGIN
                           sumlocpos := 0;  
                           j := 1;   
                           WHILE  j <= size DO BEGIN
                              locpos := 3 * poslist[j];
                              IF j <> 1
                              THEN locpos := locpos - 3 * poslist[j-1]
                              ELSE locpos := locpos - branchpos;
                              FOR k := 1 TO locpos-1 DO WRITE(OUTPUT, '-');
                              IF Rarer[ChOrder[poslist[j]]]
                              THEN WRITE(OUTPUT, '1')
                              ELSE WRITE(OUTPUT, '0'); 
                              sumlocpos := sumlocpos + locpos;
                              j := j + 1;
                              END; 
                           FOR j := sumlocpos+1 TO n DO  WRITE(OUTPUT, '-');
                           IF m > 0 THEN branchpos := branchpos + m;
                           END;
                        WRITE(OUTPUT, '-'); 
                        END;
                     END 
                  ELSE IF (NOT p^.tip) AND (last^.ycoord > i)
                         AND (first^.ycoord < i) 
                         AND ((NOT (i = p^.ycoord)) OR (p = root))
                     THEN BEGIN
                        WRITE(OUTPUT, '!');
                        FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                        plus := FALSE;     
                        IF m > 0 THEN branchpos := branchpos + m;
                        END
                     ELSE BEGIN 
                        FOR j := 1 TO n DO WRITE(OUTPUT, ' ');
                        plus := FALSE;
                        IF m > 0 THEN branchpos := branchpos + m;
                        END;
                  IF q <> p THEN p := q;
               UNTIL DONE;
               IF (p^.ycoord = i) AND (p^.tip)
               THEN FOR j := 1 TO NmLngth DO WRITE(OUTPUT, Nayme[p^.index, j]);
               WRITELN(OUTPUT);
            END; (* drawline *)
   
         BEGIN (* printree *)
            tipy := 1;
            coordinates (root);
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '  Tree and');
            IF Factors THEN WRITE(OUTPUT, ' binary');
            WRITELN(OUTPUT, ' characters:');
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   ');
            FOR i := 1 TO MaxChars DO WRITE(OUTPUT, ChOrder[i]:3);
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   ');
            FOR i := 1 TO MaxChars DO 
                IF (Rarer[ChOrder[i]]) 
                THEN WRITE(OUTPUT, '1':3)
                ELSE WRITE(OUTPUT, '0':3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            FOR i := 1 TO tipy-down DO drawline (i);
            WRITELN(OUTPUT);
         END; (* printree *)
     
         PROCEDURE treeout (p : nodeptr);
           (* write out file with representation of final tree *)
           VAR i, n : INTEGER;
             c : CHAR;
             q : nodeptr;
         BEGIN (* treeout *)
           IF p^.tip
           THEN BEGIN
             n := 0;
             FOR i := 1 TO nmlngth DO
               IF Nayme[p^.index, i] <> ' '
               THEN n := i;
             FOR i := 1 TO n DO BEGIN
               c := Nayme[p^.index, i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
             END
           ELSE BEGIN
             q := p^.next;
             WRITE(treefile, '(');
             WHILE q <> p DO BEGIN
                treeout (q^.back);
                q := q^.next;
                IF q <> p THEN WRITE(treefile, ',');
                END;
             WRITE(treefile, ')');
             END;
           IF p = root
           THEN WRITELN(treefile, ';')
         END; (* treeout *)
     
      BEGIN (* DoAll *)
         Init (ChOrder, Count, MaxChars);
         ChSort (ChOrder, Count);
         FOR i := 1 TO NumSpp DO SpOrder [i] := i;
         FOR i := 1 TO NumChars DO
            IF Chars [ActChar[i]]
            THEN BEGIN
              IF NOT Processed [ActChar[i]]
              THEN BEGIN
                 Rarer [i] := Ingroupstate (i);
                 Processed [ActChar[i]] := TRUE;
                 END;
              END;
         PrintClique (Chars);
         makeset;  
         setuptree; 
         reconstruct;
         IF noroot
         THEN reroot (treenode[outgrno]);
         printree;
         IF trout
         THEN treeout (root);
      END; (* DoAll *)

      PROCEDURE Gen2 (i, CurSize : INTEGER; Chars, Candidates,
                       Excluded : Vector);
         (* finds largest size cliques and prints them out *)
         VAR CurSize2, j, k, Actual, Possible : INTEGER;
            Futile : BOOLEAN;
            Chars2, Cands2, Excl2, Cprime, Exprime : Vector;
      BEGIN (* Gen2 *)
         CurSize2 := CurSize;
         Chars2 := Chars;
         Cands2 := Candidates;
         Excl2 := Excluded;
         j := i;
         WHILE (j <= ActualChars) DO BEGIN
            IF Cands2[j]
            THEN BEGIN
               Chars2[j] := TRUE;
               Cands2[j] := FALSE;
               CurSize2 := CurSize2 + weight[j];
               Possible := CountStates (Cands2);
               Intersect (Cands2, Comp[j], Cprime);
               Actual := CountStates (Cprime);
               Intersect (Excl2, Comp[j], Exprime);
               Futile := FALSE;
               FOR k := 1 TO j-1 DO BEGIN
                  IF (Exprime[k] AND (NOT Futile))
                  THEN BEGIN
                     Intersect (Cprime, Comp[k], Temp);
                     Futile := (CountStates(Temp) = Actual);
                     END;
                  END;
               IF (CurSize2 + Actual >= Cliqmin) AND (NOT Futile)
               THEN IF Actual > 0
                  THEN Gen2 (j+1, CurSize2, Chars2, Cprime, Exprime)
                  ELSE DoAll (Chars2, Processed, Rarer);
               IF (Possible > Actual)
               THEN BEGIN
                  Chars2[j] := FALSE;
                  Excl2[j] := TRUE;
                  CurSize2 := CurSize2 - weight[j];
                  END
               ELSE j := ActualChars;
               END;
            j := j + 1;
            END;
      END; (* Gen2 *)

   BEGIN (* GetMaxCliques *)
      WRITELN(OUTPUT);
      IF Clmin
      THEN BEGIN
         WRITELN(OUTPUT, 'Cliques with at least', Cliqmin:3, ' characters');
         WRITELN(OUTPUT, '------- ---- -- ----- -- ----------');
         END
      ELSE BEGIN
         Cliqmin := 0;
         WRITELN(OUTPUT, 'Largest Cliques');
         WRITELN(OUTPUT, '------- -------');
         FOR i := 1 TO ActualChars DO BEGIN
            Chars [i] := FALSE;
            Excluded [i] := FALSE;
            Candidates [i] := TRUE;
            END;
         Gen1 (1, 0, Chars, Candidates, Excluded);
         END;
      FOR i := 1 TO ActualChars DO BEGIN
         Chars [i] := FALSE;
         Candidates [i] := TRUE;
         Processed [i] := FALSE;
         Excluded [i] := FALSE;
         END;
      Gen2 (1, 0, Chars, Candidates, Excluded);
      WRITELN(OUTPUT);
   END; (* GetMaxCliques *)

BEGIN (* Main Program *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET (Infile);
   (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
   REWRITE (treefile);
   Anerror := FALSE;
   ReadData;
   IF NOT Anerror
   THEN BEGIN
      SetUp (Comp);
      GetMaxCliques (Comp);
      IF trout
      THEN BEGIN
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'Tree(s) written on tree file');
         WRITELN(OUTPUT);
         END;
      END;
   IF trout
   THEN BEGIN
   (* CLOSE(treefile);   Turbo Pascal and maybe others too *)
      END;
END.

