PROGRAM Consense (infile, OUTPUT, treefile);
  (* version 3.3. (c) Copyright 1986, 1989 by the University of
     Washington.  Written by Joseph Felsenstein and Hisashi Horino.
     Permission is granted to copy and use this program provided no fee
     is charged for it and provided that this copyright notice is not
     removed. *)

CONST maxsp = 60;  (* maximum number of species              *)
   nmlngth = 20;   (* max. no. of characters in species name *)
   maxgrp = 2000;  (* max. no. of groups in all trees found  *)

TYPE spset = SET OF 1..maxsp;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      next, back: nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      nodeset : spset;      (* used by accumulate *)
      xcoord, ycoord, ymin, ymax : INTEGER;   (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsp] OF nodeptr;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, numopts, outgrno, i, ntrees : INTEGER; (* outgrno indicates outgroup *)
   anerror, trout, first, noroot, outgropt, didreroot : BOOLEAN;
   ch : CHAR;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   grouping, group2 : ARRAY[1..maxgrp] OF spset; (* to store groups found *)
   timesseen, times2 : ARRAY[1..maxgrp] OF INTEGER; (* how often they're seen *)
   garbage : nodeptr;

  PROCEDURE gnu (VAR p : nodeptr);
     (* this and the following are do-it-yourself garbage collectors.
        Make a new node or pull one off the garbage list *)
  BEGIN (* gnu *)
     IF garbage <> nil
     THEN BEGIN
        p := garbage;
        garbage := garbage^.next;
        END
     ELSE NEW (p);
     p^.next := nil;
  END; (* gnu *)

  PROCEDURE chuck (p : nodeptr);
     (* collect garbage on p -- put it on front of garbage list *)
  BEGIN (* chuck *)
     p^.next := garbage;
     garbage := p;
  END; (* chuck *)

  PROCEDURE gdispose(p : nodeptr);
     (* go through tree throwing away nodes *)
     VAR q, r: nodeptr;
  BEGIN (* gdispose *)
      IF NOT p^.tip 
      THEN BEGIN 
         q := p^.next;
         WHILE(q <> p) DO BEGIN
            gdispose(q^.back);
            r := q;
            q := q^.next;         
            chuck(r); 
            END;
         chuck(q);
         END;
   END; (*  gdispose *)

   PROCEDURE setuptree;
      (* initialization of tree pointers, variables *)
      VAR i : INTEGER;
   BEGIN (* setuptree *)
      FOR i := 1 TO maxsp DO BEGIN
         gnu (treenode[i]);
         treenode[i]^.next := NIL;
         treenode[i]^.back := NIL;
         treenode[i]^.index := i;
         treenode[i]^.tip := FALSE;
         END;
   END; (* setuptree *)

   PROCEDURE treeread;
      (* read in user-defined tree and set it up *)
      VAR ch : CHAR;
         nextnode : INTEGER;

      PROCEDURE getch(VAR c: CHAR);
          (* get next nonblank character *)
      BEGIN (* getch *)
         REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, c);
         UNTIL (c <> ' ');
      END;  (* getch *)

      PROCEDURE addelement (VAR p : nodeptr; q: nodeptr);
         (* recursive procedure adds nodes to user-defined tree *)
         VAR  r : nodeptr;
            pfirst: nodeptr;
            i, n : INTEGER;
            found, notlast : BOOLEAN;
            str : ARRAY[1..nmlngth] OF CHAR;
      BEGIN (* addelement *)
        IF ch = '('
        THEN BEGIN
           nextnode := nextnode + 1;
           gnu(p);
           pfirst := p;
           notlast := true;
           WHILE notlast DO BEGIN
              gnu(p^.next);
              r := p^.next;
              r^.tip := false;
              getch(ch);
              IF NOT anerror THEN addelement(r^.back, r);
              IF r^.back <> NIL THEN p := r;
              IF ch = ')'
              THEN BEGIN
                 notlast := false;
                 REPEAT
                    getch(ch);
                 UNTIL (ch = ',') OR (ch = ')') OR (ch = ';');
                 END;
              END;
           p^.next := pfirst;
           p := pfirst;
           END
        ELSE BEGIN
           FOR i := 1 TO nmlngth DO str[i] := ' ';
           n := 1;
           REPEAT
              IF (ch = '_') THEN ch := ' ';
              str[n] := ch;
              IF EOLN(infile) THEN READLN(infile);
              READ(infile, ch);
              n := n + 1;
           UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') OR
                  (ch = ';') OR (n > nmlngth));
           IF ch = ':'
           THEN REPEAT
                 getch(ch);
              UNTIL (ch = ',') OR (ch = ')') OR (ch = ';');
           n := n -1;
           IF first 
           THEN BEGIN
              spp := spp + 1;
              IF spp > maxsp
              THEN BEGIN
                 WRITE(OUTPUT, 'ERROR: TOO MANY SPECIES. INCREASE',
                          ' CONSTANT maxsp');
                 anerror := TRUE;
                 END
              ELSE BEGIN
                 p := treenode[spp];
                 p^.tip := TRUE;
                 FOR i := 1 TO nmlngth DO name[spp,i] := ' ';
                 FOR i := 1 TO n DO name[spp,i] := str[i];
                 WRITE(OUTPUT, '  ');
                 FOR i := 1 TO n DO WRITE(OUTPUT, str[i]);
                 WRITELN(OUTPUT);
                 END;
              END
           ELSE BEGIN
              n := 1;
              REPEAT
                 found := TRUE;
                 FOR i := 1 TO nmlngth DO
                    found := found AND (str[i] = name[n,i]);
                 IF found 
                 THEN p := treenode[n]
                 ELSE n := n + 1;
              UNTIL (n > spp) OR found;
              IF n > spp 
              THEN BEGIN
                 anerror := TRUE;
                 WRITE(OUTPUT, 'ERROR: CANNOT FIND SPECIES: ');
                 FOR i := 1 TO nmlngth DO
                    WRITE(OUTPUT, str[i]);
                 WRITELN(OUTPUT);
                 END;
              END;
           END;
        IF p <> NIL THEN
        p^.back := q;
    END; (* addelement *)                                 

   BEGIN (* treeread *)
      getch(ch);
      addelement (root, NIL);
      READLN(infile);
      IF (outgrno > spp) 
      THEN BEGIN
         anerror := TRUE;
         WRITE(OUTPUT, 'ERROR IN OUTGROUP OPTION: SPECIES NUMBER',
                        outgrno:3, 'DOES NOT EXIST');
         END;
   END; (* treeread *)

   PROCEDURE reroot (outgroup : nodeptr);
      (* reorients tree, putting outgroup in desired position. *)
      VAR i: INTEGER;
         nroot: BOOLEAN;
         p, q: nodeptr;
   BEGIN (* reroot *)
      nroot := FALSE;
      p := root^.next;
      WHILE(p <> root) DO
         IF (outgroup^.back = p)
         THEN BEGIN
            nroot := TRUE;
            p := root;
            END
         ELSE p := p^.next;
      IF NOT nroot
      THEN BEGIN
         p := root;
         i := 0;
         WHILE ( p^.next <> root) DO BEGIN
            p := p^.next;
            i := i+1;
            END;
         IF (i = 2)
         THEN BEGIN
            root^.next^.back^.back := p^.back;
            p^.back^.back := root^.next^.back;
            q := root^.next;
            END
         ELSE BEGIN
            p^.next := root^.next;
            gnu(root^.next); 
            q := root^.next;
            gnu(q^.next); 
            p := q^.next;
            p^.next := root;
            q^.tip := false;
            p^.tip := false; 
            END;
         q^.back := outgroup;
         p^.back := outgroup^.back;
         outgroup^.back^.back := p;
         outgroup^.back := q;
         END;
   END; (* reroot *)

   PROCEDURE accumulate (r : nodeptr);
      VAR  q : nodeptr;

      PROCEDURE enterset;
         VAR s : spset;
            i, times : INTEGER;
            hashcode : REAL;
            done : BOOLEAN;
      BEGIN (* enterset *)
         s := r^.nodeset;
         IF (s <> [1..spp])
         THEN BEGIN
            times := 1;
            hashcode := 0.0;
            FOR i := spp DOWNTO 1 DO BEGIN
               hashcode := hashcode/2.0;
               IF (i IN s)
               THEN hashcode := hashcode + 0.5;
               END;
            i := TRUNC(0.8*maxgrp*hashcode) + 1;
            done := FALSE;
            WHILE NOT done DO BEGIN
               IF (s = grouping[i])
               THEN BEGIN
                  timesseen[i] := timesseen[i] + times;
                  done := TRUE;
                  END
               ELSE IF (timesseen[i] = 0) AND (grouping[i] = [])
                  THEN BEGIN
                     grouping[i] := s;
                     timesseen[i] := times;
                     done := TRUE;
                     END
               ELSE i := i + 1;
               IF i > maxgrp
               THEN BEGIN
                  WRITELN(OUTPUT, 'ERROR! HASH TABLE OVERFLOW');
                  done := TRUE;
                  END;
               END;
            END;
      END; (* enterset *)

   BEGIN (* accumulate *)
      IF r^.tip
      THEN r^.nodeset := [r^.index]
      ELSE BEGIN
         q := r^.next;
         WHILE(q <> r) DO BEGIN
            accumulate(q^.back);  
            q := q^.next;
            END;
         q := r^.next;
         r^.nodeset := [];  
         WHILE(q <>r) DO BEGIN
            r^.nodeset := r^.nodeset + q^.back^.nodeset;
            q := q^.next; 
            END;         
         END;
         IF NOT r^.tip
         THEN BEGIN
            IF r^.next^.next <> r
            THEN enterset;
            END
         ELSE enterset;
   END; (* accumulate *)

   PROCEDURE consensus;
      VAR i, n, n2 : INTEGER;
         fullset : spset;

      PROCEDURE compress;
         (* push all the nonempty subsets to the front end of their array *)
         VAR i, j : INTEGER;
      BEGIN (* compress *)
         i := 1;
         j := 1;
         REPEAT
            WHILE grouping[i] <> [] DO i := i + 1;
            IF j <= i
            THEN j := i + 1;
            WHILE ((grouping[j] = []) AND (j < maxgrp)) DO
               j := j + 1;
            IF (j < maxgrp)
            THEN BEGIN
               grouping[i] := grouping[j];
               timesseen[i] := timesseen[j];
               grouping[j] := [];
               timesseen[j] := 0;
               END
         UNTIL j = maxgrp;
         n := i - 1;
      END; (* compress *)

      PROCEDURE sort;
         (* Shell sort keeping grouping, timesseen in same order *)
         VAR gap, i, j, itemp : INTEGER;
           stemp : spset;
      BEGIN (* sort *)
         gap := n DIV 2;
         WHILE (gap > 0) DO BEGIN
            FOR i := gap TO n-1 DO BEGIN
               j := i - gap + 1;
               WHILE (j>0) DO BEGIN
                  IF timesseen[j] < timesseen[j+gap]
                  THEN BEGIN
                     stemp := grouping[j];
                     grouping[j] := grouping[j+gap];
                     grouping[j+gap] := stemp;
                     itemp := timesseen[j];
                     timesseen[j] := timesseen[j+gap];
                     timesseen[j+gap] := itemp;
                     END;
                  j := j - gap;
                  END;
               END;
            gap := gap DIV 2;
            END;
      END; (* sort *)

      PROCEDURE eliminate;
         (* eliminate groups incompatible with preceding ones *)
         VAR i, j : INTEGER;
      BEGIN (* eliminate *)
         FOR i := 2 TO n DO BEGIN
            FOR j := 1 TO i-1 DO
               IF timesseen[j] > 0
               THEN IF ((grouping[i] * grouping[j]) <> [])
                  AND ((grouping[i] - grouping[j]) <> [])
                  AND ((grouping[j] - grouping[i]) <> [])
                  AND ((NOT noroot) OR ((fullset-grouping[i])
                     * (fullset-grouping[j]) <> []))
                  THEN BEGIN
                     n2 := n2 + 1;
                     times2[n2] := timesseen[i];
                     group2[n2] := grouping[i];
                     timesseen[i] := 0;
                     grouping[i] := [];
                     END;
            END;
      END; (* eliminate *)

      PROCEDURE printset;
         (* print out a set of species *)
         VAR i, j, size : INTEGER;
      BEGIN (* printset *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'Set (species in order)   ');
         FOR i := 1 TO spp-25 DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '  How many times out of', ntrees:4);
         WRITELN(OUTPUT);
         FOR i := 1 TO n DO BEGIN
            IF (timesseen[i] > 0)
            THEN BEGIN
               size := 0;
               FOR j := 1 TO spp DO
                  IF j IN grouping[i]
                  THEN size := size + 1;
               IF (size <> 1) AND NOT (noroot AND (size = (spp - 1)))
               THEN BEGIN
                  FOR j := 1 TO spp DO BEGIN
                     IF j IN grouping[i]
                     THEN WRITE(OUTPUT, '*')
                     ELSE WRITE(OUTPUT, '.');
                     IF (j MOD 10) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  FOR j := 1 TO 23-spp DO WRITE(OUTPUT, ' ');
                  WRITELN(OUTPUT, '    ', timesseen[i]:4);
                  END;
               END;
            END;
      END; (* printset *)

      PROCEDURE reconstruct;
         (* reconstruct tree from the subsets *)
         VAR nextnode : INTEGER;
            s : spset;

         PROCEDURE bigsubset (VAR st : spset);
            (* find a maximal subset of st among the groupings *)
            VAR i : INTEGER;
               su : spset;
         BEGIN (* bigsubset *)
            su := [];
            FOR i := 1 TO n DO
              IF (st >= grouping[i]) AND (grouping[i] <> st)
              THEN IF (grouping[i] >= su) AND (grouping[i] <> su)
                 THEN su := grouping[i];
            st := su;
         END; (* bigsubset *)

         PROCEDURE recontraverse (VAR p : nodeptr; st : spset);
            (* traverse to add next node of consensus tree *)
            VAR i, j, k : INTEGER;
               found : BOOLEAN;
               tempset : spset;
               st2 : spset;
               q: nodeptr;

         BEGIN (* recontraverse *)
            k := 0;
            FOR i := 1 TO spp DO
               IF i IN st
               THEN BEGIN
                  k := k + 1;
                  j := i;
                  END;
            IF k = 1 
            THEN  p := treenode[j]
            ELSE BEGIN
               gnu(p);
               nextnode := nextnode + 1;
               p^.index := 0;
               FOR i := 1 TO n DO
                  IF grouping[i] = st
                  THEN p^.index := timesseen[i];
               tempset := st;
               q := p;
               st2 := st;
               IF tempset <> [] THEN bigsubset(tempset);
               WHILE (st2 <> []) AND (tempset <> []) DO BEGIN
                  gnu (q^.next);
                  q := q^.next;
                  recontraverse(q^.back, tempset);
                  q^.back^.back := q;
                  st2 := st2 - tempset;
                  tempset := st2;
                  found := FALSE;
                  i := 1;
                  WHILE (NOT found) AND (i <= n) DO BEGIN
                     IF grouping[i] = tempset THEN found := TRUE
                     ELSE i := i + 1;
                     END;
                  IF (tempset <> []) AND (NOT found)
                  THEN bigsubset(tempset); 
                  END;
               q^.next := p;
               END; 
         END; (* recontraverse *)

      BEGIN (* reconstruct *)
         nextnode := spp + 1;
         s := [1..spp];
         recontraverse (root, s);
      END; (* reconstruct *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 4;
         VAR tipy, i : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
           VAR q, first, last: nodeptr;         
              maxx : INTEGER;
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := 0;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               END
            ELSE BEGIN
               q := p^.next; 
               maxx := 0;
               WHILE(q <> p) DO BEGIN
                  coordinates (q^.back);
                  IF NOT q^.back^.tip
                  THEN IF q^.back^.xcoord > maxx
                     THEN maxx := q^.back^.xcoord;
                  q := q^.next;
                  END;
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := maxx + over;
               p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra, done : BOOLEAN;
               r, first, last : nodeptr;
               found: BOOLEAN;
         BEGIN (* drawline *)
            p := root;
            q := root;
            WRITE(OUTPUT, '  ');
            extra := FALSE;
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  found := FALSE;
                  r := p^.next;
                  WHILE (r <> p) AND NOT found DO 
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN 
                        q := r^.back;
                        found := TRUE;
                        END
                     ELSE r := r^.next;
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := p^.xcoord - q^.xcoord;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT DONE)
               THEN BEGIN
                  WRITE(OUTPUT, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     IF q^.index >= 10
                     THEN WRITE(OUTPUT, q^.index:2)
                     ELSE WRITE(OUTPUT, '-', q^.index:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(OUTPUT, '-');
                  END
               ELSE IF (NOT p^.tip) AND (last^.ycoord > i)
                      AND (first^.ycoord < i) 
                      AND ((NOT (i = p^.ycoord)) OR (p = root))
                  THEN BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END
                  ELSE FOR j := 1 TO n DO WRITE(OUTPUT, ' ');
               IF q <> p THEN p := q;
            UNTIL DONE;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'CONSENSUS TREE:');
         WRITELN(OUTPUT, 'the numbers at the forks indicate the number');
         WRITELN(OUTPUT, 'of times the group consisting of the species');
         WRITELN(OUTPUT, 'which are to the right of that fork occurred');
         WRITELN(OUTPUT, 'among the trees, out of ', ntrees:3, ' trees');
         tipy := 1;
         coordinates (root);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '  remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (though rerooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n : INTEGER;
            c : CHAR;
            q : nodeptr;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            FOR i := 1 TO nmlngth DO
               IF name[p^.index, i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := name[p^.index, i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            q := p^.next;
            WHILE(q <> p) DO BEGIN
               treeout (q^.back);
               q := q^.next;
               IF q <> p THEN  WRITE(treefile, ',');
            END;
            WRITE(treefile, ')');
            END;
         IF p^.tip
         THEN n := ntrees
         ELSE n := p^.index;
         IF p = root
         THEN WRITELN(treefile, ';')
         ELSE IF n > 99
            THEN WRITE(treefile, ':', n:3)
            ELSE IF n > 9
               THEN WRITE(treefile, ':', n:2)
               ELSE WRITE(treefile, ':', n:1);
      END; (* treeout *)

   BEGIN (* consensus *)
      fullset := [1..spp];
      n2 := 0;
      compress;
      sort;
      eliminate;
      compress;
      reconstruct;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, 'Sets included in the consensus tree');
      printset;
      FOR i := 1 TO n2 DO BEGIN
         grouping[i] := group2[i];
         timesseen[i] := times2[i];
         END;
      n := n2;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, 'Sets NOT included in consensus tree:');
      printset;
      WRITELN(OUTPUT);
      printree;
      IF trout
      THEN BEGIN
         treeout (root);
         WRITELN(OUTPUT, 'Tree also written onto file');
         WRITELN(OUTPUT);
         END;
   END; (* consensus *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   WRITELN(OUTPUT);
   WRITE(OUTPUT, 'Majority-rule and strict consensus tree');
   WRITELN(OUTPUT, ' program, version 3.3');
   WRITELN(OUTPUT);
   READ(infile, ntrees);
   anerror := FALSE;
   trout := FALSE;
   noroot := TRUE;
   outgropt := FALSE;
   outgrno := 1;
   numopts := 0;
   WHILE NOT EOLN(infile) DO BEGIN
      READ(infile, ch);
      IF (ch = 'O') OR (ch = 'R') OR (ch = 'Y')
      THEN CASE ch OF
         'O' : BEGIN
               outgropt := TRUE;
               numopts := numopts + 1;
               END;
         'R' : noroot := FALSE;
         'Y' : trout := TRUE
         END;
      END;
   READLN(infile);
   IF outgropt
   THEN READLN(infile, ch, outgrno);
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'treefile');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   didreroot := FALSE;
   first := TRUE;
   spp := 0;
   garbage := NIL;
   setuptree;
   FOR i := 1 TO maxgrp DO BEGIN
      grouping[i] := [];
      timesseen[i] := 0;
      END;
   WRITELN(OUTPUT, 'Species in order: ');
   WRITELN(OUTPUT);
   i := 1;
   WHILE  (NOT anerror) AND (i <= ntrees) DO BEGIN
      treeread;
      IF NOT anerror
      THEN BEGIN
         IF noroot
         THEN BEGIN
            reroot(treenode[outgrno]);
            didreroot := outgropt;
            END;
         accumulate(root);
         first := FALSE;
         gdispose(root);
         i := i + 1;
         END;
      END;
   WRITELN(OUTPUT);
   IF NOT anerror THEN consensus;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and maybe others too *)
      END;
END. (* main *)

