PROGRAM Contml (infile, treefile, OUTPUT);
  (* version 3.3. (c) Copyright 1986, 1990 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
     granted to copy and use this program provided no fee is charged for it
     and provided that this copyright notice is not removed. *)

CONST maxsp = 30;      (* maximum number of species *)
   maxsp2 = 58;        (* 2 x maxsp - 2 *)
   maxchr = 200;       (* maximum number of characters or alleles *)
   epsilon = 0.000001; (* small number *)
   smoothings = 4;     (* number of passes through smoothing algorithm *)
   namelength = 10;    (* number of characters max. in species name *)
   maxtrees = 10;      (* maximum number of user trees in KHT test *)

TYPE phenotype = ARRAY [1..maxchr] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;
   longer = ARRAY[0..5] OF INTEGER;

VAR infile, treefile : TEXT;
   numsp, numsp1, numsp2, loci, inseed, totalleles, df, outgrno : INTEGER;
   alleles, locus : ARRAY[1..maxchr] OF INTEGER;
   x : ARRAY[1..maxsp] OF phenotype;
   names : ARRAY[1..maxsp] OF naym;
   all, contchars, global, jumble,
      outgropt, trout, usertree, anerror : BOOLEAN;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getinput;
      (* read input file *)

      PROCEDURE getnums;
         (* read species numbers and number of characters or loci *)
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,'Continuous character Maximum Likelihood');
         WRITELN(OUTPUT, ' method version 3.3');
         WRITELN(OUTPUT);
         READ(infile, numsp, loci);
         WRITELN(OUTPUT, numsp:4, ' Populations, ', loci:4, ' Loci');
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: adjust CONSTants');
         IF loci > maxchr
         THEN WRITELN(OUTPUT, 'TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (numsp > maxsp) OR (loci > maxchr);
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         (* read options information *)
         VAR ch : CHAR;
            i, extranum : INTEGER;

         PROCEDURE inputoutgroup;
            (* input the number of the outgroup *)
         BEGIN (* inputoutgroup *)
            READ(infile, outgrno);
            IF (outgrno < 1) OR (outgrno > numsp)
            THEN BEGIN
               WRITELN(OUTPUT, 'BAD OUTGROUP NUMBER: ', outgrno:4);
               anerror := TRUE;
               END;
         END; (* inputoutgroup *)

      BEGIN (* getoptions *)
         all := FALSE;
         contchars := FALSE;
         global := FALSE;
         jumble := FALSE;
         outgropt := FALSE;
         outgrno := 1;
         trout := FALSE;
         usertree := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'C') OR (ch = 'G') OR (ch = 'J')
               OR (ch = 'O') OR (ch = 'U') OR (ch = 'Y')
            THEN CASE ch OF
                  'A' : all := TRUE;
                  'C' : contchars := TRUE;
                  'G' : global := TRUE;
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'U' : usertree := TRUE;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         FOR i := 1 TO extranum DO
            IF NOT anerror
            THEN BEGIN
               READLN(infile);
               READ(infile, ch);
               IF ch = 'J'
               THEN READ(infile, inseed);
               IF ch = 'O'
               THEN inputoutgroup;
               END;
         WRITELN(OUTPUT);
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT, 'Random number seed = ', inseed:8);
            WRITELN(OUTPUT);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
      END; (* getoptions *)

      PROCEDURE getalleles;
         (* set up number of alleles at loci *)
         VAR i, j : INTEGER;
      BEGIN (* getalleles *)
         IF contchars
         THEN BEGIN
            totalleles := loci;
            FOR i := 1 TO loci DO BEGIN
               locus[i] := i;
               alleles[i] := 1;
               END;
            df := loci;
            END
         ELSE BEGIN
            totalleles := 0;
            READLN(infile);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'Numbers of alleles at the loci:');
            WRITELN(OUTPUT, '------- -- ------- -- --- -----');
            WRITELN(OUTPUT);
            FOR i := 1 TO loci DO BEGIN
               IF EOLN(infile)
               THEN READLN(infile);
               READ(infile, alleles[i]);
               IF alleles[i] <= 0
               THEN BEGIN
                  WRITELN(OUTPUT, 'BAD NUMBER OF ALLELES: ',
                                  alleles[i]:4, ' AT LOCUS ', i:3);
                  anerror := TRUE;
                  END;
               FOR j := totalleles+1 TO totalleles+alleles[i] DO
                  locus[j] := i;
               totalleles := totalleles + alleles[i];
               WRITE(OUTPUT, alleles[i]:4);
               END;
            df := totalleles - loci;
            IF totalleles > maxchr
            THEN BEGIN
               WRITELN(OUTPUT);
               WRITELN(OUTPUT, 'TOO MANY LOCI OR ALLELES: adjust CONSTants');
               anerror := TRUE;
               END;
            END;
         WRITELN(OUTPUT);
      END; (* getalleles *)

      PROCEDURE getdata;
         (* read species data *)
         VAR i, j, k, l, m, n, p : INTEGER;
            sum : REAL;
      BEGIN (* getdata *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'Name');
         IF contchars
         THEN WRITELN(OUTPUT, '                       Phenotypes')
         ELSE WRITELN(OUTPUT, '                 Gene Frequencies');
         WRITE(OUTPUT, '----');
         IF contchars
         THEN WRITELN(OUTPUT, '                       ----------')
         ELSE WRITELN(OUTPUT, '                 ---- -----------');
         WRITELN(OUTPUT);
         IF NOT contchars
         THEN BEGIN
            FOR j := 1 TO namelength-8 DO WRITE(OUTPUT, ' ');
            WRITE(OUTPUT, 'locus:');
            p := 1;
            FOR j := 1 TO loci DO
               IF NOT anerror
               THEN BEGIN
                  IF all
                  THEN n := alleles[j]
                  ELSE n := alleles[j] - 1;
                  FOR k := 1 TO n DO BEGIN
                     WRITE(OUTPUT, j:10);
                     IF ((p MOD 6) = 0) AND (all OR (p < df))
                     THEN BEGIN
                        WRITELN(OUTPUT);
                        FOR l := 1 TO namelength-2 DO
                           WRITE(OUTPUT, ' ');
                        END;
                     p := p + 1;
                     END;
                  END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            END;
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               READLN(infile);
               FOR j := 1 TO namelength DO
                  IF NOT anerror
                  THEN BEGIN
                     READ(infile, names[i][j]);
                     anerror := anerror OR EOF(infile) OR EOLN(infile);
                     IF anerror
                     THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                        'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME')
                     ELSE WRITE(OUTPUT, names[i][j]);
                     END;
               m := 1;
               p := 1;
               FOR j := 1 TO loci DO
                  IF NOT anerror
                  THEN BEGIN
                     sum := 0.0;
                     IF contchars
                     THEN n := 1
                     ELSE IF all
                        THEN n := alleles[j]
                        ELSE n := alleles[j] - 1;
                     FOR k := 1 TO n DO BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, x[i][m]);
                        sum := sum + x[i][m];
                        WRITE(OUTPUT, x[i][m]:10:5);
                        IF (NOT contchars) AND (x[i][m] < 0.0)
                        THEN BEGIN
                           WRITELN(OUTPUT);
                           WRITE(OUTPUT, 'LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': AN ALLELE');
                           WRITELN(OUTPUT, ' FREQUENCY IS NEGATIVE');
                           anerror := TRUE;
                           END;
                        IF ((p MOD 6) = 0) AND (all OR (p < df))
                        THEN BEGIN
                           WRITELN(OUTPUT);
                           FOR l := 1 TO namelength DO
                              WRITE(OUTPUT, ' ');
                           END;
                        IF NOT contchars
                        THEN BEGIN
                           IF x[i][m] >= epsilon
                           THEN x[i][m] := SQRT(x[i][m]);
                           END;
                        p := p + 1;
                        m := m + 1;
                        END;
                     IF all AND (ABS(sum - 1.0) > epsilon)
                     THEN BEGIN
                        WRITELN(OUTPUT);
                        WRITELN(OUTPUT, 'LOCUS', j:3, ' IN SPECIES',
                              i:3, ': FREQUENCIES DO NOT ADD UP TO 1');
                        anerror := TRUE;
                        END;
                     IF (NOT all) AND (NOT contchars)
                     THEN BEGIN
                        x[i][m] := 1.0 - sum;
                        IF x[i][m] >= epsilon
                        THEN x[i][m] := SQRT(x[i][m]);
                        IF x[i][m] < 0.0
                        THEN BEGIN
                           IF x[i][m] > -epsilon
                           THEN BEGIN
                              FOR l := 1 TO m-1 DO
                                 x[i][l] := x[i][l]/SQRT(sum);
                              x[i][m] := 0.0;
                              END
                           ELSE BEGIN
                              WRITELN(OUTPUT);
                              WRITELN(OUTPUT, ' LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': FREQUENCIES ADD UP TO MORE THAN 1');
                              anerror := TRUE;
                              END;
                           END;
                        m := m + 1;
                        END;
                     END;
               WRITELN(OUTPUT);
               END;
         READLN(infile);
         WRITELN(OUTPUT);
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      IF NOT anerror
      THEN getoptions;
      IF NOT anerror
      THEN getalleles;
      IF NOT anerror
      THEN getdata;
   END; (* getinput *)

   PROCEDURE maketree;
      (* construct the tree *)
      TYPE nodeptr = ^node;
         node = RECORD
            next, back : nodeptr;
            tip        : BOOLEAN;
            number     : INTEGER;
            view       : phenotype;
            nayme      : naym;
            v, deltav  : REAL;
            bigv, dist : REAL;
            xcoord     : REAL;
            ycoord, ymin, ymax : INTEGER
            END;
         tree = RECORD
            nodep      : ARRAY[1..maxsp2] OF nodeptr;
            likelihood : REAL;
            start      : nodeptr
            END;

      VAR curtree, priortree, bestree : tree;
         i, j, k, nextsp, numtrees, num, which, maxwhich : INTEGER;
         succeeded : BOOLEAN;
         maxlogl, sum, sum2, sd : REAL;
         logl : ARRAY [1..maxtrees] OF REAL;
         logf : ARRAY [1..maxtrees] OF ARRAY [1..maxchr] OF REAL;

      PROCEDURE setuptree (VAR a : tree);
         (* initialize a tree *)
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            likelihood := -99999.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
         (* hook together two nodes *)
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate (VAR t : tree) : REAL;
         (* evaluate likelihood of a tree *)
         VAR i : INTEGER;
            sum : REAL;

         PROCEDURE sumlikely (p, q : nodeptr);
            (* sum contribution to likelihood over forks in tree *)
            VAR i : INTEGER;
               term, sumsq, vee : REAL;
         BEGIN (* sumlikely *)
            IF NOT p^.tip
            THEN sumlikely (p^.next^.back, p^.next^.next^.back);
            IF NOT q^.tip
            THEN sumlikely (q^.next^.back, q^.next^.next^.back);
            IF p^.back = q
            THEN vee := p^.v
            ELSE vee := p^.v + q^.v;
            vee := vee + p^.deltav + q^.deltav;
            IF vee <= 1.0E-10
            THEN BEGIN
               WRITELN(OUTPUT, 'ERROR: CHECK FOR TWO IDENTICAL ',
                         ' SPECIES AND ELIMINATE ONE FROM THE DATA');
               anerror := TRUE;
               END;
            sumsq := 0.0;
            IF usertree AND (which <= maxtrees)
            THEN FOR i := 1 TO loci DO
                  logf[which][i] := logf[which][i]
                     - (alleles[i]-1) * LN(vee)/2.0;
            FOR i := 1 TO totalleles DO BEGIN
               term := SQR(p^.view[i] - q^.view[i]);
               IF usertree AND (which <= maxtrees)
               THEN logf[which][locus[i]] := logf[which][locus[i]]
                     - term/(2.0*vee);
               sumsq := sumsq + term;
               END;
            sum := sum - df*LN(vee)/2.0 - sumsq/(2.0*vee);
         END; (* sumlikely *)

      BEGIN (* evaluate *)
         sum := 0.0;
         IF usertree AND (which <= maxtrees)
         THEN FOR i := 1 TO loci DO
               logf[which][i] := 0.0;
         sumlikely (t.start, t.start^.back);
         IF usertree
         THEN BEGIN
            logl[which] := sum;
            IF which = 1
            THEN BEGIN
               maxwhich := 1;
               maxlogl := sum;
               END
            ELSE IF sum > maxlogl
               THEN BEGIN
                  maxwhich := which;
                  maxlogl := sum;
                  END;
            END;
         t.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE update (p : nodeptr);
         (* update branch lengths around a node *)
         VAR q : nodeptr;
            negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            (* compute distances among three neighbors of a node *)
            VAR i : INTEGER;

            FUNCTION distance (p, q : nodeptr) : REAL;
               (* distance between two nodes *)
               VAR i : INTEGER;
                  sum : REAL;
            BEGIN (* distance *)
               sum := 0.0;
               FOR i := 1 TO totalleles DO
                  sum := sum + SQR(p^.view[i] - q^.view[i]);
               distance := sum;
            END; (* distance *)

         BEGIN (* makedists *)
            FOR i := 1 TO 3 DO BEGIN
               q := p^.next;
               p^.dist := distance (p^.back, q^.back);
               p := q;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            (* make new branch length *)
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.bigv := (p^.dist + r^.dist - q^.dist)/(2*df);
               p^.back^.bigv := p^.bigv;
               IF p^.bigv < p^.back^.deltav THEN negatives := TRUE;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            (* iterate branch lengths if some are to be zero *)
            VAR q, r, temp : nodeptr;
               i, j : INTEGER;
               f1, f2, vtot : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO smoothings DO BEGIN
               FOR j := 1 TO 3 DO BEGIN
                  vtot := q^.bigv + r^.bigv;
                  IF vtot > 0.0
                  THEN f1 := q^.bigv/vtot
                  ELSE f1 := 0.5;
                  f2 := 1.0 - f1;
                  p^.bigv := (f1*r^.dist + f2*p^.dist - f1*f2*q^.dist)/df;
                  p^.bigv := p^.bigv - vtot*f1*f2;
                  IF p^.bigv < p^.back^.deltav
                  THEN p^.bigv := p^.back^.deltav;
                  p^.back^.bigv := p^.bigv;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  END;
               END;
         END; (* correctv *)

         PROCEDURE littlev (p : nodeptr);
            (* remove part of it that belongs to other barnches *)
            VAR i : INTEGER;
         BEGIN (* littlev *)
            FOR i := 1 TO 3 DO BEGIN
               WITH p^ DO BEGIN
                  v := bigv - back^.deltav;
                  back^.v := v;
                  END;
               p := p^.next;
               END;
         END; (* littlev *)

         PROCEDURE nuview (p : nodeptr);
            (* renew information about subtrees *)
            VAR i, j : INTEGER;
               q, r, a, b, temp : nodeptr;
               v1, v2, vtot, f1, f2 : REAL;
         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               a := q^.back;
               b := r^.back;
               v1 := q^.bigv;
               v2 := r^.bigv;
               vtot := v1 + v2;
               IF vtot > 0.0
               THEN f1 := v2/vtot
               ELSE f1 := 0.5;
               f2 := 1.0 - f1;
               FOR j := 1 TO totalleles DO
                  p^.view[j] := f1*a^.view[j] + f2*b^.view[j];
               p^.deltav := v1*f1;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF negatives THEN correctv (p);
            littlev (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
         (* go through tree getting new branch lengths and views *)
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr);
         (* put p and q together and iterate info. on resulting tree *)
         VAR i : INTEGER;
      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         (* make a copy of a tree *)
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            (* make a copy of a node *)
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
         (* initialize branch lengths and views in a tip *)
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            view := x[m];
            nayme := names[m];
            deltav := 0.0;
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         (* initialize and hook up a new tip *)
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[nextsp + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
         (* make and initialize a three-species tree *)
      BEGIN (* buildsimpletree *)
         setuptip (enterorder[1], t);
         setuptip (enterorder[2], t);
         hookup (t.nodep[enterorder[1]], t.nodep[enterorder[2]]);
         buildnewtip (enterorder[3], t);
         insert(t.nodep[enterorder[3]]^.back, t.nodep[enterorder[1]]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
         (* traverse through a tree, finding best place to add p *)
      BEGIN (* addtraverse *)
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         copy (priortree, curtree);
         IF (NOT q^.tip) AND contin
         THEN BEGIN
            addtraverse (p, q^.next^.back, contin);
            addtraverse (p, q^.next^.next^.back, contin);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p : nodeptr);
         (* rearranges the tree, globally or locally *)
         VAR q, r : nodeptr;

         PROCEDURE remove (VAR p, q : nodeptr);
            (* remove p and record in q where it was *)
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update (q);
            update (q^.back);
         END; (* remove *)

      BEGIN (* rearrange *)
         IF (NOT p^.tip) AND (NOT p^.back^.tip)
         THEN BEGIN
            r := p^.next^.next;
            remove (r, q);
            copy (curtree, priortree);
            addtraverse (r, q^.next^.back, global AND (nextsp = numsp));
            addtraverse (r, q^.next^.next^.back, global AND (nextsp = numsp));
            copy (bestree, curtree);
            IF global AND (nextsp = numsp)
            THEN WRITE(OUTPUT, '.');
            IF global AND (nextsp = numsp) AND (NOT succeeded)
            THEN BEGIN
               IF r^.back^.tip
               THEN BEGIN
                  r := r^.next^.next;
                  remove (r, q);
                  q := q^.back;
                  copy (curtree, priortree);
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  q := q^.back;
                  IF NOT q^.tip
                  THEN BEGIN
                     addtraverse (r, q^.next^.back, TRUE);
                     addtraverse (r, q^.next^.next^.back, TRUE);
                     END;
                  copy (bestree, curtree);
                  END;
               END;
            END;
         IF NOT p^.tip
         THEN BEGIN
            rearrange (p^.next^.back);
            rearrange (p^.next^.next^.back);
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 60;
         VAR tipy, i : INTEGER;
             scale, tipmax : REAL;

         PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
            (* establishes coordinates of nodes *)
            VAR q, first, last: nodeptr;         
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := lengthsum;
               p^.ycoord := tipy;
               p^.ymin := tipy;
               p^.ymax := tipy;
               tipy := tipy + down;
               IF lengthsum > tipmax
               THEN tipmax := lengthsum;
               END
            ELSE BEGIN
               q := p^.next; 
               REPEAT
                  coordinates (q^.back, lengthsum+q^.v);
                  q := q^.next;
               UNTIL ((p <> curtree.start^.back) AND (p = q))
                  OR ((p = curtree.start^.back) AND (p^.next = q));
               first := p^.next^.back; 
               q := p;
               WHILE q^.next <> p DO 
                  q := q^.next;
               last := q^.back;
               p^.xcoord := lengthsum; 
               p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
               p^.ymin := first^.ymin;
               p^.ymax := last^.ymax;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
               extra : BOOLEAN;
               r, first, last : nodeptr;
               done : BOOLEAN;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            q := curtree.start^.back;
            IF (i = p^.ycoord) AND (p = curtree.start^.back)
            THEN WRITE(OUTPUT, '-')
            ELSE WRITE(OUTPUT, ' ');
            extra := FALSE;
            REPEAT
               IF NOT p^.tip
               THEN BEGIN
                  r := p^.next;
                  done := FALSE;
                  REPEAT
                     IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
                     THEN BEGIN
                        q := r^.back;
                        done := TRUE;
                        END;
                     r := r^.next;
                  UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
                     OR ((p = curtree.start^.back) AND (r = p^.next)));
                  first := p^.next^.back; 
                  r := p;
                  WHILE r^.next <> p DO 
                     r := r^.next;
                  last := r^.back;
                  IF p = curtree.start^.back THEN last := p^.back;
                  END;
               done := (p^.tip) OR (p = q);
               n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
               IF (n < 2) AND (NOT q^.tip) THEN n := 2;
               IF extra
               THEN BEGIN
                  n := n - 1;
                  extra := FALSE;
                  END;
               IF (q^.ycoord = i) AND (NOT done)
               THEN BEGIN
                  WRITE(OUTPUT, '+');
                  IF NOT (q^.tip)
                  THEN BEGIN
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     IF q^.number-numsp >= 10
                     THEN WRITE(OUTPUT, q^.number-numsp:2)
                     ELSE WRITE(OUTPUT, '-', q^.number-numsp:1);
                     extra := TRUE;
                     END
                  ELSE FOR j := 1 TO n-1 DO WRITE(OUTPUT, '-');
                  END
               ELSE IF (NOT p^.tip)
                  THEN BEGIN
                     IF (i = p^.ycoord) AND (p = curtree.start^.back)
                     THEN BEGIN
                        WRITE(OUTPUT, p^.number-numsp:1);
                        n := n - 1;
                        END;
                     IF (last^.ycoord > i) AND (first^.ycoord < i)
                        AND ((NOT (i = p^.ycoord)) OR (p = curtree.start))
                     THEN BEGIN
                        WRITE(OUTPUT, '!');
                        FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                        END
                     ELSE FOR j := 1 TO n DO WRITE(OUTPUT, ' ');
                     END;
               IF q <> p THEN p := q;
            UNTIL done;
            IF (p^.ycoord = i) AND (p^.tip)
            THEN FOR j := 1 TO namelength DO WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         WRITELN(OUTPUT);
         tipy := 1;
         tipmax := 0.0;
         coordinates (curtree.start^.back, 0.0);
         scale := over/(tipmax + 0.0001);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;
         (* print out branch lengths etc. *)
         VAR chilow, chihigh : REAL;

         PROCEDURE describe (p : nodeptr);
            (* print out information for one branch *)
            VAR i : INTEGER;
               q : nodeptr;
               bigv, delta : REAL;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, (q^.number - numsp):3, '       ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, (p^.number - numsp):4, '      ');
            WRITE(OUTPUT, q^.v:15:5);
            delta := p^.deltav + p^.back^.deltav;
            bigv := p^.v + delta;
            WRITELN(OUTPUT, '   (', chilow*bigv-delta:12:5, ',',
               chihigh*bigv-delta:12:5,')');
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'remember: ');
         IF outgropt
         THEN WRITE(OUTPUT, '(although rooted by outgroup) ');
         WRITELN(OUTPUT, 'this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'Ln Likelihood = ', curtree.likelihood:11:5);
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'examined ',numtrees:4, ' trees');
            END;
         IF df = 1
         THEN BEGIN
            chilow := 0.000982;
            chihigh := 5.02389;
            END
         ELSE IF df = 2
            THEN BEGIN
               chilow := 0.05064;
               chihigh := 7.3777;
               END
            ELSE BEGIN
               chilow := 1.0 - 2.0/(9*df);
               chihigh := chilow;
               chilow := chilow - 1.95996 * SQRT(2.0/(9*df));
               chihigh := chihigh + 1.95996 * SQRT(2.0/(9*df));
               chilow := chilow * chilow * chilow;
               chihigh := chihigh * chihigh * chihigh;
               END;
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Between       And             Length');
         WRITELN(OUTPUT, '      Approx. Confidence Limits');
         WRITE(OUTPUT, ' -------       ---             ------');
         WRITELN(OUTPUT, '      ------- ---------- ------');
         describe (curtree.start^.back^.next^.back);
         describe (curtree.start^.back^.next^.next^.back);
         describe (curtree.start);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* summarize *)

      PROCEDURE treeout (p : nodeptr);
         (* write out file with representation of final tree *)
         VAR i, n, w : INTEGER;
            c : CHAR;
            x : REAL;
      BEGIN (* treeout *)
         IF p^.tip
         THEN BEGIN
            n := 0;
            FOR i := 1 TO namelength DO
               IF p^.nayme[i] <> ' '
               THEN n := i;
            FOR i := 1 TO n DO BEGIN
               c := p^.nayme[i];
               IF c = ' '
               THEN c := '_';
               WRITE(treefile, c);
               END;
            END
         ELSE BEGIN
            WRITE(treefile, '(');
            treeout (p^.next^.back);
            WRITE(treefile, ',');
            treeout (p^.next^.next^.back);
            IF p = curtree.start^.back
            THEN BEGIN
               WRITE(treefile, ',');
               treeout (p^.back);
               END;
            WRITE(treefile, ')');
            END;
         x := p^.v;
         IF x > 0.0
         THEN w := TRUNC(0.4343*LN(x))
         ELSE IF x = 0.0
            THEN w := 0
            ELSE w := TRUNC(0.4343*LN(-x)) + 1;
         IF w < 0 THEN w := 0;
         IF p = curtree.start^.back
         THEN WRITELN(treefile, ';')
         ELSE WRITE(treefile, ':', x:w+7:5);
      END; (* treeout *)

      PROCEDURE treeread;
         (* read in a user tree *)
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* skip forward in user tree until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            (* add one node to the user tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..namelength] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               READ(infile, ch);
               END
            ELSE BEGIN
               FOR i := 1 TO namelength DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') 
                        OR (n > namelength));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO namelength DO
                     found := found AND (str[i] = names[n][i]);
                  IF NOT found
                  THEN n := n + 1;
               UNTIL (n > numsp) OR found;
               IF n > numsp
               THEN BEGIN
                  WRITE(OUTPUT, 'CANNOT FIND SPECIES: ');
                  FOR i := 1 TO namelength DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               hookup (curtree.nodep[n], p);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         (* evaluate user-defined tree, iterating branch lengths *)
         VAR i : INTEGER;
            dummy : REAL;

         PROCEDURE initrav (p : nodeptr);
            (* traverse to initialize *)

            PROCEDURE nodeinit (p : nodeptr);
               (* initialize a node *)
               VAR q, r : nodeptr;
                  i : INTEGER;
            BEGIN (* nodeinit *)
               IF NOT p^.tip
               THEN BEGIN
                  q := p^.next^.back;
                  r := p^.next^.next^.back;
                  nodeinit (q);
                  nodeinit (r);
                  FOR i := 1 TO totalleles DO
                     p^.view[i] := 0.5 * q^.view[i] + 0.5 * r^.view[i];
                  p^.v := 0.1;
                  p^.back^.v := 0.1;
                  END;
            END; (* nodeinit *)

         BEGIN (* initrav *)
            IF p^.tip
            THEN nodeinit (p^.back)
            ELSE BEGIN
               initrav (p^.next^.back);
               initrav (p^.next^.next^.back);
               END;
         END; (* initrav *)

      BEGIN (* treevaluate *)
         initrav (curtree.start);
         initrav (curtree.start^.back);
         FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         READLN(infile, numtrees);
         WRITE(OUTPUT, 'User-defined tree');
         IF numtrees > 1
         THEN WRITE(OUTPUT, 's');
         WRITELN(OUTPUT);
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         FOR which := 1 TO numtrees DO
            IF NOT anerror
            THEN BEGIN
               treeread;
               treevaluate;
               printree;
               summarize;
               END;
         IF (numtrees > 1) AND (loci > 1)
         THEN BEGIN
           WRITE(OUTPUT, 'Tree    Ln L      Diff Ln L     Its S.D.');
           WRITELN(OUTPUT, '   Significantly worse?');
           WRITELN(OUTPUT);
           IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
           FOR i := 1 TO num DO BEGIN
              WRITE(OUTPUT, i:3, logl[i]:12:5);
              IF maxwhich = i
              THEN WRITELN(OUTPUT, '  <------ best')
              ELSE BEGIN
                 sum := 0.0;
                 sum2 := 0.0;
                 FOR j := 1 TO loci DO BEGIN
                    sum := sum + logf[maxwhich][j]-logf[i][j];
                    sum2 := sum2 + SQR(logf[maxwhich][j]-logf[i][j]);
                    END;
                 sd := SQRT((loci/(loci-1))*(sum2-SQR(sum/loci)));
                 WRITE(OUTPUT, (logl[i]-maxlogl):12:5, sd:12:4);
                 IF sum > 1.95996*sd
                 THEN WRITELN(OUTPUT, '           Yes')
                 ELSE WRITELN(OUTPUT, '            No');
                 END;
              END;
           WRITELN(OUTPUT);
           WRITELN(OUTPUT);
           END;
         END
      ELSE BEGIN
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         FOR i := 1 TO numsp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO numsp DO BEGIN
            j := TRUNC(randum(seed)*numsp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         nextsp := 3;
         buildsimpletree (curtree);
         curtree.start := curtree.nodep[enterorder[1]];
         numtrees := 1;
         nextsp := 4;
         WRITELN(OUTPUT, 'Adding species:');
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[1]][i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[2]][i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '   ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, names[enterorder[3]][i]);
         WRITELN(OUTPUT);
         WHILE (nextsp <= numsp) AND (NOT anerror) DO BEGIN
            buildnewtip (enterorder[nextsp], curtree);
            copy (curtree, priortree);
            bestree.likelihood := -99999.0;
            addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                                   curtree.start^.back, TRUE);
            copy (bestree, curtree);
            WRITE(OUTPUT, '   ');
            FOR j := 1 TO namelength DO
               WRITE(OUTPUT, names[enterorder[nextsp]][j]);
            WRITELN(OUTPUT);
            IF global AND (nextsp = numsp)
            THEN BEGIN
               WRITELN(OUTPUT, 'Doing global rearrangements');
               WRITE(OUTPUT, '  !');
               FOR i := 1 TO numsp-2 DO WRITE(OUTPUT, '-');
               WRITELN(OUTPUT, '!');
               WRITE(OUTPUT, '   ');
               END;
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start^.back);
               IF global AND (nextsp = numsp)
               THEN WRITELN(OUTPUT);
               END;
            IF (nextsp = numsp)
            THEN BEGIN
               curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               IF trout AND (nextsp = numsp)
               THEN BEGIN
                  treeout (curtree.start^.back);
                  WRITELN(OUTPUT, 'Tree also written onto file');
                  WRITELN(OUTPUT);
                  END;
               END;
            nextsp := nextsp + 1;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   anerror := FALSE;
   getinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  Turbo Pascal only *)
      REWRITE(treefile);
      END;
   IF NOT anerror
   THEN maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  Turbo Pascal and some others too *)
      END;
END.

