PROGRAM DNADist (infile, outfile, OUTPUT);
   (* version 3.31. (c) Copyright 1989, 1990 by Joseph Felsenstein.  Written by
      Joseph Felsenstein.  Permission is granted to copy and use this program
      provided no fee is charged for it and provided that this copyright notice
      is not removed. *)

   CONST maxsp = 20;      (* maximum number of species *)
      maxsites = 1500;    (* maximum number of sites *)
      maxpatterns = 512;  (* max number of different site patterns *)
      maxcategories = 9;  (* maximum number of site types *)
      iterations = 100;   (* number of iterates of EM for each distance *)
      cycles = 30;        (* number of cycles to solve each EM step *)
      nmlngth = 10;       (* number of characters max. in species name *)
      epsilon = 0.00001;  (* used in makedist *)

   TYPE base = (A, C, G, T);
      sitelike = ARRAY[A..T] OF REAL;
      phenotype = ARRAY [1..maxpatterns] OF sitelike;
      sequence = ARRAY[1..maxsp] OF ARRAY [1..maxsites] OF CHAR;
      naym = ARRAY[1..nmlngth] OF CHAR;
      longer = ARRAY [0..5] OF INTEGER;
      nodeptr = ^node;
      node = RECORD
         x          : phenotype;
         nayme      : naym
         END;

   VAR infile, outfile : TEXT;
      rr, numsp, sites, endsite, categs, inseed, repsno, weightsum : INTEGER;
      anerror, freqsfrom, boot, halfjack, jukes,
         kimura, lower, ml, weights : BOOLEAN;
      nodep : ARRAY[1..maxsp] OF nodeptr;
      xi, xv, ttratio, freqa, freqc, freqg, freqt, freqr,
         freqy, freqar, freqcy, freqgr, freqty, fracchange, sumrates : REAL;
      y : sequence;
      seed : longer;
      category, oldweight, weight, alias, ally,
         location : ARRAY[1..maxsites] OF INTEGER;
      rate : ARRAY[1..maxcategories] OF REAL;
      d : ARRAY[1..maxsp] OF ARRAY[1..maxsp] OF REAL;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   PROCEDURE getinput;
      (* read input file *)

      PROCEDURE getnums;
         (* input number of species, number of sites *)
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,'Nucleic acid sequence Distance Matrix program,');
         WRITELN(OUTPUT, ' version 3.3');
         WRITELN(OUTPUT);
         READ(infile, numsp, sites);
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: adjust CONSTants');
         IF sites > maxsites
         THEN WRITELN(OUTPUT, 'TOO MANY SITES: adjust CONSTants');
         WRITELN(OUTPUT, numsp:3, ' species, ', sites:4, ' sites');
         anerror := (numsp > maxsp) OR (sites > maxsites);
      END; (* getnums *)

      FUNCTION letter (ch : CHAR) : BOOLEAN;
        (* tests whether ch is a letter
           -- works for both ASCII and EBCDIC codes *)
      BEGIN (* letter *)
       letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
             OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
             OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
      END; (* letter *)

      PROCEDURE uppercase (VAR ch : CHAR);
         (* convert ch to upper case -- either ASCII or EBCDIC *)
      BEGIN (* uppercase *)
         IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                         OR  ('a' < 'A') AND (ch < 'A'))
         THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
      END; (* uppercase *)

      PROCEDURE getoptions;
         (* read options information *)
         VAR ch : CHAR;
            i, j, extranum, origseed : INTEGER;

         PROCEDURE inputcategories;
            (* reads the categories for each site *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputcategories *)
            FOR i := 1 TO nmlngth DO READ(infile, ch);
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL (ch <> ' ');
               category[i] := ORD(ch) - ORD('0');
               END;
            READLN(infile);
         END; (* inputcategories *)

         PROCEDURE printcategories;
            (* print out list of categories of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printcategories *)
            WRITELN(OUTPUT, 'Rate categories');
            WRITELN(OUTPUT);
            FOR i := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO sites DO BEGIN
               WRITE(OUTPUT, category[i]:1);
               IF (i MOD 60) = 0
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR j := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
                  END
               ELSE IF (i MOD 10) = 0
                  THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printcategories *)

         PROCEDURE inputweights;
            (* input the character weights, 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            weightsum := 0;
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               oldweight[i] := 1;
               IF (ch = '0') OR (ch = '1')
               THEN oldweight[i] := ORD(ch) - ORD('0')
               ELSE BEGIN
                  WRITELN(OUTPUT, 'BAD WEIGHT CHARACTER: ', ch,
                     ' -- WEIGHTS IN DNADIST MUST BE 0 OR 1');
                  anerror := TRUE;
                  END;
               weightsum := weightsum + oldweight[i];
               END;
            weights := TRUE;
            READLN(infile);
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'Sites are weighted as follows:');
            FOR i := 1 TO sites DO BEGIN
               IF ((i-1) MOD 60) = 0
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR j := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
                  END;
               WRITE(OUTPUT, oldweight[i]:1);
               IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
               THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* getoptions *)
         categs := 1;
         rate[1] := 1.0;
         FOR i := 1 TO sites DO category[i] := 1;
         freqsfrom := FALSE;
         boot := FALSE;
         halfjack := FALSE;
         jukes := FALSE;
         kimura := TRUE;
         lower := FALSE;
         ml := FALSE;
         weights := FALSE;
         FOR i := 1 TO sites DO oldweight[i] := 1;
         weightsum := sites;
         repsno := 1;
         ttratio := 2.0;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'B') OR (ch = 'C') OR (ch = 'F') OR (ch = 'H')
               OR (ch = 'J') OR (ch = 'K') OR (ch = 'L') OR (ch = 'M')
               OR (ch = 'T') OR (ch = 'W')
            THEN CASE ch OF
                  'B' : BEGIN
                        boot := TRUE;
                        extranum := extranum + 1;
                        END;
                  'C' : extranum := extranum + 1;
                  'F' : freqsfrom := TRUE;
                  'H' : BEGIN
                        halfjack := TRUE;
                        extranum := extranum + 1;
                        END;
                  'J' : BEGIN
                        jukes := TRUE;
                        kimura := FALSE;
                        ml := FALSE;
                        END;
                  'K' : BEGIN
                        kimura := TRUE;
                        jukes := FALSE;
                        ml := FALSE;
                        END;
                  'L' : lower := TRUE;
                  'M' : BEGIN
                        ml := TRUE;
                        jukes := FALSE;
                        kimura := FALSE;
                        END;
                  'T' : extranum := extranum + 1;
                  'W' : extranum := extranum + 1
                  END
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(OUTPUT);
                  WRITELN(OUTPUT, 'MAKE SURE YOU WANTED OPTION CHARACTER: ',
                          ch);
                  WRITELN(OUTPUT, '(It was passed along to output file)');
                  END;
            END;
         READLN(infile);
         WRITELN(OUTPUT);
         IF jukes
         THEN WRITELN(OUTPUT, '  Jukes-Cantor Distance');
         IF kimura
         THEN WRITELN(OUTPUT, '  Kimura 2-parameter Distance');
         IF ml
         THEN WRITELN(OUTPUT, '  Maximum Likelihood Distance');
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            anerror := anerror OR (NOT ((ch = 'B') OR (ch = 'H') OR (ch = 'C')
                            OR (ch = 'T') OR (ch = 'W')));
            IF anerror
            THEN WRITELN(OUTPUT, 'ERROR: INCORRECT AUXILIARY',
                  ' OPTIONS LINE WHICH STARTS WITH ', ch);
            IF (ch = 'B') OR (ch = 'H')
            THEN BEGIN
               READLN(infile, repsno, inseed);
               origseed := inseed;
               FOR j := 0 TO 5 DO seed[j] := 0;
               j := 0;
               REPEAT
                  seed[j] := inseed MOD 64;
                  inseed := inseed DIV 64;
                  j := j + 1;
               UNTIL inseed = 0;
               END;
            IF ch = 'C'
            THEN BEGIN
               READ(infile, categs);
               FOR j := 1 TO categs DO BEGIN
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, rate[j]);
                  END;
               READLN(infile);
               END;
            IF ch = 'T'
            THEN BEGIN
               IF jukes
               THEN BEGIN
                  WRITE(OUTPUT, 'WARNING: CANNOT SET TRANSITION/');
                  WRITE(OUTPUT, 'TRANSVERSION RATIO FOR');
                  WRITELN(OUTPUT, ' JUKES-CANTOR DISTANCE');
                  anerror := TRUE;
                  END
               ELSE READLN(infile, ttratio);
               END;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF boot OR halfjack
         THEN BEGIN
            WRITELN(OUTPUT);
            IF boot THEN WRITE(OUTPUT, 'Bootstrap: ');
            IF halfjack THEN WRITE(OUTPUT, 'Delete-half jackknife: ');
            WRITELN(OUTPUT, repsno:5, ' replicates');
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'Random number seed =', origseed:8);
            WRITELN(OUTPUT);
            END;
         IF categs > 1
         THEN BEGIN
            IF jukes OR kimura
            THEN BEGIN
               WRITE(OUTPUT, 'WARNING: CANNOT HAVE DIFFERENT');
               WRITE(OUTPUT, ' RATE CATEGORIES WITH');
               WRITELN(OUTPUT, ' JUKES-CANTOR OR KIMURA DISTANCES');
               anerror := TRUE;
               END
            ELSE BEGIN
               WRITELN(OUTPUT);
               WRITELN(OUTPUT, 'Site category   Rate of change');
               WRITELN(OUTPUT);
               FOR i := 1 TO categs DO
                  WRITELN(OUTPUT, i:12, rate[i]:13:3);
               WRITELN(OUTPUT);
               inputcategories;
               printcategories;
               END;
            END;
         IF (jukes OR kimura) AND freqsfrom
         THEN BEGIN
            WRITE(OUTPUT, 'WARNING: CANNOT USE EMPIRICAL BASE FREQUENCIES');
            WRITELN(OUTPUT, ' WITH JUKES-CANTOR OR KIMURA DISTANCES');
            anerror := TRUE;
            END;
         IF jukes
         THEN ttratio := 0.5;
         IF weights
         THEN printweights;
      END; (* getoptions *)

      PROCEDURE getbasefreqs;
         (* compute or read in base frequencies *)
         VAR aa, bb : REAL;
      BEGIN (* getbasefreqs *)
         WRITELN(OUTPUT);
         IF freqsfrom
         THEN WRITE(OUTPUT, 'Empirical');
         WRITELN(OUTPUT, ' Base Frequencies:');
         WRITELN(OUTPUT);
         IF NOT freqsfrom
         THEN BEGIN
            IF kimura OR jukes
            THEN BEGIN
               freqa := 0.25;
               freqc := 0.25;
               freqg := 0.25;
               freqt := 0.25;
               END
            ELSE READLN(infile, freqa, freqc, freqg, freqt);
            END;
         WRITELN(OUTPUT, '    A    ', freqa:10:5);
         WRITELN(OUTPUT, '    C    ', freqc:10:5);
         WRITELN(OUTPUT, '    G    ', freqg:10:5);
         WRITELN(OUTPUT, '   T(U)  ', freqt:10:5);
         freqr := freqa + freqg;
         freqy := freqc + freqt;
         freqar := freqa/freqr;
         freqcy := freqc/freqy;
         freqgr := freqg/freqr;
         freqty := freqt/freqy;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'Transition/transversion ratio = ', ttratio:10:6);
         WRITELN(OUTPUT);
         aa := ttratio*freqr*freqy - (freqa*freqg + freqc*freqt);
         bb := freqa*freqgr + freqc*freqty;
         xi := aa/(aa+bb);
         xv := 1.0 - xi;
         ttratio := xi / xv;
         IF (xi <= 0.0) AND (xi >= -epsilon)
         THEN xi := 0.0;
         IF xi < 0.0
         THEN BEGIN
            WRITELN(OUTPUT, 'WARNING: This transition/transversion ratio');
            WRITELN(OUTPUT, 'is impossible with these base frequencies!');
            xi := 3/5;
            xv := 2/5;
            WRITELN(OUTPUT, ' Transition/transversion parameter reset');
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT, '(Transition/transversion parameter = ',
            xi/xv:10:6, ')');
         WRITELN(OUTPUT);
         fracchange := xi*(2*freqa*freqgr + 2*freqc*freqty)
                       + xv*(1.0 - freqa*freqa - freqc*freqc
                                  - freqg*freqg - freqt*freqt);
      END; (* getbasefreqs *)

     PROCEDURE setuptree;
         VAR i : INTEGER;
      BEGIN (* setuptree *)
         FOR i := 1 TO numsp DO
            NEW(nodep[i]);
      END; (* setuptree *)

      PROCEDURE inputdata;
         (* Input the names and sequences for each species *)
        VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread : BOOLEAN;
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         WRITE(OUTPUT, 'Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Sequences');
         WRITE(OUTPUT, '----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '---------');
         WRITELN(OUTPUT);
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            FOR i := 1 TO numsp DO
               IF NOT anerror
               THEN BEGIN
                  IF basesread = 0
                  THEN FOR j := 1 TO nmlngth DO
                        IF NOT anerror
                        THEN BEGIN
                           READ(infile, nodep[i]^.nayme[j]);
                           anerror := anerror OR EOF(infile) OR EOLN(infile);
                           IF anerror
                           THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                               'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                           END;
                  j := basesread;
                  WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                     THEN BEGIN
                        WRITELN(OUTPUT, 'ERROR: BAD BASE:', charstate,
                           ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                        anerror := TRUE;
                        END;
                     j := j + 1;
                     IF charstate = '.' THEN charstate := y[1][j];
                     y[i][j] := charstate;
                     END;
                  END;                           
               IF i = 1 THEN basesnew := j;
               READLN(infile);
               anerror := anerror OR (j <> basesnew);
               IF anerror
               THEN WRITELN(OUTPUT, 'ERROR: SEQUENCES OUT OF ALIGNMENT');
               END;
            basesread := basesnew;
            allread := basesread = sites;
            END;
         IF NOT anerror
         THEN BEGIN
            FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO numsp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(OUTPUT, nodep[j]^.nayme[k]);
                  WRITE(OUTPUT, '   ');
                  l := 60*i;
                  IF l > sites THEN l := sites;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j][k] = y[1][k])
                     THEN charstate := '.'
                     ELSE charstate := y[j][k];
                     WRITE(OUTPUT, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort of sites lexicographically *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := sites DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO sites-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     tied := (oldweight[jj] = oldweight[jg]);
                     flip := (oldweight[jj] < oldweight[jg])
                             OR (tied AND (category[jj] > category[jg]));
                     tied := tied AND (category[jj] = category[jg]);
                     k := 1;
                     WHILE (k <= numsp) AND tied DO BEGIN
                        flip := y[k][jj] > y[k][jg];
                        tied := tied AND (y[k][jj] = y[k][jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < sites DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= sites) AND tied DO BEGIN
                  tied := (oldweight[alias[i]] = oldweight[alias[j]])
                     AND (category[alias[i]] = category[alias[j]]);
                  k := 1;
                  WHILE (k <= numsp) AND tied DO BEGIN
                     tied := tied AND (y[k][alias[i]] = y[k][alias[j]]);
                     k := k + 1;
                     END;
                  IF tied
                  THEN BEGIN
                     ally[alias[j]] := alias[i];
                     j := j + 1;
                     END;
                  END;
               i := j;
               END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* move so one representative of each pattern of
               sites comes first *)
            VAR i, j, itemp : INTEGER;
               done, found, completed : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               IF NOT (ally[alias[i]] = alias[i])
               THEN BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= sites
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := (ally[alias[j]] = alias[j]);
                        j := j + 1;
                        completed := j > sites;
                        IF j <= sites
                        THEN completed := oldweight[alias[j]] = 0;
                     UNTIL found OR completed;
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END
                  ELSE done := TRUE;
                  END;
               i := i + 1;
               done := done OR (i >= sites);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO sites DO BEGIN
            alias[i] := i;
            ally[i] := i;
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         endsite := 0;
         FOR i := 1 TO sites DO
            IF (ally[i] = i) AND (oldweight[i] > 0)
            THEN endsite := endsite + 1;
         FOR i := 1 TO endsite DO location[alias[i]] := i;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN(OUTPUT, 'TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
         sumrates := 0.0;
         FOR i := 1 TO sites DO
           sumrates := sumrates + oldweight[i]*rate[category[i]];
         FOR i := 1 TO categs DO rate[i] := rate[i] * (weightsum/sumrates);
         sumrates := weightsum;
      END; (* makeweights *)

      PROCEDURE makevalues;
         (* set up fractional likelihoods at tips *)
         VAR i, j, k : INTEGER;
            b : base;
      BEGIN (* makevalues *)
         FOR k := 1 TO endsite DO BEGIN
            j := alias[k];
            FOR i := 1 TO numsp DO BEGIN
               FOR b := A TO T DO
                  nodep[i]^.x[k][b] := 0.0;
               CASE y[i][j] OF
                  'A' : nodep[i]^.x[k][A] := 1.0;
                  'C' : nodep[i]^.x[k][C] := 1.0;
                  'G' : nodep[i]^.x[k][G] := 1.0;
                  'T' : nodep[i]^.x[k][T] := 1.0;
                  'U' : nodep[i]^.x[k][T] := 1.0;
                  'M' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][C] := 1.0;
                        END;
                  'R' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        END;
                  'W' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'S' : BEGIN
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        END;
                  'Y' : BEGIN
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'K' : BEGIN
                        nodep[i]^.x[k][G] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'B' : BEGIN
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'D' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'H' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][T] := 1.0;
                        END;
                  'V' : BEGIN
                        nodep[i]^.x[k][A] := 1.0;
                        nodep[i]^.x[k][C] := 1.0;
                        nodep[i]^.x[k][G] := 1.0;
                        END;
                  'N' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  'X' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  '?' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  'O' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0;
                  '-' : FOR b := A TO T DO
                           nodep[i]^.x[k][b] := 1.0
                  END;
               END;
            END;
      END; (* makevalues *)

      PROCEDURE empiricalfreqs;
         (* Get empirical base frequencies from the data *)
         VAR i, j, k, l : INTEGER;
            sum, suma, sumc, sumg, sumt, w : REAL;
      BEGIN (* empiricalfreqs *)
         freqa := 0.25;
         freqc := 0.25;
         freqg := 0.25;
         freqt := 0.25;
         FOR k := 1 TO 8 DO BEGIN
            suma := 0.0;
            sumc := 0.0;
            sumg := 0.0;
            sumt := 0.0;
            FOR i := 1 TO numsp DO
               FOR j := 1 TO sites DO BEGIN
                  l := location[ally[j]];
                  w := oldweight[l];
                  sum := freqa*nodep[i]^.x[l][A];
                  sum := sum + freqc*nodep[i]^.x[l][C];
                  sum := sum + freqg*nodep[i]^.x[l][G];
                  sum := sum + freqt*nodep[i]^.x[l][T];
                  suma := suma + w*freqa*nodep[i]^.x[l][A] / sum;
                  sumc := sumc + w*freqc*nodep[i]^.x[l][C] / sum;
                  sumg := sumg + w*freqg*nodep[i]^.x[l][G] / sum;
                  sumt := sumt + w*freqt*nodep[i]^.x[l][T] / sum;
                  END;
            sum := suma + sumc + sumg + sumt;
            freqa := suma / sum;
            freqc := sumc / sum;
            freqg := sumg / sum;
            freqt := sumt / sum;
            END;
      END; (* empiricalfreqs *)

   BEGIN (* getinput *)
      getnums;
      IF NOT anerror THEN getoptions;
      IF (NOT freqsfrom) AND (NOT anerror)
      THEN getbasefreqs;
      IF NOT anerror THEN setuptree;
      IF NOT anerror THEN inputdata;
      IF NOT anerror THEN makeweights;
      IF NOT anerror
      THEN BEGIN
         makevalues;
         IF freqsfrom
         THEN BEGIN
            empiricalfreqs;
            getbasefreqs
            END;
         END;
   END; (* getinput *)

   PROCEDURE bootweights;
      (* sets up weights by resampling data *)
      VAR i, j : INTEGER;
         p, q, r : REAL;
   BEGIN (* bootweights *)
      FOR i := 1 TO endsite DO weight[i] := 0;
      IF halfjack
      THEN BEGIN
         p := 0.5;
         q := weightsum/2;
         r := weightsum;
         FOR i := 1 TO weightsum DO BEGIN
            IF randum(seed) < p
            THEN BEGIN
               weight[location[ally[alias[i]]]]
                  := weight[location[ally[alias[i]]]] + 1;
               q := q - 1;
               END;
            r := r - 1;
            IF i < weightsum THEN p := q / r;
            END;
         END
      ELSE IF boot
         THEN BEGIN
            FOR i := 1 TO weightsum DO BEGIN
               j := TRUNC(weightsum*randum(seed)) + 1;
               weight[location[ally[alias[j]]]]
                  := weight[location[ally[alias[j]]]] + 1;
               END;
            END
         ELSE FOR i := 1 TO sites DO
               weight[location[ally[i]]] := weight[location[ally[i]]]
                                             + oldweight[i];
   END; (* bootweights *)

   PROCEDURE makedists;
      (* compute distance matrix *)
      VAR i, j : INTEGER;
         v : REAL;

      PROCEDURE makev (m, n : INTEGER; VAR v : REAL);
         (* compute one distance *)
         VAR i, j, it, numerator, denominator, num1, num2 : INTEGER;
            sum, denom, sum1, sum2, sumyr, rat, y, z, lz,
               yy, zz, y1, z1, yold, aa, bb, cc, vv,
               tt, p1, p2, p3, q1, q2, r1, r2, r3 : REAL;
            prod, prod2, prod3 : ARRAY[1..maxsites] OF REAL;
            done, quick, jukesquick, kimquick : BOOLEAN;
            b : base;
            p, q : nodeptr;
            xx1, xx2 : sitelike;
      BEGIN (* makev *)
         p := nodep[m];
         q := nodep[n];
         quick := TRUE;
         IF jukes OR kimura
         THEN BEGIN
            numerator := 0;
            denominator := 0;
            FOR i := 1 TO endsite DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               sum := 0.0;
               sum1 := 0.0;
               sum2 := 0.0;
               FOR b := A TO T DO BEGIN
                  sum1 := sum1 + xx1[b];
                  sum2 := sum2 + xx2[b];
                  sum := sum + xx1[b]*xx2[b];
                  END;
               quick := quick AND
                  ((sum1 = 1.0) OR (sum1 = 4.0)) AND
                  ((sum2 = 1.0) OR (sum2 = 4.0));
               IF (sum1 = 1.0) AND (sum2 = 1.0)
               THEN BEGIN
                  numerator := numerator + TRUNC(weight[i]*sum);
                  denominator := denominator + weight[i];
                  END;
               END;
            END;
         jukesquick := jukes AND quick;
         kimquick := kimura AND quick;
         IF jukesquick AND (4*numerator <= denominator)
         THEN BEGIN
            WRITE(OUTPUT, ' WARNING: INFINITE DISTANCE BETWEEN ');
            WRITELN(OUTPUT, ' SPECIES ', m:3, ' AND ', n:3);
            anerror := TRUE;
            END;
         IF jukesquick AND (NOT anerror)
         THEN vv := - 0.75 * LN((4.0*(numerator/denominator)-1.0)/3.0);
         IF kimquick
         THEN BEGIN
            num1 := 0;
            num2 := 0;
            denominator := 0;
            FOR i := 1 TO endsite DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               sum := 0.0;
               sum1 := 0.0;
               sum2 := 0.0;
               FOR b := A TO T DO BEGIN
                  sum1 := sum1 + xx1[b];
                  sum2 := sum2 + xx2[b];
                  sum := sum + xx1[b]*xx2[b];
                  END;
               sumyr := (xx1[A] + xx1[G])*(xx2[A] + xx2[G])
                        + (xx1[C] + xx1[T])*(xx2[C] + xx2[T]);
               IF (sum1 = 1.0) AND (sum2 = 1.0)
               THEN BEGIN
                  num1 := num1 + TRUNC(weight[i]*sum);
                  num2 := num2 + TRUNC(weight[i]*(sumyr-sum));
                  denominator := denominator + weight[i];
                  END;
               END;
            tt := 1.0 - num1/denominator;
            IF tt > 0.0
            THEN FOR i := 1 TO iterations DO BEGIN
               p1 := EXP(-tt);
               p2 := EXP(-xv*tt) - EXP(-tt);
               p3 := 1.0 - EXP(-xv*tt);
               q1 := p1 + p2/2.0 + p3/4.0;
               q2 := p2/2.0 + p3/4.0;
               r1 := num1*p1/q1;
               r2 := num1*p2/(2.0*q1) + num2*p2/(2.0*q2);
               r3 := denominator - r1 - r2;
               FOR j := 1 TO cycles DO
                  tt := (r2*(xi*tt/(1.0-EXP(-xi*tt)))
                         + r3*(xv*tt/(1.0-EXP(-xv*tt))+xi*tt))/denominator;
               END;
            vv := fracchange * tt;
            END;
         IF NOT (jukesquick OR kimquick)
         THEN BEGIN
            y := 0.1;
            FOR i := 1 TO endsite DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               sum1 := freqa*xx1[A] + freqc*xx1[C]
                         + freqg*xx1[G] + freqt*xx1[T];
               sum2 := freqa*xx2[A] + freqc*xx2[C]
                         + freqg*xx2[G] + freqt*xx2[T];
               prod[i] := sum1*sum2;
               prod2[i] := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar
                            +xx2[G]*freqgr) + (xx1[C]*freqc+xx1[T]
                            *freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
               prod3[i] := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                            + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
               END;
            IF y < epsilon
            THEN y := 10.0*epsilon;
            IF y = 1.0
            THEN y := 1.0 - 10.0*epsilon;
            done := FALSE;
            it := 1;
            WHILE (it < iterations) AND (NOT done) DO BEGIN
               sum := 0.0;
               denom := 0.0;
               z := 1.0 - y;
               IF z > 0.0
               THEN lz := LN(z);
               FOR i := 1 TO endsite DO BEGIN
                  rat := rate[category[alias[i]]];
                  IF z > 0.0
                  THEN BEGIN
                     zz := EXP(rat*xi*lz);
                     z1 := EXP(rat*xv*lz);
                     END
                  ELSE BEGIN
                     zz := 0.0;
                     z1 := 0.0;
                     END;
                  yy := 1.0 - zz;
                  y1 := 1.0 - z1;
                  cc := prod[i];
                  bb := prod2[i];
                  aa := prod3[i];
                  sum := sum + weight[i]*rat*(xi*(bb*z1+cc*y1)+xv*cc)
                                    /(aa*z1*zz + bb*z1*yy + cc*y1);
                  denom := denom + weight[i]*rat;
                  END;
               yold := y;
               y := 1.0 - EXP(lz*sum/denom);
               done := ABS(yold-y) < epsilon;
               it := it + 1;
               END;
            vv := -LN(1.0-y)*fracchange;
            END;
         v := vv;
      END; (* makev *)

   BEGIN (* makedists *)
      WRITELN(OUTPUT, 'Distances calculated for species');
      FOR i := 1 TO numsp DO d[i][i] := 0.0;
      FOR i := 1 TO numsp-1 DO BEGIN
         WRITE(OUTPUT, '    ');
         FOR j := 1 TO nmlngth DO
            WRITE(OUTPUT, nodep[i]^.nayme[j]);
         WRITE(OUTPUT, '   ');
         FOR j := i+1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               makev (i, j, v);
               d[i][j] := v;
               d[j][i] := v;
               WRITE(OUTPUT, '.');
               END;
         WRITELN(OUTPUT);
         END;
      WRITE(OUTPUT, '    ');
      FOR j := 1 TO nmlngth DO
         WRITE(OUTPUT, nodep[numsp]^.nayme[j]); 
      WRITELN(OUTPUT);
   END; (* makedists *)

   PROCEDURE writedists;
      (* write out distances *)
      VAR i, j, k : INTEGER;
   BEGIN (* writedists *)
     WRITE(outfile, numsp:5);
     IF lower
     THEN WRITE(outfile, '  L');
     WRITELN(outfile);
     FOR i := 1 TO numsp DO BEGIN
        FOR j := 1 TO nmlngth DO
           WRITE(outfile, nodep[i]^.nayme[j]);
        IF lower
        THEN k := i-1
        ELSE k := numsp;
        FOR j := 1 TO k DO BEGIN
           WRITE(outfile, d[i][j]:8:4);
           IF (((j+1) MOD 9) = 0) AND (j < k)
           THEN WRITELN(outfile);
           END;
        WRITELN(outfile);
        END;
      WRITELN(OUTPUT);
      IF rr = repsno
      THEN BEGIN
         WRITELN(OUTPUT, 'Distances written to file');
         WRITELN(OUTPUT);
         END;
   END; (* writedists *)

BEGIN (* DNA Distances by Maximum Likelihood *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   (* ASSIGN(outfile, 'OUTFILE');   Turbo Pascal only *)
   REWRITE(outfile);
   getinput;
   IF boot OR halfjack
   THEN WRITE(outfile, repsno:5);
   IF boot
   THEN WRITELN(outfile, ' BY');
   IF halfjack
   THEN WRITELN(outfile, ' HY');
   FOR rr := 1 TO repsno DO BEGIN
      IF repsno > 1
      THEN BEGIN
         WRITELN(OUTPUT, 'Replicate number', rr:5);
         WRITELN(OUTPUT);
         END;
      IF NOT anerror
      THEN bootweights;
      IF NOT anerror
      THEN makedists;
      IF NOT anerror
      THEN writedists;
      END;
(* CLOSE(infile);  -- may need this here *)
   (* CLOSE(outfile);   Turbo Pascal and maybe others too *)
END. (* DNA Distances by Maximum Likelihood *)

