PROGRAM DNAInvar (infile, OUTPUT);
  (* version 3.31. (c) Copyright 1990 by Joseph Felsenstein.  Written by
     Joseph Felsenstein.  Permission is granted to copy and use this program
     provided no fee is charged for it and provided that this copyright notice
     is not removed. *)

   CONST maxsp = 4;       (* maximum number of species -- must be 4 *)
      maxsites = 5000;    (* maximum number of sites *)
      maxpatterns = 1000; (* max. number of different site patterns *)
      nmlngth = 10;       (* max. number characters in species name *)

   TYPE symbol = (xx, yy, zz, ww);
      sequence = ARRAY[1..maxsp] OF ARRAY [1..maxsites] OF CHAR;
      naym = ARRAY[1..nmlngth] OF CHAR;

   VAR infile : TEXT;
      numsp, sites, endsite : INTEGER;
      weights, anerror : BOOLEAN;
      nayme : ARRAY[1..maxsp] OF naym;
      y : sequence;
      weight, alias, aliasweight : ARRAY[1..maxsites] OF INTEGER;

   PROCEDURE getinput;
      (* read input file *)

      PROCEDURE getnums;
         (* input number of species, number of sites *)
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,'Nucleic acid sequence Invariants ');
         WRITELN(OUTPUT, 'method, version 3.3');
         WRITELN(OUTPUT);
         READ(infile, numsp, sites);
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: only 4 allowed');
         IF sites > maxsites
         THEN WRITELN(OUTPUT, 'TOO MANY SITES: adjust CONSTants');
         WRITELN(OUTPUT, numsp:4, ' Species, ', sites:4, ' Sites');
         anerror := (numsp > maxsp) OR (sites > maxsites);
      END; (* getnums *)

      FUNCTION letter (ch : CHAR) : BOOLEAN;
         (* tests whether ch is a letter
            -- works for both ASCII and EBCDIC codes *)
      BEGIN (* letter *)
         letter := ((ch >= 'A') AND (ch <= 'I'))
              OR ((ch >= 'J') AND (ch <= 'R'))
              OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
              OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
      END; (* letter *)

      PROCEDURE uppercase (VAR ch : CHAR);
         (* convert ch to upper case -- either ASCII or EBCDIC *)
      BEGIN (* uppercase *)
         IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                         OR  ('a' < 'A') AND (ch < 'A'))
         THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
      END; (* uppercase *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputweights;
            (* input the character weights, which must be 0 or 1 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO sites DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF (ch = '0')
               THEN weight[i] := 0
               ELSE IF (ch <> '1')
                  THEN BEGIN
                     WRITELN(OUTPUT, 'BAD WEIGHT CHARACTER: ',ch);
                     anerror := TRUE;
                     END;
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'   Sites are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (sites DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=sites)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'W'
            THEN extranum := extranum + 1
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
         READLN(infile);
         FOR i := 1 TO sites DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            IF NOT anerror
            THEN BEGIN
               READ(infile, ch);
               uppercase (ch);
               IF ch = 'W'
               THEN inputweights;
               anerror := NOT (ch = 'W');
               IF anerror
               THEN WRITELN(OUTPUT, 'ERROR: INCORRECT AUXILIARY',
                     ' OPTIONS LINE WHICH STARTS WITH ', ch);
               END;
            END;
         IF weights
         THEN printweights;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* Input the names and sequences for each species *)
         VAR i, j, k, l, basesread, basesnew : INTEGER;
            charstate : CHAR;
            allread : BOOLEAN;
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 37 THEN j := 37;
         WRITE(OUTPUT, 'Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Sequences');
         WRITE(OUTPUT, '----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '---------');
         WRITELN(OUTPUT);
         basesread := 0;
         allread := FALSE;
         WHILE NOT (allread OR anerror) DO BEGIN
            allread := TRUE;
            IF EOLN(infile) THEN READLN(infile);
            FOR i := 1 TO numsp DO
               IF NOT anerror
               THEN BEGIN
                  IF basesread = 0
                  THEN FOR j := 1 TO nmlngth DO
                        IF NOT anerror
                        THEN BEGIN
                           READ(infile, nayme[i,j]);
                           anerror := anerror OR EOF(infile) OR EOLN(infile);
                           IF anerror
                           THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                               'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                           END;
                  j := basesread;
                  WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
                     AND (NOT anerror) DO BEGIN
                     READ(infile, charstate);
                     IF (charstate <> ' ') AND
                        ((charstate < '0') OR (charstate > '9'))
                     THEN BEGIN
                        uppercase (charstate);
                        IF NOT ((charstate = 'A') OR (charstate = 'B')
                           OR (charstate = 'C') OR (charstate = 'D')
                           OR (charstate = 'G') OR (charstate = 'H')
                           OR (charstate = 'K') OR (charstate = 'M')
                           OR (charstate = 'N') OR (charstate = 'R')
                           OR (charstate = 'S') OR (charstate = 'T')
                           OR (charstate = 'U') OR (charstate = 'V')
                           OR (charstate = 'W') OR (charstate = 'X')
                           OR (charstate = 'Y') OR (charstate = '?')
                           OR (charstate = 'O') OR (charstate = '-')
                           OR (charstate = '.'))
                        THEN BEGIN
                           WRITELN(OUTPUT, 'ERROR: BAD BASE:', charstate,
                              ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                           anerror := TRUE;
                           END;
                        j := j + 1;
                        IF charstate = '.' THEN charstate := y[1][j];
                        y[i][j] := charstate;
                        END;
                     END;
                  IF i = 1 THEN basesnew := j;
                  READLN(infile);
                  anerror := anerror OR (j <> basesnew);
                  IF anerror
                  THEN WRITELN(OUTPUT, 'ERROR: SEQUENCES OUT OF ALIGNMENT');
                  END;
            basesread := basesnew;
            allread := basesread = sites;
            END;
         IF NOT anerror
         THEN BEGIN
            FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
               FOR j := 1 TO numsp DO BEGIN
                  FOR k := 1 TO nmlngth DO WRITE(OUTPUT, nayme[j,k]);
                  WRITE(OUTPUT, '   ');
                  l := 60*i;
                  IF l > sites THEN l := sites;
                  FOR k := 60*(i-1)+1 TO l DO
                  IF NOT anerror
                  THEN BEGIN
                     IF (j > 1) AND (y[j][k] = y[1][k])
                     THEN charstate := '.'
                     ELSE charstate := y[j][k];
                     WRITE(OUTPUT, charstate);
                     IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
      END; (* inputdata *)

      PROCEDURE makeweights;
         (* make up weights vector to avoid duplicate computations *)
         VAR i : INTEGER;

         PROCEDURE sitesort;
            (* Shell sort keeping sites, weights in same order *)
            VAR gap, i, j, jj, jg, k, itemp : INTEGER;
               flip, tied : BOOLEAN;
         BEGIN (* sitesort *)
            gap := sites DIV 2;
            WHILE (gap > 0) DO BEGIN
               FOR i := gap TO sites-1 DO BEGIN
                  j := i - gap + 1;
                  flip := TRUE;
                  WHILE ((j>0) AND flip) DO BEGIN
                     jj := alias[j];
                     jg := alias[j+gap];
                     flip := FALSE;
                     k := 1;
                     tied := TRUE;
                     WHILE (k <= numsp) AND tied DO BEGIN
                        flip := y[k][jj] > y[k][jg];
                        tied := tied AND (y[k][jj] = y[k][jg]);
                        k := k + 1;
                        END;
                     IF flip
                     THEN BEGIN
                        itemp := alias[j];
                        alias[j] := alias[j+gap];
                        alias[j+gap] := itemp;
                        itemp := aliasweight[j];
                        aliasweight[j] := aliasweight[j+gap];
                        aliasweight[j+gap] := itemp;
                        j := j - gap;
                        END;
                     END;
                  END;
               gap := gap DIV 2;
               END;
         END; (* sitesort *)

         PROCEDURE sitecombine;
            (* combine sites that have identical patterns *)
            VAR i, j, k : INTEGER;
               tied : BOOLEAN;
         BEGIN (* sitecombine *)
            i := 1;
            WHILE i < sites DO BEGIN
               j := i+1;
               tied := TRUE;
               WHILE (j <= sites) AND tied DO BEGIN
                  k := 1;
                  WHILE (k <= numsp) AND tied DO BEGIN
                     tied := tied AND (y[k][alias[i]] = y[k][alias[j]]);
                     k := k + 1;
                     END;
                  IF tied AND (aliasweight[j] > 0)
                  THEN BEGIN
                     aliasweight[i] := aliasweight[i] + aliasweight[j];
                     aliasweight[j] := 0;
                     END;
                  j := j + 1;
                  END;
              i := j - 1;
              END;
         END; (* sitecombine *)

         PROCEDURE sitescrunch;
            (* Bubble sort so positively weighted sites come first *)
            VAR i, j, itemp : INTEGER;
               done, found : BOOLEAN;
         BEGIN (* sitescrunch *)
            done := FALSE;
            i := 1;
            j := 2;
            WHILE NOT done DO BEGIN
               found := FALSE;
               IF aliasweight[i] > 0
               THEN i := i + 1
               ELSE BEGIN
                  IF j <= i
                  THEN j := i + 1;
                  IF j <= sites
                  THEN BEGIN
                     found := FALSE;
                     REPEAT
                        found := aliasweight[j] > 0;
                        j := j + 1;
                     UNTIL found OR (j > sites);
                     IF found
                     THEN BEGIN
                        j := j - 1;
                        itemp := alias[i];
                        alias[i] := alias[j];
                        alias[j] := itemp;
                        itemp := aliasweight[i];
                        aliasweight[i] := aliasweight[j];
                        aliasweight[j] := itemp;
                        END
                     ELSE done := TRUE;
                     END;
                  END;
               done := done OR (i >= sites);
               END;
         END; (* sitescrunch *)

      BEGIN (* makeweights *)
         FOR i := 1 TO sites DO BEGIN
            alias[i] := i;
            aliasweight[i] := weight[i];
            END;
         sitesort;
         sitecombine;
         sitescrunch;
         FOR i := 1 TO sites DO BEGIN
            weight[i] := aliasweight[i];
            IF weight[i] > 0
            THEN endsite := i;
            END;
         IF endsite > maxpatterns
         THEN BEGIN
            WRITELN(OUTPUT, 'TOO MANY PATTERNS: increase',
                 ' CONSTant maxpatterns to at least ', endsite:5);
            anerror := TRUE;
            END;
      END; (* makeweights *)

   BEGIN (* getinput *)
      getnums;
      IF NOT anerror THEN inputoptions;
      IF NOT anerror THEN inputdata;
      IF NOT anerror THEN makeweights;
   END; (* getinput *)

   PROCEDURE makeinv;
      (* print out patterns and compute invariants *)
      VAR f : ARRAY[xx..ww] OF ARRAY[xx..ww] OF ARRAY[xx..ww] OF INTEGER;

      PROCEDURE printpatterns;
         (* print out patterns *)
         VAR i, j : INTEGER;
      BEGIN (* printpatterns *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '   Pattern   Number of times');
         WRITELN(OUTPUT);
         FOR i := 1 TO endsite DO BEGIN
            WRITE(OUTPUT, '     ');
            FOR j := 1 TO numsp DO
               WRITE(OUTPUT, y[j][alias[i]]);
            WRITELN(OUTPUT, '  ', weight[i]:8);
            END;
         WRITELN(OUTPUT);
      END; (* printpatterns *)

      PROCEDURE makesymmetries;
         (* get frequencies of symmetrized patterns *)
         VAR i, j : INTEGER;
            drop, usedz : BOOLEAN;
            ch, ch1, zchar : CHAR;
            s1, s2, s3 : symbol;
            t : ARRAY[2..maxsp] OF symbol;
      BEGIN (* makesymmetries *)
         FOR s1 := xx TO ww DO
            FOR s2 := xx TO ww DO
               FOR s3 := xx TO ww DO
                  f[s1][s2][s3] := 0;
         FOR i := 1 TO endsite DO BEGIN
            drop := FALSE;
            FOR j := 1 TO numsp DO BEGIN
               ch := y[j][alias[i]];
               drop := drop OR NOT ((ch = 'A') OR (ch = 'C')
                            OR (ch = 'G') OR (ch = 'T') OR (ch = 'U'));
               END;
            ch1 := y[1][alias[i]];
            IF NOT drop
            THEN BEGIN
               usedz := FALSE;
               zchar := ' ';
               FOR j := 2 TO numsp DO BEGIN
                  ch := y[j][alias[i]];
                  IF ch = ch1
                  THEN t[j] := xx
                  ELSE IF ((ch1 = 'A') AND (ch = 'G'))
                          OR ((ch1 = 'G') AND (ch = 'A'))
                          OR ((ch1 = 'C') AND ((ch = 'T') OR (ch = 'U')))
                          OR (((ch1 = 'T') OR (ch1 = 'U')) AND (ch = 'C'))
                     THEN t[j] := yy
                     ELSE IF NOT usedz
                        THEN BEGIN
                           t[j] := zz;
                           usedz := TRUE;
                           zchar := ch;
                           END
                        ELSE IF usedz AND (ch = zchar)
                           THEN t[j] := zz
                           ELSE IF usedz AND (ch <> zchar)
                           THEN t[j] := ww;
                  END;
               f[t[2]][t[3]][t[4]] := f[t[2]][t[3]][t[4]] + weight[i];
               END;
            END;
      END; (* makesymmetries *)

      PROCEDURE printsymmetries;
         (* print out symmetrized pattern numbers *)
         VAR s1, s2, s3 : symbol;

         PROCEDURE printsymbol(s : symbol);
            (* print 1, 2, 3, 4 as appropriate *)
         BEGIN (* printsymbol *)
            CASE s OF
               xx : WRITE(OUTPUT, '1');
               yy : WRITE(OUTPUT, '2');
               zz : WRITE(OUTPUT, '3');
               ww : WRITE(OUTPUT, '4')
               END;
         END; (* printsymbol *)

      BEGIN (* printsymmetries *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'Symmetrized patterns (1, 2 = the two purines  ');
         WRITELN(OUTPUT, 'and  3, 4 = the two pyrimidines');
         WRITE(OUTPUT, '                  or  1, 2 = the two pyrimidines  ');
         WRITELN(OUTPUT, 'and  3, 4 = the two purines)');
         WRITELN(OUTPUT);
         FOR s1 := xx TO ww DO
            FOR s2 := xx TO ww DO
               FOR s3 := xx TO ww DO
                  IF f[s1][s2][s3] > 0
                  THEN BEGIN
                     WRITE(OUTPUT, '     1');
                     printsymbol(s1);
                     printsymbol(s2);
                     printsymbol(s3);
                     WRITELN('   ', f[s1][s2][s3]:7);
                     END;
      END; (* printsymmetries *)

      PROCEDURE invariants;
         (* compute invariants *)
         VAR m, n, p, q : INTEGER;
            L1, L2, L3 : REAL;

         PROCEDURE tabulate (mm, nn, pp, qq : INTEGER);
            (* make quadratic invariant, table, chi-square *)
            VAR total : INTEGER;
               mr, nr, pr, qr, k : REAL;
         BEGIN (* tabulate *)
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'   Contingency Table');
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, mm:7, nn:6);
            WRITELN(OUTPUT, pp:7, qq:6);
            WRITELN(OUTPUT);
            mr := mm;
            nr := nn;
            pr := pp;
            qr := qq;
            total := mm + nn + pp + qq;
            WRITELN(OUTPUT, '   Quadratic invariant = ', (nr*pr-mr*qr):15:1);
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   Chi-square = ');
            k := total * SQR(mr*qr - nr*pr)/((mr+nr)*(mr+pr)*(nr+qr)*(pr+qr));
            WRITE(OUTPUT, k:10:5);
            IF (mr*qr > nr*pr) AND (k > 2.71)
            THEN WRITELN(OUTPUT, ' (P < 0.05)')
            ELSE WRITELN(OUTPUT, ' (not significant)');
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* tabulate *)

         PROCEDURE writetree (i, j, k, l : INTEGER);
            (* write out tree topology ((i,j),(k,l)) using names *)

            PROCEDURE writename (m : INTEGER);
               (* write out a species name *)
               VAR i, n : INTEGER;
            BEGIN (* writename *)
               n := nmlngth;
               WHILE nayme[m][n] = ' ' DO n := n - 1;
               IF n = 0 THEN n := 1;
               FOR i := 1 TO n DO
                  WRITE(OUTPUT, nayme[m][i]);
            END; (* writename *)

         BEGIN(* writetree *)
            WRITE(OUTPUT, '((');
            writename(i);
            WRITE(OUTPUT, ',');
            writename(j);
            WRITE(OUTPUT, '),(');
            writename(k);
            WRITE(OUTPUT, ',');
            writename(l);
            WRITELN(OUTPUT, '))');
         END; (* writetree *)

         PROCEDURE exacttest (m, n : INTEGER);
            (* exact binomial test that m <= n *)
            VAR i : INTEGER;
               p, sum : REAL;
         BEGIN (* exacttest *)
            p := 1.0;
            FOR i := 1 TO m+n DO p := p/2.0;
            sum := p;
            FOR i := 1 TO n DO BEGIN
               p := p*(m+n-i+1)/i;
               sum := sum + p;
               END;
            WRITE(OUTPUT, '      ', sum:7:4);
            IF sum <= 0.05
            THEN WRITELN(OUTPUT, '              yes')
            ELSE WRITELN(OUTPUT, '               no');
         END; (* exacttest *)

      BEGIN (* invariants *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'Tree topologies (unrooted): ');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '    I:  ');
         writetree(1, 2, 3, 4);
         WRITE(OUTPUT, '   II:  ');
         writetree(1, 3, 2, 4);
         WRITE(OUTPUT, '  III:  ');
         writetree(1, 4, 2, 3);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, 'Lake''s linear invariants');
         WRITELN(OUTPUT, ' (these are expected to be zero for the',
                           ' two incorrect tree topologies.');
         WRITELN(OUTPUT, '  This is tested by testing the equality',
                           ' of the two parts');
         WRITELN(OUTPUT, '  of each expression using',
                           ' a one-sided exact binomial test.');
         WRITELN(OUTPUT, '  The null hypothesis is that the first',
                           ' part is no larger than the second.)');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Tree                           ');
         WRITELN(OUTPUT, '  Exact test P value    Significant?');
         WRITELN(OUTPUT);
         m := f[yy][zz][ww]+f[xx][zz][zz];
         n := f[yy][zz][zz]+f[xx][zz][ww];
         WRITE(OUTPUT, '   I  ', m:5, '    - ',n:5, '   = ', m-n:5);
         exacttest(m, n);
         m := f[zz][yy][ww]+f[zz][xx][zz];
         n := f[zz][yy][zz]+f[zz][xx][ww];
         WRITE(OUTPUT, '   II ', m:5, '    - ',n:5, '   = ', m-n:5);
         exacttest(m, n);
         m := f[zz][ww][yy]+f[zz][zz][xx];
         n := f[zz][zz][yy]+f[zz][ww][xx];
         WRITE(OUTPUT, '   III', m:5, '    - ',n:5, '   = ', m-n:5);
         exacttest(m, n);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'Cavender''s quadratic invariants (type L)');
         WRITELN(OUTPUT, ' using purines vs. pyrimidines');
         WRITELN(OUTPUT, ' (these are expected to be zero, and thus have a',
                         ' nonsignificant');
         WRITELN(OUTPUT, '  chi-square, for the correct tree topology)');
         WRITELN(OUTPUT, 'They will be misled if there are substantially');
         WRITELN(OUTPUT, 'different evolutionary rate between sites, or');
         WRITELN(OUTPUT, 'different purine:pyrimidine ratios from 1:1.');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '  Tree I:');
         m := f[xx][xx][xx] + f[xx][yy][yy] + f[xx][zz][zz];
         n := f[xx][xx][yy] + f[xx][xx][zz] + f[xx][yy][xx]
              + f[xx][yy][zz] + f[xx][zz][xx] + f[xx][zz][yy]
              + f[xx][zz][ww];
         p := f[yy][xx][xx] + f[yy][yy][yy] + f[yy][zz][zz]
              + f[zz][xx][xx] + f[zz][yy][yy] + f[zz][zz][zz]
              + f[zz][ww][ww];
         q := f[yy][xx][yy] + f[yy][xx][zz] + f[yy][yy][xx]
              + f[yy][yy][zz] + f[yy][zz][xx] + f[yy][zz][yy]
              + f[yy][zz][ww] + f[zz][xx][yy] + f[zz][xx][zz]
              + f[zz][xx][ww] + f[zz][yy][xx] + f[zz][yy][zz]
              + f[zz][yy][ww] + f[zz][zz][xx] + f[zz][zz][yy]
              + f[zz][zz][ww] + f[zz][ww][xx] + f[zz][ww][yy]
              + f[zz][ww][zz];
         L1 := n*p - m*q;
         tabulate (m, n, p, q);
         WRITELN(OUTPUT, '  Tree II:');
         m := f[xx][xx][xx] + f[yy][xx][yy] + f[zz][xx][zz];
         n := f[xx][xx][yy] + f[xx][xx][zz] + f[yy][xx][xx]
              + f[yy][xx][zz] + f[zz][xx][xx] + f[zz][xx][yy]
              + f[zz][xx][ww];
         p := f[xx][yy][xx] + f[yy][yy][yy] + f[zz][yy][zz]
              + f[xx][zz][xx] + f[yy][zz][yy] + f[zz][zz][zz]
              + f[zz][ww][zz];
         q := f[xx][yy][yy] + f[xx][yy][zz] + f[yy][yy][xx]
              + f[yy][yy][zz] + f[zz][yy][xx] + f[zz][yy][yy]
              + f[zz][yy][ww] + f[xx][zz][yy] + f[xx][zz][zz]
              + f[xx][zz][ww] + f[yy][zz][xx] + f[yy][zz][zz]
              + f[yy][zz][ww] + f[zz][zz][xx] + f[zz][zz][yy]
              + f[zz][zz][ww] + f[zz][ww][xx] + f[zz][ww][yy]
              + f[zz][ww][ww];
         L2 := n*p - m*q;
         tabulate (m, n, p, q);
         WRITELN(OUTPUT, '  Tree III:');
         m := f[xx][xx][xx] + f[yy][yy][xx] + f[zz][zz][xx];
         n := f[yy][xx][xx] + f[zz][xx][xx] + f[xx][yy][xx]
              + f[zz][yy][xx] + f[xx][zz][xx] + f[yy][zz][xx]
              + f[zz][ww][xx];
         p := f[xx][xx][yy] + f[yy][yy][yy] + f[zz][zz][yy]
              + f[xx][xx][zz] + f[yy][yy][zz] + f[zz][zz][zz]
              + f[zz][zz][ww];
         q := f[yy][xx][yy] + f[zz][xx][yy] + f[xx][yy][yy]
              + f[zz][yy][yy] + f[xx][zz][yy] + f[yy][zz][yy]
              + f[zz][ww][yy] + f[yy][xx][zz] + f[zz][xx][zz]
              + f[xx][zz][ww] + f[xx][yy][zz] + f[zz][yy][zz]
              + f[zz][yy][ww] + f[xx][zz][zz] + f[yy][zz][zz]
              + f[zz][ww][ww] + f[zz][xx][ww] + f[yy][zz][ww]
              + f[zz][ww][zz];
         L3 := n*p - m*q;
         tabulate (m, n, p, q);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, 'Cavender''s quadratic invariants (type K)');
         WRITELN(OUTPUT, ' using purines vs. pyrimidines');
         WRITELN(OUTPUT, ' (these are expected to be zero for the correct',
                             ' tree topology)');
         WRITELN(OUTPUT, 'They will be misled if there are substantially');
         WRITELN(OUTPUT, 'different evolutionary rate between sites, or');
         WRITELN(OUTPUT, 'different purine:pyrimidine ratios from 1:1.');
         WRITELN(OUTPUT, 'No statistical test is done on them here.');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '  Tree I:   ', (L2 - L3):15:1);
         WRITELN(OUTPUT, '  Tree II:  ', (L3 - L1):15:1);
         WRITELN(OUTPUT, '  Tree III: ', (L1 - L2):15:1);
         WRITELN(OUTPUT);
      END; (* invariants *)

   BEGIN (* makeinv *)
      printpatterns;
      makesymmetries;
      printsymmetries;
      invariants;
   END; (* makeinv *)

BEGIN (* DNA Invariants *)
   (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
   RESET(infile);
   getinput;
   IF NOT anerror
   THEN makeinv;
(* CLOSE(infile);  -- may need this here *)
END. (* DNA Invariants *)

