PROGRAM DNAML (infile, treefile, OUTPUT);
  (* version 3.31. (c) Copyright 1986, 1990 by the University of Washington 
    and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is 
    granted to copy and use this program provided no fee is charged for it 
    and provided that this copyright notice is not removed. *)

  CONST maxsp = 10;     (* maximum number of species *)
    maxsp2 =  18;       (* 2 x maximum number of species - 2 *)
    maxsites = 2000;    (* maximum number of sites *)
    maxpatterns = 512;  (* max number of different site patterns *)
    maxcategories = 9;  (* maximum number of site types *)
    maxtrees = 10;      (* maximum number of user trees for test *)
    smoothings = 20;    (* number of passes through smoothing algorithm *)
    iterations = 10;    (* number of iterates of EM for each branch *)
    nmlngth = 10;       (* max. number of characters in species name *)
    epsilon = 0.000001; (* used in update *)

  TYPE base = (A, C, G, T);
    sitelike = ARRAY[A..T] OF REAL;
    phenotype = ARRAY [1..maxpatterns] OF sitelike;
    sequence = ARRAY[1..maxsp] OF ARRAY [1..maxsites] OF CHAR;
    naym = ARRAY[1..nmlngth] OF CHAR;
    longer = ARRAY[0..5] OF INTEGER;
    nodeptr = ^node;
    node = RECORD
      next, back : nodeptr;
      tip        : BOOLEAN;
      number     : INTEGER;
      x          : phenotype;
      nayme      : naym;
      v          : REAL;
      xcoord, ycoord, ymin, ymax : INTEGER;
      END;
    tree = RECORD
      nodep      : ARRAY[1..maxsp2] OF nodeptr;
      likelihood : REAL;
      start      : nodeptr
      END;

  VAR infile, treefile : TEXT;
    numsp, numsp1, numsp2, sites, endsite,
      weightsum, categs, inseed, outgrno : INTEGER;
    anerror, freqsfrom, global, jumble, lengths,
      outgropt, weights, trout, usertree : BOOLEAN;
    curtree, bestree : tree;
    xi, xv, ttratio, freqa, freqc, freqg, freqt, freqr, freqy,
             freqar, freqcy, freqgr, freqty, fracchange, sumrates : REAL;
    seed : longer;
    enterorder : ARRAY[1..maxsp] OF INTEGER;
    y : sequence;
    category, weight, alias, aliasweight : ARRAY[1..maxsites] OF INTEGER;
    rate : ARRAY[1..maxcategories] OF REAL;

  FUNCTION randum (VAR seed : longer) : REAL;
    (* random number generator -- slow but machine independent *)
    VAR i, j, k, sum : INTEGER;
      mult, newseed : longer;
      x : REAL;
  BEGIN (* randum -- slow but machine independent *)
    mult[0] := 13;
    mult[1] := 24;
    mult[2] := 22;
    mult[3] :=  6;
    FOR i := 0 TO 5 DO
      newseed[i] := 0;
    FOR i := 0 TO 5 DO BEGIN
      sum := newseed[i];
      k := i;
      IF i > 3 THEN k := 3;
      FOR j := 0 TO k DO
        sum := sum + mult[j] * seed[i-j];
      newseed[i] := sum;
      FOR j := i TO 4 DO BEGIN
        newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
        newseed[j] := newseed[j] MOD 64;
        END;
      END;
    seed := newseed;
    seed[5] := seed[5] MOD 4;
    x := 0.0;
    FOR i := 0 TO 5 DO
      x := x / 64.0 + seed[i];
    x := x / 4.0;
    randum := x;
  END; (* randum *)

  PROCEDURE getinput;
    (* read input file *)

    PROCEDURE getnums;
      (* input number of species, number of sites *)
    BEGIN (* getnums *)
      WRITELN(OUTPUT);
      WRITE(OUTPUT,'Nucleic acid sequence Maximum Likelihood');
      WRITELN(OUTPUT, ' method, version 3.31');
      WRITELN(OUTPUT);
      READ(infile, numsp, sites);
      IF numsp > maxsp
      THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: adjust CONSTants');
      IF sites > maxsites
      THEN WRITELN(OUTPUT, 'TOO MANY SITES: adjust CONSTants');
      WRITELN(OUTPUT, numsp:4, ' Species, ', sites:4, ' Sites');
      anerror := (numsp > maxsp) OR (sites > maxsites);
      numsp1 := numsp + 1;
      numsp2 := 2*numsp - 2;
    END; (* getnums *)

    FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
        -- works for both ASCII and EBCDIC codes *)
    BEGIN (* letter *)
       letter := ((ch >= 'A') AND (ch <= 'I'))
             OR ((ch >= 'J') AND (ch <= 'R'))
             OR ((ch >= 'S') AND (ch <= 'Z'))
             OR ((ch >= 'a') AND (ch <= 'i'))
             OR ((ch >= 'j') AND (ch <= 'r'))
             OR ((ch >= 's') AND (ch <= 'z'));
    END; (* letter *)

    PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
    BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                 OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
    END; (* uppercase *)

    PROCEDURE getoptions;
      VAR ch : CHAR;
        i, j, extranum : INTEGER;

      PROCEDURE inputcategories;
        (* reads the categories for each site *)
        VAR i : INTEGER;
          ch : CHAR;
      BEGIN (* inputcategories *)
        READLN(infile);
        FOR i := 1 TO nmlngth DO READ(infile, ch);
        FOR i := 1 TO sites DO BEGIN
          REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
          UNTIL (ch <> ' ');
          category[i] := ORD(ch) - ORD('0');
          END;
      END; (* inputcategories *)

      PROCEDURE printcategories;
        (* print out list of categories of sites *)
        VAR i, j : INTEGER;
      BEGIN (* printcategories *)
        WRITELN(OUTPUT, 'Rate categories');
        FOR i := 1 TO sites DO BEGIN
          IF ((i-1) MOD 60) = 0
          THEN BEGIN
            WRITELN(OUTPUT);
            FOR j := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
            END;
          IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
          THEN WRITE(OUTPUT, ' ');
          WRITE(OUTPUT, category[i]:1);
          END;
        WRITELN(OUTPUT);
        WRITELN(OUTPUT);
      END; (* printcategories *)

      PROCEDURE inputweights;
        (* input the character weights, 0 or 1 *)
        VAR ch : CHAR;
          i : INTEGER;
      BEGIN (* inputweights *)
        FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
        weightsum := 0;
        FOR i := 1 TO sites DO BEGIN
          REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
          UNTIL ch <> ' ';
          weight[i] := 1;
          IF (ch = '0') OR (ch = '1')
          THEN weight[i] := ORD(ch) - ORD('0')
          ELSE BEGIN
            WRITELN(OUTPUT, 'BAD WEIGHT CHARACTER: ', ch,
              ' -- WEIGHTS IN DNAML MUST BE 0 OR 1');
            anerror := TRUE;
            END;
          weightsum := weightsum + weight[i];
          END;
        READLN(infile);
        weights := TRUE;
      END; (* inputweights *)

      PROCEDURE printweights;
        (* print out the weights of sites *)
        VAR i, j : INTEGER;
      BEGIN (* printweights *)
        WRITELN(OUTPUT);
        WRITELN(OUTPUT,'   Sites are weighted as follows:');
        FOR i := 1 TO sites DO BEGIN
          IF ((i-1) MOD 60) = 0
          THEN BEGIN
            WRITELN(OUTPUT);
            FOR j := 1 TO nmlngth+3 DO WRITE(OUTPUT, ' ');
            END;
          WRITE(OUTPUT, weight[i]:1);
          IF ((i MOD 10) = 0) AND ((i MOD 60) <> 0)
          THEN WRITE(OUTPUT, ' ');
          END;
        WRITELN(OUTPUT);
        WRITELN(OUTPUT);
      END; (* printweights *)

    BEGIN (* getoptions *)
      categs := 1;
      rate[1] := 1.0;
      FOR i := 1 TO sites DO category[i] := 1;
      freqsfrom := FALSE;
      global := FALSE;
      jumble := FALSE;
      lengths := FALSE;
      trout := FALSE;
      usertree := FALSE;
      weights := FALSE;
      FOR i := 1 TO sites DO weight[i] := 1;
      weightsum := sites;
      ttratio := 2.0;
      outgropt := FALSE;
      outgrno := 1;
      extranum := 0;
      WHILE NOT EOLN(infile) DO BEGIN
        READ(infile, ch);
        uppercase (ch);
        IF (ch = 'C') OR (ch = 'F') OR (ch = 'G') OR (ch = 'J') OR
          (ch = 'L') OR (ch = 'O') OR (ch = 'T') OR (ch = 'U') OR
          (ch = 'W') OR (ch = 'Y')
        THEN CASE ch OF
            'C' : extranum := extranum + 1;
            'F' : freqsfrom := TRUE;
            'G' : global := TRUE;
            'J' : BEGIN
                jumble := TRUE;
                extranum := extranum + 1;
                END;
            'L' : lengths := TRUE;
            'O' : BEGIN
                outgropt := TRUE;
                extranum := extranum + 1;
                END;
            'T' : extranum := extranum + 1;
            'U' : usertree := TRUE;
            'W' : extranum := extranum + 1;
            'Y' : trout := TRUE
            END
        ELSE IF ch <> ' '
          THEN BEGIN
            WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
            anerror := TRUE;
            END;
        END;
      FOR i := 1 TO extranum DO
        IF NOT anerror
        THEN BEGIN
          READLN(infile);
          READ(infile, ch);
          uppercase (ch);
          anerror := NOT ((ch = 'C') OR (ch = 'J')
            OR (ch = 'O') OR (ch = 'T') OR (ch = 'W'));
          IF anerror
          THEN WRITELN(OUTPUT, 'ERROR: INCORRECT AUXILIARY',
                ' OPTIONS LINE WHICH STARTS WITH ', ch);
          IF ch = 'C'
          THEN BEGIN
            READ(infile, categs);
            FOR j := 1 TO categs DO BEGIN
              IF EOLN(infile) THEN READLN(infile);
              READ(infile, rate[j]);
              END;
            END;
          IF ch = 'J'
          THEN READ(infile, inseed);
          IF ch = 'O'
          THEN BEGIN
            READ(infile, outgrno);
            IF (outgrno < 1) OR (outgrno > numsp)
            THEN BEGIN
              WRITELN(OUTPUT, 'BAD OUTGROUP NUMBER: ', outgrno:6);
              anerror := TRUE;
              END;
            END;
          IF ch = 'T'
          THEN READ(infile, ttratio);
          IF ch = 'W'
          THEN inputweights;
          END;
      IF jumble
      THEN BEGIN
        WRITELN(OUTPUT);
        WRITELN(OUTPUT, 'Random number seed = ', inseed:8);
        FOR i := 0 TO 5 DO seed[i] := 0;
        i := 0;
        REPEAT
          seed[i] := inseed MOD 64;
          inseed := inseed DIV 64;
          i := i + 1;
        UNTIL inseed = 0;
        END;
      IF categs > 1
      THEN BEGIN
        WRITELN(OUTPUT);
        WRITELN(OUTPUT, 'Site category   Rate of change');
        WRITELN(OUTPUT);
        FOR i := 1 TO categs DO
          WRITELN(OUTPUT, i:12, rate[i]:13:3);
        WRITELN(OUTPUT);
        inputcategories;
        printcategories;
        END;
      IF weights
      THEN printweights;
    END; (* getoptions *)

    PROCEDURE getbasefreqs;
      VAR aa, bb : REAL;
    BEGIN (* getbasefreqs *)
      IF NOT freqsfrom THEN READLN(infile);
      WRITELN(OUTPUT);
      IF freqsfrom
      THEN WRITE(OUTPUT, 'Empirical ');
      WRITELN(OUTPUT, 'Base Frequencies:');
      WRITELN(OUTPUT);
      IF NOT freqsfrom
      THEN BEGIN
        READ(infile, freqa);
        READ(infile, freqc);
        READ(infile, freqg);
        READ(infile, freqt);
        END;
      WRITELN(OUTPUT, '   A    ', freqa:10:5);
      WRITELN(OUTPUT, '   C    ', freqc:10:5);
      WRITELN(OUTPUT, '   G    ', freqg:10:5);
      WRITELN(OUTPUT, '  T(U)  ', freqt:10:5);
      freqr := freqa + freqg;
      freqy := freqc + freqt;
      freqar := freqa/freqr;
      freqcy := freqc/freqy;
      freqgr := freqg/freqr;
      freqty := freqt/freqy;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, 'Transition/transversion ratio = ', ttratio:10:6);
      WRITELN(OUTPUT);
      aa := ttratio*freqr*freqy - (freqa*freqg + freqc*freqt);
      bb := freqa*freqgr + freqc*freqty;
      xi := aa/(aa+bb);
      xv := 1.0 - xi;
      ttratio := xi / xv;
      IF xi <= 0.0
      THEN BEGIN
        WRITELN(OUTPUT, 'WARNING: This transition/transversion ratio');
        WRITELN(OUTPUT, 'is impossible with these base frequencies!');
        xi := 3/5;
        xv := 2/5;
        WRITELN(OUTPUT, ' Transition/transversion parameter reset');
        WRITELN(OUTPUT);
        END;
      WRITELN(OUTPUT, '(Transition/transversion parameter = ',
        xi/xv:10:6, ')');
      WRITELN(OUTPUT);
      fracchange := xi*(2*freqa*freqgr + 2*freqc*freqty)
                + xv*(1.0 - freqa*freqa - freqc*freqc
                       - freqg*freqg - freqt*freqt);
    END; (* getbasefreqs *)

    PROCEDURE setuptree (VAR a : tree);
      VAR i, j : INTEGER;
        p, q : nodeptr;
    BEGIN (* setuptree *)
      FOR i := 1 TO numsp DO BEGIN
        NEW(a.nodep[i]);
        a.nodep[i]^.tip := TRUE;
        a.nodep[i]^.number := i;
        END;
      FOR i := numsp1 TO numsp2 DO BEGIN
        q := NIL;
        FOR j := 1 TO 3 DO BEGIN
          NEW(p);
          p^.tip := FALSE;
          p^.number := i;
          p^.next := q;
          q := p;
          END;
        p^.next^.next^.next := p;
        a.nodep[i] := p;
        END;
      a.likelihood := -999999.0;
      a.start := a.nodep[1];
    END; (* setuptree *)

    PROCEDURE getdata;
      (* read sequences *)
      VAR i, j, k, l, basesread, basesnew : INTEGER;
        ch : CHAR;
        allread : BOOLEAN;
    BEGIN (* getdata *)
      WRITELN(OUTPUT);
      j := nmlngth - 5 + ((sites + ((sites-1) DIV 10)) DIV 2);
      IF j < nmlngth - 1 THEN j := nmlngth - 1;
      IF j > 37 THEN j := 37;
      WRITE(OUTPUT, 'Name');
      FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
      WRITELN(OUTPUT, 'Sequences');
      WRITE(OUTPUT, '----');
      FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
      WRITELN(OUTPUT, '---------');
      WRITELN(OUTPUT);
      setuptree (curtree);
      basesread := 0;
      allread := FALSE;
      WHILE NOT (allread OR anerror) DO BEGIN
        allread := TRUE;
        IF EOLN(infile) THEN READLN(infile);
        FOR i := 1 TO numsp DO
          IF NOT anerror
          THEN BEGIN
            IF basesread = 0
            THEN FOR j := 1 TO nmlngth DO
                IF NOT anerror
                THEN BEGIN
                  READ(infile, curtree.nodep[i]^.nayme[j]);
                  anerror := anerror OR EOF(infile) OR EOLN(infile);
                  IF anerror
                  THEN WRITELN(OUTPUT, 'ERROR: END-OF-LINE OR ',
                     'END-OF-FILE IN THE MIDDLE OF A SPECIES NAME');
                  END;
            j := basesread;
            WHILE (j < sites) AND (NOT (EOLN(infile) OR EOF(infile)))
              AND (NOT anerror) DO BEGIN
              READ(infile, ch);
              IF (ch <> ' ') AND
                ((ch < '0') OR (ch > '9'))
              THEN BEGIN
                uppercase (ch);
                IF NOT ((ch = 'A') OR (ch = 'B') OR (ch = 'C') OR (ch = 'D')
                  OR (ch = 'G') OR (ch = 'H') OR (ch = 'K') OR (ch = 'M')
                  OR (ch = 'N') OR (ch = 'R') OR (ch = 'S') OR (ch = 'T')
                  OR (ch = 'U') OR (ch = 'V') OR (ch = 'W') OR (ch = 'X')
                  OR (ch = 'Y') OR (ch = '?') OR (ch = 'O') OR (ch = '-')
                  OR (ch = '.'))
                THEN BEGIN
                  WRITELN(OUTPUT, 'ERROR: BAD BASE:', ch,
                    ' AT POSITION', j:5, ' OF SPECIES ', i:3);
                  anerror := TRUE;
                  END;
                j := j + 1;
                IF ch = '.' THEN ch := y[1][j];
                y[i][j] := ch;
                END;
              END;
            IF i = 1 THEN basesnew := j;
            READLN(infile);
            anerror := anerror OR (j <> basesnew);
            IF anerror
            THEN WRITELN(OUTPUT, 'ERROR: SEQUENCES OUT OF ALIGNMENT');
            END;
        basesread := basesnew;
        allread := basesread = sites;
        END;
      IF NOT anerror
      THEN BEGIN
        FOR i := 1 TO ((sites-1) DIV 60)+1 DO BEGIN
          FOR j := 1 TO numsp DO BEGIN
            FOR k := 1 TO nmlngth DO
              WRITE(OUTPUT, curtree.nodep[j]^.nayme[k]);
            WRITE(OUTPUT, '   ');
            l := 60*i;
            IF l > sites THEN l := sites;
            FOR k := 60*(i-1)+1 TO l DO
              IF NOT anerror
              THEN BEGIN
                IF (j > 1) AND (y[j][k] = y[1][k])
                THEN ch := '.'
                ELSE ch := y[j][k];
                WRITE(OUTPUT, ch);
                IF ((k MOD 10) = 0) AND ((k MOD 60) <> 0)
                THEN WRITE(OUTPUT, ' ');
                END;
            WRITELN(OUTPUT);
            END;
          WRITELN(OUTPUT);
          END;
        END;
      IF (NOT usertree)
      THEN setuptree(bestree);
    END; (* getdata *)

    PROCEDURE makeweights;
      (* make up weights vector to avoid duplicate computations *)
      VAR i : INTEGER;

      PROCEDURE sitesort;
        (* Shell sort keeping sites, weights in same order *)
        VAR gap, i, j, jj, jg, k, itemp : INTEGER;
          flip, tied : BOOLEAN;
      BEGIN (* sitesort *)
        gap := sites DIV 2;
        WHILE (gap > 0) DO BEGIN
          FOR i := gap TO sites-1 DO BEGIN
            j := i - gap + 1;
            flip := TRUE;
            WHILE ((j>0) AND flip) DO BEGIN
              jj := alias[j];
              jg := alias[j+gap];
              flip := (category[jj] > category[jg]);
              tied := (category[jj] = category[jg]);
              k := 1;
              WHILE (k <= numsp) AND tied DO BEGIN
                flip := y[k][jj] > y[k][jg];
                tied := tied AND (y[k][jj] = y[k][jg]);
                k := k + 1;
                END;
              IF flip
              THEN BEGIN
                itemp := alias[j];
                alias[j] := alias[j+gap];
                alias[j+gap] := itemp;
                itemp := aliasweight[j];
                aliasweight[j] := aliasweight[j+gap];
                aliasweight[j+gap] := itemp;
                j := j - gap;
                END;
              END;
            END;
          gap := gap DIV 2;
          END;
      END; (* sitesort *)

      PROCEDURE sitecombine;
        (* combine sites that have identical patterns *)
        VAR i, j, k : INTEGER;
          tied : BOOLEAN;
      BEGIN (* sitecombine *)
        i := 1;
        WHILE i < sites DO BEGIN
          j := i+1;
          tied := TRUE;
          WHILE (j <= sites) AND tied DO BEGIN
            tied := tied AND (category[alias[i]] = category[alias[j]]);
            k := 1;
            WHILE (k <= numsp) AND tied DO BEGIN
              tied := tied AND (y[k][alias[i]] = y[k][alias[j]]);
              k := k + 1;
              END;
            IF tied AND (aliasweight[j] > 0)
            THEN BEGIN
              aliasweight[i] := aliasweight[i] + aliasweight[j];
              aliasweight[j] := 0;
              alias[j] := alias[i];
              END;
            j := j + 1;
            END;
          i := j - 1;
          END;
      END; (* sitecombine *)

      PROCEDURE sitescrunch;
        (* move so positively weighted sites come first *)
        VAR i, j, itemp : INTEGER;
          done, found : BOOLEAN;
      BEGIN (* sitescrunch *)
        done := FALSE;
        i := 1;
        j := 2;
        WHILE NOT done DO BEGIN
          found := FALSE;
          IF aliasweight[i] > 0
          THEN i := i + 1
          ELSE BEGIN
            IF j <= i
            THEN j := i + 1;
            IF j <= sites
            THEN BEGIN
              found := FALSE;
              REPEAT
                found := aliasweight[j] > 0;
                j := j + 1;
              UNTIL found OR (j > sites);
              IF found
              THEN BEGIN
                j := j - 1;
                itemp := alias[i];
                alias[i] := alias[j];
                alias[j] := itemp;
                itemp := aliasweight[i];
                aliasweight[i] := aliasweight[j];
                aliasweight[j] := itemp;
                END
              ELSE done := TRUE;
              END
            ELSE done := TRUE;
            END;
          done := done OR (i >= sites);
          END;
      END; (* sitescrunch *)

    BEGIN (* makeweights *)
      FOR i := 1 TO sites DO BEGIN
        alias[i] := i;
        aliasweight[i] := weight[i];
        END;
      sitesort;
      sitecombine;
      sitescrunch;
      FOR i := 1 TO sites DO BEGIN
        weight[i] := aliasweight[i];
        IF weight[i] > 0
        THEN endsite := i;
        END;
      IF endsite > maxpatterns
      THEN BEGIN
        WRITELN(OUTPUT, 'TOO MANY PATTERNS: increase',
            ' CONSTant maxpatterns to at least ', endsite:5);
        anerror := TRUE;
        END;
      sumrates := 0.0;
      FOR i := 1 TO sites DO
        sumrates := sumrates + weight[i]*rate[category[alias[i]]];
      FOR i := 1 TO categs DO rate[i] := rate[i] * (weightsum/sumrates);
    END; (* makeweights *)

    PROCEDURE makevalues;
      (* set up fractional likelihoods at tips *)
      VAR i, j, k : INTEGER;
        b : base;
    BEGIN (* makevalues *)
      FOR k := 1 TO endsite DO BEGIN
        j := alias[k];
        FOR i := 1 TO numsp DO BEGIN
          FOR b := A TO T DO
            curtree.nodep[i]^.x[k][b] := 0.0;
          CASE y[i][j] OF
            'A' : curtree.nodep[i]^.x[k][A] := 1.0;
            'C' : curtree.nodep[i]^.x[k][C] := 1.0;
            'G' : curtree.nodep[i]^.x[k][G] := 1.0;
            'T' : curtree.nodep[i]^.x[k][T] := 1.0;
            'U' : curtree.nodep[i]^.x[k][T] := 1.0;
            'M' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][C] := 1.0;
                END;
            'R' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                END;
            'W' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'S' : BEGIN
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                END;
            'Y' : BEGIN
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'K' : BEGIN
                curtree.nodep[i]^.x[k][G] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'B' : BEGIN
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'D' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'H' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][T] := 1.0;
                END;
            'V' : BEGIN
                curtree.nodep[i]^.x[k][A] := 1.0;
                curtree.nodep[i]^.x[k][C] := 1.0;
                curtree.nodep[i]^.x[k][G] := 1.0;
                END;
            'N' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            'X' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            '?' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            'O' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0;
            '-' : FOR b := A TO T DO curtree.nodep[i]^.x[k][b] := 1.0
            END;
          END;
        END;
    END; (* makevalues *)

    PROCEDURE empiricalfreqs;
      (* Get empirical base frequencies from the data *)
      VAR i, j, k : INTEGER;
        sum, suma, sumc, sumg, sumt, w : REAL;
    BEGIN (* empiricalfreqs *)
      freqa := 0.25;
      freqc := 0.25;
      freqg := 0.25;
      freqt := 0.25;
      FOR k := 1 TO 8 DO BEGIN
        suma := 0.0;
        sumc := 0.0;
        sumg := 0.0;
        sumt := 0.0;
        FOR i := 1 TO numsp DO
          FOR j := 1 TO endsite DO BEGIN
            w := weight[j];
            sum := freqa*curtree.nodep[i]^.x[j][A];
            sum := sum + freqc*curtree.nodep[i]^.x[j][C];
            sum := sum + freqg*curtree.nodep[i]^.x[j][G];
            sum := sum + freqt*curtree.nodep[i]^.x[j][T];
            suma := suma + w*freqa*curtree.nodep[i]^.x[j][A] / sum;
            sumc := sumc + w*freqc*curtree.nodep[i]^.x[j][C] / sum;
            sumg := sumg + w*freqg*curtree.nodep[i]^.x[j][G] / sum;
            sumt := sumt + w*freqt*curtree.nodep[i]^.x[j][T] / sum;
            END;
        sum := suma + sumc + sumg + sumt;
        freqa := suma / sum;
        freqc := sumc / sum;
        freqg := sumg / sum;
        freqt := sumt / sum;
        END;
    END; (* empiricalfreqs *)

  BEGIN (* getinput *)
    getnums;
    IF NOT anerror THEN getoptions;
    IF (NOT freqsfrom) AND (NOT anerror)
    THEN getbasefreqs;
    IF NOT anerror THEN getdata;
    IF NOT anerror THEN makeweights;
    IF NOT anerror
    THEN BEGIN
      makevalues;
      IF freqsfrom
      THEN BEGIN
        empiricalfreqs;
        getbasefreqs
        END;
      END;
  END; (* getinput *)

  PROCEDURE maketree;
    VAR i, j, k, nextsp, numtrees, num, which, maxwhich : INTEGER;
      maxlogl, sum, sum2, sd : REAL;
      succeeded, smoothed : BOOLEAN;
      logl : ARRAY [1..maxtrees] OF REAL;
      logf : ARRAY [1..maxtrees] OF ARRAY [1..maxpatterns] OF REAL;

    FUNCTION evaluate (VAR tr : tree) : REAL;
      VAR sum, y, yy, z, lz, zz, y1, z1, rat,
          prod12, prod1, prod2, prod3, term : REAL;
        i : INTEGER;
        p, q : nodeptr;
        x1, x2 : sitelike;
    BEGIN (* evaluate *)
      sum := 0.0;
      p := tr.start;
      q := p^.back;
      y := p^.v;
      z := 1.0 - y;
      lz := LN(z);
      FOR i := 1 TO endsite DO BEGIN
        rat := rate[category[alias[i]]];
        IF z > 0.0
        THEN BEGIN
          zz := EXP(rat*xi*lz);
          z1 := EXP(rat*xv*lz);
          END
        ELSE BEGIN
          zz := 0.0;
          z1 := 0.0;
          END;
        yy := 1.0 - zz;
        y1 := 1.0 - z1;
        x1 := p^.x[i];
        prod1 := freqa*x1[A] + freqc*x1[C] + freqg*x1[G] + freqt*x1[T];
        x2 := q^.x[i];
        prod2 := freqa*x2[A] + freqc*x2[C] + freqg*x2[G] + freqt*x2[T];
        prod3 := (x1[A]*freqa+x1[G]*freqg)*(x2[A]*freqar+x2[G]
                *freqgr) + (x1[C]*freqc+x1[T]*freqt)
                      *(x2[C]*freqcy+x2[T]*freqty);
        prod12 := freqa*x1[A]*x2[A] + freqc*x1[C]*x2[C]
               + freqg*x1[G]*x2[G] + freqt*x1[T]*x2[T];
        term := LN(z1*(zz*prod12 + yy*prod3)+y1*prod1*prod2);
        IF usertree AND (which <= maxtrees)
        THEN logf[which][i] := term;
        sum := sum + weight[i]*term;
        END;
      tr.likelihood := sum;
      IF usertree
      THEN BEGIN
        logl[which] := sum;
        IF which = 1
        THEN BEGIN
          maxwhich := 1;
          maxlogl := sum;
          END
        ELSE IF sum > maxlogl
          THEN BEGIN
            maxwhich := which;
            maxlogl := sum;
            END;
        END;
      evaluate := sum;
    END; (* evaluate *)

    PROCEDURE nuview (p : nodeptr);
      VAR i : INTEGER;
        w1, w2, lw1, lw2, ww1, ww2, zz1, zz2, vv1, vv2,
          yy1, yy2, rat, sum1, sum2, sumr1, sumr2,
          sumy1, sumy2 : REAL;
        q, r : nodeptr;
        xx1, xx2, xx3 : sitelike;
    BEGIN (* nuview *)
      q := p^.next^.back;
      r := p^.next^.next^.back;
      w1 := 1.0 - q^.v;
      w2 := 1.0 - r^.v;
      IF w1 > 0.0
      THEN lw1 := LN(w1);
      IF w2 > 0.0
      THEN lw2 := LN(w2);
      FOR i := 1 TO endsite DO BEGIN
        rat := rate[category[alias[i]]];
        IF w1 <= 0.0
        THEN BEGIN
          ww1 := 0.0;
          zz1 := 0.0;
          END
        ELSE BEGIN
          ww1 := EXP(rat*xi*lw1);
          zz1 := EXP(rat*xv*lw1);
          END;
        yy1 := 1.0 - zz1;
        vv1 := 1.0 - ww1;
        IF w2 <= 0.0
        THEN BEGIN
          ww2 := 0.0;
          zz2 := 0.0;
          END
        ELSE BEGIN
          ww2 := EXP(rat*xi*lw2);
          zz2 := EXP(rat*xv*lw2);
          END;
        yy2 := 1.0 - zz2;
        vv2 := 1.0 - ww2;
        xx1 := q^.x[i];
        xx2 := r^.x[i];
        sum1 := yy1*(freqa*xx1[A] + freqc*xx1[C]
              + freqg*xx1[G] + freqt*xx1[T]);
        sum2 := yy2*(freqa*xx2[A] + freqc*xx2[C]
              + freqg*xx2[G] + freqt*xx2[T]);
        sumr1 :=vv1*(freqar*xx1[A] + freqgr*xx1[G]);
        sumr2 :=vv2*(freqar*xx2[A] + freqgr*xx2[G]);
        sumy1 :=vv1*(freqcy*xx1[C] + freqty*xx1[T]);
        sumy2 :=vv2*(freqcy*xx2[C] + freqty*xx2[T]);
        xx3[A] := (sum1 + zz1*(ww1*xx1[A] + sumr1))
                *(sum2 + zz2*(ww2*xx2[A] + sumr2));
        xx3[C] := (sum1 + zz1*(ww1*xx1[C] + sumy1))
                *(sum2 + zz2*(ww2*xx2[C] + sumy2));
        xx3[G] := (sum1 + zz1*(ww1*xx1[G] + sumr1))
                *(sum2 + zz2*(ww2*xx2[G] + sumr2));
        xx3[T] := (sum1 + zz1*(ww1*xx1[T] + sumy1))
                *(sum2 + zz2*(ww2*xx2[T] + sumy2));
        p^.x[i] := xx3;
        END;
    END; (* nuview *)

    PROCEDURE update (p : nodeptr);

      PROCEDURE makenewv (p : nodeptr);
        VAR i, it : INTEGER;
          sum, denom, sum1, sum2, rat, y, z, lz, yy, zz,
            y1, z1, yold, aa, bb, cc : REAL;
          prod, prod2, prod3 : ARRAY[1..maxpatterns] OF REAL;
          done : BOOLEAN;
          q : nodeptr;
          xx1, xx2 : sitelike;
      BEGIN (* makenewv *)
        q := p^.back;
        y := p^.v;
        FOR i := 1 TO endsite DO BEGIN
          xx1 := p^.x[i];
          xx2 := q^.x[i];
          sum1 := freqa*xx1[A] + freqc*xx1[C]
                 + freqg*xx1[G] + freqt*xx1[T];
          sum2 := freqa*xx2[A] + freqc*xx2[C]
                 + freqg*xx2[G] + freqt*xx2[T];
          prod[i] := sum1*sum2;
          prod2[i] := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar
                   +xx2[G]*freqgr) + (xx1[C]*freqc+xx1[T]
                   *freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
          prod3[i] := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                   + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
          END;
        IF y < epsilon
        THEN y := 10.0*epsilon;
        IF y = 1.0
        THEN y := 1.0 - 10.0*epsilon;
        done := FALSE;
        it := 1;
        WHILE (it < iterations) AND (NOT done) DO BEGIN
          sum := 0.0;
          denom := 0.0;
          z := 1.0 - y;
          IF z > 0.0
          THEN lz := LN(z);
          FOR i := 1 TO endsite DO BEGIN
            rat := rate[category[alias[i]]];
            IF z > 0.0
            THEN BEGIN
              zz := EXP(rat*xi*lz);
              z1 := EXP(rat*xv*lz);
              END
            ELSE BEGIN
              zz := 0.0;
              z1 := 0.0;
              END;
            yy := 1.0 - zz;
            y1 := 1.0 - z1;
            cc := prod[i];
            bb := prod2[i];
            aa := prod3[i];
            sum := sum + weight[i]*rat*(xi*(bb*z1+cc*y1)+xv*cc)
                        /(aa*z1*zz + bb*z1*yy + cc*y1);
            denom := denom + weight[i];
            END;
          yold := y;
          y := 1.0 - EXP(lz*sum/denom);
          done := ABS(yold-y) < epsilon;
          it := it + 1;
          END;
        smoothed := smoothed AND done;
        p^.v := y;
        q^.v := y;
      END; (* makenewv *)

    BEGIN (* update *)
      IF NOT p^.tip
      THEN nuview(p);
      IF NOT p^.back^.tip
      THEN nuview(p^.back);
      IF NOT lengths
      THEN makenewv(p);
    END; (* update *)

    PROCEDURE smooth (p : nodeptr);
    BEGIN (* smooth *)
      update (p);
      IF NOT p^.tip
      THEN BEGIN
        smooth (p^.next^.back);
        smooth (p^.next^.next^.back);
        END;
    END; (* smooth *)

    PROCEDURE hookup (p, q : nodeptr);
    BEGIN (* hookup *)
      p^.back := q;
      q^.back := p;
    END; (* hookup *)

    PROCEDURE insert (p, q: nodeptr);
      VAR i : INTEGER;
        r : nodeptr;
    BEGIN (* insert *)
      r := p^.next^.next;
      hookup (r, q^.back);
      hookup (p^.next, q);
      q^.v := 0.5 * q^.v;
      q^.back^.v := q^.v;
      r ^.v := q^.v;
      r^.back^.v := r^.v;
      smoothed := FALSE;
      i := 1;
      WHILE (i < smoothings) AND (NOT smoothed) DO BEGIN
        smoothed := TRUE;
        smooth (p);
        smooth (p^.back);
        i := i + 1;
        END;
    END; (* insert *)

    PROCEDURE remove (VAR p, q : nodeptr);
      (* remove p and record in q where it was *)
    BEGIN (* remove *)
      q := p^.next^.back;
      hookup (q, p^.next^.next^.back);
      p^.next^.back := NIL;
      p^.next^.next^.back := NIL;
      update (q);
      update (q^.back);
    END; (* remove *)

    PROCEDURE copy ( VAR a, b : tree);
      VAR i : INTEGER;

      PROCEDURE copynode( VAR c, d : nodeptr);
        VAR necks : nodeptr;
      BEGIN (* copynode *)
        necks := d^.next;
        d^ := c^;
        d^.next := necks;
      END; (* copynode *)

    BEGIN (* copy *)
      FOR i := 1 TO numsp DO
        b.nodep[i]^ := a.nodep[i]^;
      FOR i := numsp1 TO numsp2 DO BEGIN
        copynode (a.nodep[i], b.nodep[i]);
        copynode (a.nodep[i]^.next, b.nodep[i]^.next);
        copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
        END;
      b.likelihood := a.likelihood;
      b.start := a.start;
    END; (* copy *)

    PROCEDURE buildnewtip (m : INTEGER; VAR tr : tree);
      VAR p : nodeptr;
    BEGIN (* buildnewtip *)
      p := tr.nodep[nextsp + numsp - 2];
      hookup (tr.nodep[m], p);
      p^.v := 0.5;
      p^.back^.v := 0.5;
    END; (* buildnewtip *)

    PROCEDURE buildsimpletree (VAR tr : tree);
    BEGIN (* buildsimpletree *)
      hookup (tr.nodep[enterorder[1]], tr.nodep[enterorder[2]]);
      tr.nodep[enterorder[1]]^.v := 0.5;
      tr.nodep[enterorder[2]]^.v := 0.5;
      buildnewtip (enterorder[3], tr);
      insert(tr.nodep[enterorder[3]]^.back, tr.nodep[enterorder[1]]);
    END; (* buildsimpletree *)

    PROCEDURE addtraverse (p, q : nodeptr; contin : BOOLEAN);
    BEGIN (* addtraverse *)
      insert(p, q);
      numtrees := numtrees + 1;
      IF evaluate(curtree) > bestree.likelihood
      THEN copy (curtree, bestree);
      remove(p, q);
      IF (NOT q^.tip) AND contin
      THEN BEGIN
        addtraverse (p, q^.next^.back, contin);
        addtraverse (p, q^.next^.next^.back, contin);
        END;
    END; (* addtraverse *)

    PROCEDURE rearrange (p : nodeptr);
      (* rearranges the tree, globally or locally *)
      VAR q, r : nodeptr;
    BEGIN (* rearrange *)
      IF (NOT p^.tip) AND (NOT p^.back^.tip)
      THEN BEGIN
        r := p^.next^.next;
        remove (r, q);
        addtraverse (r, q^.next^.back,
                   global AND (nextsp = numsp));
        addtraverse (r, q^.next^.next^.back,
                   global AND (nextsp = numsp));
        copy (bestree, curtree);
        IF global AND (nextsp = numsp)
        THEN WRITE(OUTPUT, '.');
        IF global AND (nextsp = numsp) AND (NOT succeeded)
        THEN BEGIN
          IF r^.back^.tip
          THEN BEGIN
            r := r^.next^.next;
            remove (r, q);
            q := q^.back;
            IF NOT q^.tip
            THEN BEGIN
              addtraverse (r, q^.next^.back, TRUE);
              addtraverse (r, q^.next^.next^.back, TRUE);
              END;
            q := q^.back;
            IF NOT q^.tip
            THEN BEGIN
              addtraverse (r, q^.next^.back, TRUE);
              addtraverse (r, q^.next^.next^.back, TRUE);
              END;
            copy (bestree, curtree);
            END;
          END;
        END;
      IF NOT p^.tip
      THEN BEGIN
        rearrange (p^.next^.back);
        rearrange (p^.next^.next^.back);
        END;
    END; (* rearrange *)

    PROCEDURE printree;
      (* prints out diagram of the tree *)
      CONST down = 2;
        over = 60;
      VAR tipy, i : INTEGER;
         scale, tipmax : REAL;

      PROCEDURE coordinates (p : nodeptr; lengthsum : REAL);
        (* establishes coordinates of nodes *)
        VAR q, first, last: nodeptr;         
      BEGIN (* coordinates *)
        IF p^.tip
        THEN BEGIN
          p^.xcoord := TRUNC(over*lengthsum+0.5);
          p^.ycoord := tipy;
          p^.ymin := tipy;
          p^.ymax := tipy;
          tipy := tipy + down;
          IF lengthsum > tipmax
          THEN tipmax := lengthsum;
          END
        ELSE BEGIN
          q := p^.next; 
          REPEAT
            coordinates (q^.back, lengthsum-fracchange*LN(1.0-q^.v));
            q := q^.next;
          UNTIL ((p <> curtree.start^.back) AND (p = q))
            OR ((p = curtree.start^.back) AND (p^.next = q));
          first := p^.next^.back; 
          q := p;
          WHILE q^.next <> p DO 
            q := q^.next;
          last := q^.back;
          p^.xcoord := TRUNC(over*lengthsum + 0.5);
          p^.ycoord := (first^.ycoord + last^.ycoord) DIV 2;
          p^.ymin := first^.ymin;
          p^.ymax := last^.ymax;
          END;
      END; (* coordinates *)

      PROCEDURE drawline (i : INTEGER);
        (* draws one row of the tree diagram by moving up tree *)
        VAR p, q : nodeptr;
          n, j : INTEGER;
          extra : BOOLEAN;
          r, first, last : nodeptr;
          done : BOOLEAN;
      BEGIN (* drawline *)
        p := curtree.start^.back;
        q := curtree.start^.back;
        IF (i = p^.ycoord) AND (p = curtree.start^.back)
        THEN WRITE(OUTPUT, '-')
        ELSE WRITE(OUTPUT, ' ');
        extra := FALSE;
        REPEAT
          IF NOT p^.tip
          THEN BEGIN
            r := p^.next;
            done := FALSE;
            REPEAT
              IF (i >= r^.back^.ymin) AND (i <= r^.back^.ymax)
              THEN BEGIN
                q := r^.back;
                done := TRUE;
                END;
              r := r^.next;
            UNTIL done OR (((p <> curtree.start^.back) AND (r = p))
              OR ((p = curtree.start^.back) AND (r = p^.next)));
            first := p^.next^.back; 
            r := p;
            WHILE r^.next <> p DO 
              r := r^.next;
            last := r^.back;
            IF p = curtree.start^.back THEN last := p^.back;
            END;
          done := (p^.tip) OR (p = q);
          n := TRUNC(scale*(q^.xcoord - p^.xcoord)+0.5);
          IF (n < 2) AND (NOT q^.tip) THEN n := 2;
          IF extra
          THEN BEGIN
            n := n - 1;
            extra := FALSE;
            END;
          IF (q^.ycoord = i) AND (NOT done)
          THEN BEGIN
            WRITE(OUTPUT, '+');
            IF NOT (q^.tip)
            THEN BEGIN
              FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
              IF q^.number-numsp >= 10
              THEN WRITE(OUTPUT, q^.number-numsp:2)
              ELSE WRITE(OUTPUT, '-', q^.number-numsp:1);
              extra := TRUE;
              END
            ELSE FOR j := 1 TO n-1 DO WRITE(OUTPUT, '-');
            END
          ELSE IF (NOT p^.tip)
            THEN BEGIN
              IF (i = p^.ycoord) AND (p = curtree.start^.back)
              THEN BEGIN
                WRITE(OUTPUT, p^.number-numsp:1);
                n := n - 1;
                END;
              IF (last^.ycoord > i) AND (first^.ycoord < i)
                AND ((NOT (i = p^.ycoord)) OR (p = curtree.start))
              THEN BEGIN
                WRITE(OUTPUT, '!');
                FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                END
              ELSE FOR j := 1 TO n DO WRITE(OUTPUT, ' ');
              END;
          IF q <> p THEN p := q;
        UNTIL done;
        IF (p^.ycoord = i) AND (p^.tip)
        THEN FOR j := 1 TO nmlngth DO WRITE(OUTPUT, p^.nayme[j]);
        WRITELN(OUTPUT);
      END; (* drawline *)

    BEGIN (* printree *)
      WRITELN(OUTPUT);
      tipy := 1;
      tipmax := 0.0;
      coordinates (curtree.start^.back, 0.0);
      scale := 1.0/TRUNC(tipmax + 0.999);
      FOR i := 1 TO tipy-down DO drawline (i);
      WRITELN(OUTPUT);
    END; (* printree *)

    PROCEDURE summarize;
       (* print out branch length information and node numbers *)

      PROCEDURE describe (p : nodeptr);
        (* print out information for one branch *)
        VAR i : INTEGER;
          q : nodeptr;
          s, sumlr : REAL;

        FUNCTION sigma(p : nodeptr) : REAL;
           (* compute standard deviation *)
          VAR sum, y, z, z1, zz, lz, qra1, qra2, qro1, qro2, rat, suma,
            sumb, sumc, sumc1, sumc2, slope, d2, d, li, ab, bc : REAL;
            i : INTEGER;
            q : nodeptr;
            xx1, xx2 : sitelike;
        BEGIN (* sigma *)
          sum := 0.0;
          sumlr := 0.0;
          slope := 0.0;
          q := p^.back;
          y := p^.v;
          z := 1.0 - y;
          IF z > 0.0
          THEN lz := LN(z);
          FOR i := 1 TO endsite DO BEGIN
            xx1 := p^.x[i];
            xx2 := q^.x[i];
            rat := rate[category[alias[i]]];
            IF z > 0.0
            THEN BEGIN
              zz := EXP(rat*xi*lz);
              z1 := EXP(rat*xv*lz);
              qra1 := EXP((rat*xv-1.0)*lz);
              qro1 := EXP((rat-1.0)*lz);
              qra2 := EXP((rat*xv-2.0)*lz);
              qro2 := EXP((rat-2.0)*lz);
              END
            ELSE BEGIN
              zz := 0.0;
              z1 := 0.0;
              qra1 := 0.0;
              qro1 := 0.0;
              qra2 := 0.0;
              qro2 := 0.0;
              END;
            sumc1 := freqa*xx1[A] + freqc*xx1[C] + freqg*xx1[G] + freqt*xx1[T];
            sumc2 := freqa*xx2[A] + freqc*xx2[C] + freqg*xx2[G] + freqt*xx2[T];
            sumc := sumc1*sumc2;
            sumb := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar+xx2[G]*freqgr)
                  + (xx1[C]*freqc+xx1[T]*freqt)*(xx2[C]*freqcy+xx2[T]*freqty);
            suma := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                  + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
            ab := suma - sumb;
            bc := sumb - sumc;
            li := sumc + z1*(zz*ab + bc);
            IF z > 0.0
            THEN BEGIN
              d := rat*qro1*ab + rat*xv*qra1*bc;
              d2 := rat*(rat-1.0)*qro2*ab + rat*xv*(rat*xv-1.0)*qra2*bc;
              END
            ELSE BEGIN
              d := 0.0;
              d2 := 0.0;
              END;
            slope := slope + weight[i] * d/li;
            sum := sum + weight[i] * (SQR(d/li) - d2/li);
            IF suma > 0.0
            THEN sumlr := sumlr + weight[i]*LN(li/suma)
            ELSE sumlr := 1.0E+10;
            END;
          IF sum > 1.0E-6
          THEN sigma := (-slope + SQRT(SQR(slope)+3.841*sum))/sum
          ELSE sigma := -1.0;
        END; (* sigma *)

      BEGIN (* describe *)
        q := p^.back;
        WRITE(OUTPUT, (q^.number - numsp):4, '          ');
        IF p^.tip
        THEN FOR i := 1 TO nmlngth DO WRITE(OUTPUT, p^.nayme[i])
        ELSE WRITE(OUTPUT, (p^.number - numsp):4, '      ');
        IF q^.v >= 1.0
        THEN WRITE(OUTPUT, '    infinity')
        ELSE WRITE(OUTPUT, -LN(1-q^.v)*fracchange:15:5);
        s := sigma(q);
        IF s < 0.0
        THEN WRITE(OUTPUT, '     (     zero,    infinity)')
        ELSE BEGIN
          WRITE(OUTPUT, '     (');
          IF q^.v - s <= 0.0
          THEN WRITE(OUTPUT, '     zero')
          ELSE WRITE(OUTPUT, -LN(1 - q^.v + s)*fracchange:9:5);
          WRITE(OUTPUT, ',');
          IF q^.v + s >= 1.0
          THEN WRITE(OUTPUT, '    infinity')
          ELSE WRITE(OUTPUT, -LN(1 - q^.v - s)*fracchange:12:5);
          WRITE(OUTPUT, ')')
          END;
        IF sumlr > 1.9205
        THEN WRITE(OUTPUT, ' *');
        IF sumlr > 2.995
        THEN WRITE(OUTPUT, '*');
        WRITELN(OUTPUT);
        IF NOT p^.tip
        THEN BEGIN
          describe (p^.next^.back);
          describe (p^.next^.next^.back);
          END;
      END; (* describe *)

    BEGIN (* summarize *)
      WRITELN(OUTPUT);
      WRITE(OUTPUT, 'remember: ');
      IF outgropt
      THEN WRITE(OUTPUT, '(although rooted by outgroup) ');
      WRITELN(OUTPUT, 'this is an unrooted tree!');
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, 'Ln Likelihood = ', curtree.likelihood:11:5);
      IF NOT usertree
      THEN BEGIN
        WRITELN(OUTPUT);
        WRITELN(OUTPUT, 'Examined ',numtrees:4, ' trees');
        END;
      WRITELN(OUTPUT);
      WRITE(OUTPUT, ' Between        And            Length');
      WRITELN(OUTPUT, '      Approx. Confidence Limits');
      WRITE(OUTPUT, ' -------        ---            ------');
      WRITELN(OUTPUT, '      ------- ---------- ------');
      WRITELN(OUTPUT);
      describe (curtree.start^.back^.next^.back);
      describe (curtree.start^.back^.next^.next^.back);
      describe (curtree.start);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT, '     *  = significantly positive, P < 0.05');
      WRITELN(OUTPUT, '     ** = significantly positive, P < 0.01');
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
    END; (* summarize *)

    PROCEDURE treeout (p : nodeptr);
      (* write out file with representation of final tree *)
      VAR i, n, w : INTEGER;
        c : CHAR;
        x : REAL;
    BEGIN (* treeout *)
      IF p^.tip
      THEN BEGIN
        n := 0;
        FOR i := 1 TO nmlngth DO
          IF p^.nayme[i] <> ' '
          THEN n := i;
        FOR i := 1 TO n DO BEGIN
          c := p^.nayme[i];
          IF c = ' '
          THEN c := '_';
          WRITE(treefile, c);
          END;
        END
      ELSE BEGIN
        WRITE(treefile, '(');
        treeout (p^.next^.back);
        WRITE(treefile, ',');
        treeout (p^.next^.next^.back);
        IF p = curtree.start^.back
        THEN BEGIN
          WRITE(treefile, ',');
          treeout (p^.back);
          END;
        WRITE(treefile, ')');
        END;
      IF p^.v >= 1.0
      THEN x := -1.0
      ELSE x := -LN(1-p^.v)*fracchange;
      IF x > 0.0
      THEN w := TRUNC(0.4343*LN(x))
      ELSE IF x = 0.0
        THEN w := 0
        ELSE w := TRUNC(0.4343*LN(-x)) + 1;
      IF w < 0 THEN w := 0;
      IF p = curtree.start^.back
      THEN WRITELN(treefile, ';')
      ELSE WRITE(treefile, ':', x:w+7:5);
    END; (* treeout *)

    PROCEDURE treeread;
      VAR ch : CHAR;
        p : nodeptr;
        i, nextnode : INTEGER;

      PROCEDURE findch (c : CHAR);
      BEGIN (* findch *)
        IF ch <> c
        THEN REPEAT
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
          UNTIL ch = c;
      END; (* findch *)

      PROCEDURE getch (VAR c : CHAR);
        (* get next nonblank character *)
      BEGIN (* getch *)
        REPEAT
          IF EOLN(infile) THEN READLN(infile);
          READ(infile, c);
        UNTIL c <> ' ';
      END; (* getch *)

      PROCEDURE addelement (p : nodeptr);
        VAR q : nodeptr;
          i, n : INTEGER;
          found : BOOLEAN;
          str : ARRAY[1..nmlngth] OF CHAR;

        PROCEDURE processlength(p : nodeptr);
          CONST point = '.';
          VAR digit, ordzero : INTEGER;
            valyew, divisor : REAL;
            pointread : BOOLEAN;
        BEGIN (* processlength *)
          ordzero := ORD('0');
          pointread := FALSE;
          valyew := 0.0;
          divisor := 1.0;
          getch(ch);
          digit := ORD(ch)-ordzero;
          WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point) DO BEGIN
            IF ch = point
            THEN pointread := TRUE
            ELSE BEGIN
              valyew := valyew*10.0 + digit;
              IF pointread
              THEN divisor := divisor*10.0;
              END;
            getch(ch);
            digit := ORD(ch)-ordzero;
            END;
          IF lengths
          THEN BEGIN
            p^.v := 1.0-EXP((-valyew/divisor)/fracchange);
            p^.back^.v := p^.v;
            END;
        END; (* processlength *)

      BEGIN (* addelement *)
        getch (ch);
        IF ch = '('
        THEN BEGIN
          nextnode := nextnode + 1;
          q := curtree.nodep[nextnode];
          hookup(p, q);
          addelement (q^.next);
          findch (',');
          addelement (q^.next^.next);
          findch (')');
          getch (ch);
          END
        ELSE BEGIN
          FOR i := 1 TO nmlngth DO str[i] := ' ';
          n := 1;
          REPEAT
            IF (ch = '_') THEN ch := ' ';
            str[n] := ch;
            IF EOLN(infile) THEN READLN(infile);
            READ(infile, ch);
            n := n + 1;
          UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')') OR (n > nmlngth));
          n := 1;
          REPEAT
            found := TRUE;
            FOR i := 1 TO nmlngth DO
              found := found AND (str[i] = curtree.nodep[n]^.nayme[i]);
            IF NOT found
            THEN n := n + 1;
          UNTIL (n > numsp) OR found;
          IF n > numsp
          THEN BEGIN
            WRITE(OUTPUT, 'Cannot find species: ');
            FOR i := 1 TO nmlngth DO WRITE(OUTPUT, str[i]);
            WRITELN(OUTPUT);
            END;
          hookup (curtree.nodep[n], p);
          IF curtree.start^.number > n
          THEN curtree.start := curtree.nodep[n];
          END;
        IF ch = ':'
        THEN processlength (p);
      END; (* addelement *)

    BEGIN (* treeread *)
      curtree.start := curtree.nodep[numsp];
      getch (ch);
      IF ch = '('
      THEN BEGIN
        nextnode := numsp + 1;
        p := curtree.nodep[nextnode];
        FOR i := 1 TO 2 DO BEGIN
          addelement (p);
          p := p^.next;
          findch (',');
          END;
        addelement (p);
        END;
      READLN(infile);
    END; (* treeread *)

    PROCEDURE treevaluate;
      VAR i : INTEGER;
        dummy : REAL;

      PROCEDURE initrav (p : nodeptr);

        PROCEDURE nodeinit (p : nodeptr);
          VAR q, r : nodeptr;
            i : INTEGER;
            b : base;
        BEGIN (* nodeinit *)
          IF NOT p^.tip
          THEN BEGIN
            q := p^.next^.back;
            r := p^.next^.next^.back;
            nodeinit (q);
            nodeinit (r);
            FOR i := 1 TO endsite DO
              FOR b := A TO T DO
                p^.x[i][b] := 0.5 * (q^.x[i][b] + r^.x[i][b]);
            p^.v := 0.1;
            p^.back^.v := 0.1;
            END;
        END; (* nodeinit *)

      BEGIN (* initrav *)
        IF p^.tip
        THEN nodeinit (p^.back)
        ELSE BEGIN
          initrav (p^.next^.back);
          initrav (p^.next^.next^.back);
          END;
      END; (* initrav *)

    BEGIN (* treevaluate *)
      IF NOT lengths
      THEN BEGIN
        initrav (curtree.start);
        initrav (curtree.start^.back);
        END;
      FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
      dummy := evaluate (curtree);
    END; (* treevaluate *)

  BEGIN (* maketree *)
    IF usertree
    THEN BEGIN
      READLN(infile, numtrees);
      WRITE(OUTPUT, 'User-defined tree');
      IF numtrees > 1
      THEN WRITE(OUTPUT, 's');
      WRITELN(OUTPUT, ':');
      WRITELN(OUTPUT);
      FOR which := 1 TO numtrees DO BEGIN
        treeread;
        treevaluate;
        printree;
        summarize;
        IF trout
        THEN treeout (curtree.start^.back);
        END;
      WRITELN(OUTPUT);
      IF (numtrees > 1) AND (weightsum > 1)
      THEN BEGIN
        WRITE(OUTPUT, 'Tree    Ln L      Diff Ln L     Its S.D.');
        WRITELN(OUTPUT, '   Significantly worse?');
        WRITELN(OUTPUT);
        IF numtrees > maxtrees THEN num := maxtrees ELSE num := numtrees;
        FOR i := 1 TO num DO BEGIN
          WRITE(OUTPUT, i:3, logl[i]:12:5);
          IF maxwhich = i
          THEN WRITELN(OUTPUT, '  <------ best')
          ELSE BEGIN
            sum := 0.0;
            sum2 := 0.0;
            FOR j := 1 TO endsite DO BEGIN
              sum := sum + weight[j]*(logf[maxwhich][j]-logf[i][j]);
              sum2 := sum2 + weight[j]*SQR(logf[maxwhich][j]-logf[i][j]);
              END;
            sd := SQRT((weightsum/(weightsum-1))
                   *(sum2-SQR(sum/weightsum)));
            WRITE(OUTPUT, (logl[i]-maxlogl):12:5, sd:12:4);
            IF sum > 1.95996*sd
            THEN WRITELN(OUTPUT, '           Yes')
            ELSE WRITELN(OUTPUT, '            No');
            END;
          END;
        WRITELN(OUTPUT);
        WRITELN(OUTPUT);
        END;
      END
    ELSE BEGIN
      FOR i := 1 TO numsp DO enterorder[i] := i;
      IF jumble
      THEN FOR i := 1 TO numsp DO BEGIN
        j := TRUNC(randum(seed)*numsp) + 1;
        k := enterorder[j];
        enterorder[j] := enterorder[i];
        enterorder[i] := k;
        END;
      WRITELN(OUTPUT, 'Adding species:');
      WRITE(OUTPUT, '   ');
      FOR i := 1 TO nmlngth DO
        WRITE(OUTPUT, curtree.nodep[enterorder[1]]^.nayme[i]);
      WRITELN(OUTPUT);
      WRITE(OUTPUT, '   ');
      FOR i := 1 TO nmlngth DO
        WRITE(OUTPUT, curtree.nodep[enterorder[2]]^.nayme[i]);
      WRITELN(OUTPUT);
      WRITE(OUTPUT, '   ');
      FOR i := 1 TO nmlngth DO
        WRITE(OUTPUT, curtree.nodep[enterorder[3]]^.nayme[i]);
      WRITELN(OUTPUT);
      nextsp := 3;
      buildsimpletree (curtree);
      curtree.start := curtree.nodep[enterorder[1]];
      numtrees := 1;
      nextsp := 4;
      WHILE nextsp <= numsp DO BEGIN
        buildnewtip (enterorder[nextsp], curtree);
        bestree.likelihood := -999999.0;
        addtraverse (curtree.nodep[enterorder[nextsp]]^.back,
                        curtree.start^.back, TRUE);
        copy (bestree, curtree);
        WRITE(OUTPUT, '   ');
        FOR j := 1 TO nmlngth DO
          WRITE(OUTPUT,
            curtree.nodep[enterorder[nextsp]]^.nayme[j]);
        WRITELN(OUTPUT);
        IF global AND (nextsp = numsp)
        THEN BEGIN
          WRITELN(OUTPUT, 'Doing global rearrangements');
          WRITE(OUTPUT, '   ');
          END;
        succeeded := TRUE;
        WHILE succeeded DO BEGIN
          succeeded := FALSE;
          rearrange (curtree.start^.back);
          END;
        IF (nextsp = numsp)
        THEN BEGIN
          curtree.start := curtree.nodep[outgrno];
          printree;
          summarize;
          IF trout AND (nextsp = numsp)
          THEN BEGIN
            treeout (curtree.start^.back);
            WRITELN(OUTPUT, 'Tree also written onto file');
            WRITELN(OUTPUT);
            END;
          END;
        nextsp := nextsp + 1;
        END;
      END;
  END; (* maketree *)

BEGIN (* DNA Maximum Likelihood *)
  (* ASSIGN(infile, 'INFILE');   Turbo Pascal only *)
  RESET(infile);
  getinput;
  IF trout
  THEN BEGIN
    (* ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
    REWRITE(treefile);
    END;
  IF NOT anerror
  THEN maketree;
(* CLOSE(infile);  -- may need this here *)
  IF trout
  THEN BEGIN
    (* CLOSE(treefile);   Turbo Pascal and maybe others too *)
    END;
END. (* DNA Maximum Likelihood *)
