PROGRAM Drawgram (INPUT, OUTPUT, treefile, fontfile, plotfile);
  (* USES CRT, GRAPH;  added for Turbo Pascal versions 4, 5 *)
  (* Version 3.3.  Copyright (c) 1986, 1990 by Joseph Felsenstein and
    Christopher A. Meacham.
    Permission is granted to copy, distribute, and modify this
    program provided that (1) this copyright message is not removed
    and (2) no fee is charged for this program. *)

  CONST maxnodes = 300;
    maxnch = 30;
    point = '.';
    stripedepth = 4;
    stripewidth = 1350;
    fontsize = 3800;
    pi = 3.141592653;
    epsilon = 0.00001;
    ebcdic = FALSE;

  TYPE plottertype = (lw, hp, tek, ibmpc, houston,
      decregis, epson, oki, citoh, toshiba, other);
    growth = (vertical, horizontal);
    treestyle = (cladogram, phenogram, curvogram, eurogram, swoopogram);
    penstatustype = (penup, pendown);
    plotstring = ARRAY[1..maxnch] OF CHAR;
    fonttype = ARRAY[1..fontsize] OF INTEGER;
    nodeptr = ^node;
    node = RECORD
      next, back : nodeptr;
      tip : BOOLEAN;
      nayme : plotstring;
      naymlength, tipsabove : INTEGER;
      xcoord, ycoord, level : REAL
      END;

  VAR treefile, fontfile, plotfile : TEXT;
    ntips, nextnode, numtochange, strpwide, strpdeep, strpdiv,
      strptop, strpbottom, oldx, oldy, GraphDriver, GraphMode,
      LoMode, HiMode, nmoves, payge, numlines : INTEGER;
    parmsready, canbeplotted, preview, previewing, dotmatrix,
      haslengths, uselengths, empty : BOOLEAN;
    xmargin, ymargin, topoflabels, rightoflabels, leftoflabels, tipspacing,
      scale, xscale, yscale, xoffset, yoffset, nodespace, stemlength,
      treedepth, xnow, ynow, xunitspercm, yunitspercm, xsize, ysize, xcorner,
      ycorner, oldxhigh, oldyhigh, oldylow, oldxlow, treeline, labelline,
      labelheight, labelrotation, oldxsize, oldysize, oldxunitspercm,
      oldyunitspercm, oldxcorner, oldycorner, expand, rooty : REAL;
    escape : CHAR;     (* DLS/JMH -- used in new drivers *)
    stripe : ARRAY[1..stripedepth] OF ARRAY[1..stripewidth] OF CHAR;
    bit : ARRAY[0..7] OF INTEGER;
    eb : ARRAY[0..127] OF INTEGER;  (*  DLS/JMH -- for EBCDIC Tek driver *)
    plotter, oldplotter, previewer : plottertype;
    grows : growth;
    style : treestyle;
    root : nodeptr;
    nodep : ARRAY[1..maxnodes] OF nodeptr;
    font : fonttype;
    penchange, oldpenchange : (yes, no);
    nodeposition : (weighted, intermediate, centered, inner, vshaped);

(*    PROCEDURE CGADriver; EXTERNAL;     Turbo Pascal only *)
      {$L cgadrv.obj }                   (* Turbo Pascal only *)        
(*    PROCEDURE EGAVGADriver; EXTERNAL;  Turbo Pascal only *)
      {$L egavgadrv.obj }                (* Turbo Pascal only *)
(*    PROCEDURE HERCDriver; EXTERNAL;    Turbo Pascal only *)
      {$L hercdrv.obj }                  (* Turbo Pascal only *)
(*    PROCEDURE ATTDriver; EXTERNAL;     Turbo Pascal only *)
      {$L attdrv.obj}                    (* Turbo Pascal only *)

  PROCEDURE setup;
    (* sets up some particular arrays *)
  BEGIN (* setup *)
    IF ebcdic
    THEN escape := CHR(39)   (* DLS/JMH *)
    ELSE escape := CHR(27);
    IF ebcdic
    THEN BEGIN
  eb[  0]:=  0; eb[  1]:=  1; eb[  2]:=  2; eb[  3]:=  3; eb[  4]:= 55;
  eb[  5]:= 45; eb[  6]:= 46; eb[  7]:= 47; eb[  8]:= 22; eb[  9]:=  5;
  eb[ 10]:= 37; eb[ 11]:= 11; eb[ 12]:= 12; eb[ 13]:= 13; eb[ 14]:= 14;
  eb[ 15]:= 15; eb[ 16]:= 16; eb[ 17]:= 17; eb[ 18]:= 18; eb[ 19]:= 19;
  eb[ 20]:= 60; eb[ 21]:= 61; eb[ 22]:= 50; eb[ 23]:= 38; eb[ 24]:= 24;
  eb[ 25]:= 25; eb[ 26]:= 63; eb[ 27]:= 39; eb[ 28]:= 28; eb[ 29]:= 29;
  eb[ 30]:= 30; eb[ 31]:= 31; eb[ 32]:= 64; eb[ 33]:= 90; eb[ 34]:=127;
  eb[ 35]:=123; eb[ 36]:= 91; eb[ 37]:=108; eb[ 38]:= 80; eb[ 39]:=125;
  eb[ 40]:= 77; eb[ 41]:= 93; eb[ 42]:= 92; eb[ 43]:= 78; eb[ 44]:=107;
  eb[ 45]:= 96; eb[ 46]:= 75; eb[ 47]:= 97; eb[ 48]:=240; eb[ 49]:=241;
  eb[ 50]:=242; eb[ 51]:=243; eb[ 52]:=244; eb[ 53]:=245; eb[ 54]:=246;
  eb[ 55]:=247; eb[ 56]:=248; eb[ 57]:=249; eb[ 58]:=122; eb[ 59]:= 94;
  eb[ 60]:= 76; eb[ 61]:=126; eb[ 62]:=110; eb[ 63]:=111; eb[ 64]:=124;
  eb[ 65]:=193; eb[ 66]:=194; eb[ 67]:=195; eb[ 68]:=196; eb[ 69]:=197;
  eb[ 70]:=198; eb[ 71]:=199; eb[ 72]:=200; eb[ 73]:=201; eb[ 74]:=209;
  eb[ 75]:=210; eb[ 76]:=211; eb[ 77]:=212; eb[ 78]:=213; eb[ 79]:=214;
  eb[ 80]:=215; eb[ 81]:=216; eb[ 82]:=217; eb[ 83]:=226; eb[ 84]:=227;
  eb[ 85]:=228; eb[ 86]:=229; eb[ 87]:=230; eb[ 88]:=231; eb[ 89]:=232;
  eb[ 90]:=233; eb[ 91]:=173; eb[ 92]:=224; eb[ 93]:=189; eb[ 94]:= 95;
  eb[ 95]:=109; eb[ 96]:=121; eb[ 97]:=129; eb[ 98]:=130; eb[ 99]:=131;
  eb[100]:=132; eb[101]:=133; eb[102]:=134; eb[103]:=135; eb[104]:=136;
  eb[105]:=137; eb[106]:=145; eb[107]:=146; eb[108]:=147; eb[109]:=148;
  eb[110]:=149; eb[111]:=150; eb[112]:=151; eb[113]:=152; eb[114]:=153;
  eb[115]:=162; eb[116]:=163; eb[117]:=164; eb[118]:=165; eb[119]:=166;
  eb[120]:=167; eb[121]:=168; eb[122]:=169; eb[123]:=192; eb[124]:= 79;
  eb[125]:=208; eb[126]:=161; eb[127]:=  7;
      END;
  END; (* setup *)

  PROCEDURE uppercase (VAR ch : CHAR);
    (* make ch upper-case *)

    FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* TRUE if ch is a letter -- works for both ASCII and EBCDIC codes *)
    BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I')) OR ((ch >= 'J') AND (ch <= 'R'))
            OR ((ch >= 'S') AND (ch <= 'Z')) OR ((ch >= 'a') AND (ch <= 'i'))
            OR ((ch >= 'j') AND (ch <= 'r')) OR ((ch >= 's') AND (ch <= 'z'));
    END; (* letter *)

  BEGIN (* uppercase *)
    IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                OR (('A' > 'a') AND (ch < 'A')))
    THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
  END; (* uppercase *)

  PROCEDURE treeread;
    (* read a tree from the treefile and set up nodes and pointers *)
    VAR ch : CHAR;

    PROCEDURE getch (VAR c : CHAR);
      (* get next nonblank character *)
    BEGIN (* getch *)
      REPEAT
        IF EOLN(treefile) THEN READLN(treefile);
        READ(treefile, c);
      UNTIL c <> ' ';
    END; (* getch *)

    PROCEDURE addelement (VAR p : nodeptr; q : nodeptr);
      (* read in and add next part of tree, it will be node p
        and will be hooked to pointer q *)
      VAR pfirst : nodeptr;
        n : INTEGER;
        notlast : BOOLEAN;

      PROCEDURE processlength(p : nodeptr);
        VAR digit, ordzero : INTEGER;
          valyew, divisor : REAL;
          pointread : BOOLEAN;
      BEGIN (* processlength *)
        ordzero := ORD('0');
        pointread := FALSE;
        valyew := 0.0;
        divisor := 1.0;
        getch(ch);
        digit := ORD(ch)-ordzero;
        WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point) DO BEGIN
          IF ch = point
          THEN pointread := TRUE
          ELSE BEGIN
            valyew := valyew*10.0 + digit;
            IF pointread
            THEN divisor := divisor*10.0;
            END;
          getch(ch);
          digit := ORD(ch)-ordzero;
          END;
        p^.level := valyew/divisor;
      END; (* processlength *)

    BEGIN (* addelement *)
      nextnode := nextnode + 1;
      NEW(p);
      nodep[nextnode] := p;
      IF ch = '('
      THEN BEGIN
        p^.tip := FALSE;
        p^.tipsabove := 0;
        pfirst := p;
        notlast := TRUE;
        WHILE notlast DO BEGIN
          NEW(p^.next);
          p := p^.next;
          p^.tip := FALSE;
          getch (ch);
          addelement (p^.back, p);
          pfirst^.tipsabove := pfirst^.tipsabove + p^.back^.tipsabove;
          IF ch = ')'
          THEN BEGIN
            notlast := FALSE;
            REPEAT
              getch (ch);
            UNTIL (ch = ':') OR (ch = ',') OR (ch = ')') OR (ch = ';');
            END;
          END;
        p^.next := pfirst;
        p := pfirst;
        END
      ELSE BEGIN
        p^.tip := TRUE;
        p^.tipsabove := 1;
        ntips := ntips + 1;
        n := 1;
        REPEAT
          IF (ch = '_') THEN ch := ' ';
          IF n < maxnch
          THEN p^.nayme[n] := ch;
          IF EOLN(treefile) THEN READLN(treefile);
          READ(treefile, ch);
          n := n + 1;
        UNTIL ((ch = ':') OR (ch = ',') OR (ch = ')'));
        IF n > maxnch
        THEN n := maxnch + 1;
        p^.naymlength := n - 1;
        END;
      IF ch = ':'
      THEN processlength (p)
      ELSE haslengths := haslengths AND (q = NIL);
      p^.back := q;
    END; (* addelement *)

  BEGIN (* treeread *)
    haslengths := TRUE;
    ntips := 0;
    nextnode := 0;
    getch (ch);
    addelement (root, NIL);
    READLN(treefile);
    uselengths := haslengths;
  END; (* treeread *)

  PROCEDURE loadfont(VAR font : fonttype);
    VAR i, charstart, dummy : INTEGER;
      ch : CHAR;
  BEGIN (* loadfont *)
    ch := 'A';
    i := 0;
    WHILE NOT(EOF(fontfile) OR (ch = ' ')) DO BEGIN
      charstart := i+1;
      READ(fontfile, ch, ch, dummy, font[charstart+2], font[charstart+3]);
      font[charstart+1] := ORD(ch);
      i := charstart+3;
      REPEAT
        IF ((i-(charstart+3)) MOD 10) =0 THEN READLN(fontfile);
        i := i+1;
        READ(fontfile,font[i]);
        UNTIL ABS(font[i]) >= 10000;
      READLN(fontfile);
      font[charstart] := i+1;
      END;
    font[charstart] := 0;
  END; (* loadfont *)

  PROCEDURE plotrparms;
    (* set up initial characteristics of plotter or printer *)
  BEGIN (* plotrparms *)
    penchange := no;
    xcorner := 0.0;
    ycorner := 0.0;
    IF NOT (previewing)
    THEN BEGIN
      strpdiv := 1;
      strpwide := 1;
      END;
    CASE plotter OF
      lw : BEGIN
        penchange := yes;
        xunitspercm := 28.346456693;
        yunitspercm := 28.346456693;
        xsize := 21.59;
        ysize := 27.94;
        END;
      hp : BEGIN
        penchange := yes;
        xunitspercm := 400.0;
        yunitspercm := 400.0;
        xsize := 24.0;
        ysize := 18.0;
        END;
      tek : BEGIN
        xunitspercm := 50.0;
        yunitspercm := 50.0;
        xsize := 20.46;
        ysize := 15.6;
        END;
      ibmpc : BEGIN
      (* for most IBM PC monitors using TURBO 4.0 BGI interface
      Written by Buz Wilson, 6 November, 1988
      based on a 25 cm x 17.5 cm physical screen size *)
(* Comment out Turbo specific code
        GraphDriver := 0;
        DetectGraph(GraphDriver, GraphMode);
        GetModeRange(GraphDriver,LoMode,HiMode);
        InitGraph(GraphDriver,HiMode,'');
        xunitspercm := GetMaxX/25;
        yunitspercm := GetMaxY/17.5;
        RestoreCRTMode;
        xsize := 25.0;
        ysize := 17.5;
end commenting out of Turbo specific code *)
        END;
      houston : BEGIN
        penchange := yes;
        xunitspercm := 100.0;
        yunitspercm := 100.0;
        xsize := 24.5;
        ysize := 17.5;
        END;
      decregis : BEGIN
        xunitspercm := 30.0;
        yunitspercm := 30.0;
        xsize := 25.0;
        ysize := 15.0;
        END;
      epson: BEGIN
        xunitspercm := 47.244;
        yunitspercm := 28.346;
        xsize := 18.70;
        ysize := 22.0;
        strpwide := 960;
        strpdeep := 8;
        strpdiv := 1;
        END;
      oki : BEGIN
        xunitspercm := 56.692;
        yunitspercm := 28.346;
        xsize := 19.0;
        ysize := 22.0;
        strpwide := 1100;
        strpdeep := 8;
        strpdiv := 1;
        END;
      citoh : BEGIN
        xunitspercm := 28.346;
        yunitspercm := 28.346;
        xsize := 22.3;
        ysize := 26.0;
        strpwide := 640;
        strpdeep := 8;
        strpdiv := 1;
        END;
      toshiba : BEGIN
        xunitspercm := 70.866;
        yunitspercm := 70.866;
        xsize := 19.0;
        ysize := 25.0;
        strpwide := 1350;
        strpdeep := 24;
        strpdiv := 4;
        END;
      other : BEGIN
        (* initial parameter settings for a new
          plotter go here *)
        END
      END;
    IF NOT previewing
    THEN BEGIN
      IF dotmatrix
      THEN numlines := ROUND(yunitspercm*ysize) DIV strpdeep
      ELSE numlines := 1;
      END;
  END; (* plotrparms *)

  PROCEDURE initialparms;
     (* initialize parameters *)

     PROCEDURE getplotter;
       VAR ch : CHAR;
     BEGIN (* getplotter *)
       WRITELN;
       WRITELN('Which plotter or printer will the tree be drawn on?');
       WRITELN('(many other brands or models are compatible with these)');
       WRITELN;
       WRITELN('   type:       to choose one compatible with:');
       WRITELN;
       WRITELN('        L         Apple Laserwriter (with Postscript)');
       WRITELN('        K         TeKtronix 4010 graphics terminal');
       WRITELN('        H         Hewlett-Packard 7470 plotter');
       WRITELN('        I         IBM PC graphics screens');
       WRITELN('        D         DEC ReGIS graphics (VT240 terminal)');
       WRITELN('        B         Houston Instruments plotter');
       WRITELN('        E         Epson MX-80 dot-matrix printer');
       WRITELN('        C         Prowriter/Imagewriter dot-matrix printer');
       WRITELN('        O         Okidata dot-matrix printer');
       WRITELN('        T         Toshiba 24-pin dot-matrix printer');
       WRITELN('        Y         other: one You have inserted code for');
       REPEAT
         WRITELN(' Choose one: ');
         READLN(ch);
         uppercase (ch);
       UNTIL (ch = 'L') OR (ch = 'K') OR (ch = 'H') OR (ch = 'I')
         OR (ch = 'D') OR (ch = 'B') OR (ch = 'E') OR (ch = 'C')
         OR (ch = 'O') OR (ch = 'T') OR (ch = 'Y');
       CASE ch OF
         'L' : plotter := lw;
         'K' : plotter := tek;
         'H' : plotter := hp;
         'I' : plotter := ibmpc;
         'D' : plotter := decregis;
         'B' : plotter := houston;
         'E' : plotter := epson;
         'C' : plotter := citoh;
         'O' : plotter := oki;
         'T' : plotter := toshiba;
         'Y' : plotter := other
         END;
       dotmatrix := (plotter = epson) OR (plotter = oki)
                   OR (plotter = citoh) OR (plotter = toshiba);
       WRITELN;
       WRITELN('Will the plot be previewed on your computer screen?');
       REPEAT
         WRITELN(' Type Y or N (for Yes or No):');
         READLN(ch);
         uppercase (ch);
       UNTIL (ch = 'Y') OR (ch = 'N');
       preview := (ch = 'Y');
       IF preview
       THEN BEGIN
         WRITELN('Which type of screen will it be previewed on?');
         WRITELN;
         WRITELN('   type:       to choose one compatible with:');
         WRITELN;
         WRITELN('        I         IBM PC graphics screens');
         WRITELN('        K         TeKtronix 4010 graphics terminal');
         WRITELN('        D         DEC ReGIS graphics (VT240 terminal)');
         WRITELN('        Y         other: one You have inserted code for');
         REPEAT
           WRITELN(' Choose one: ');
           READLN(ch);
           uppercase (ch);
         UNTIL (ch = 'I') OR (ch = 'K') OR (ch = 'D') OR (ch = 'Y');
         CASE ch OF
           'I' : previewer := ibmpc;
           'K' : previewer := tek;
           'D' : previewer := decregis;
           'Y' : previewer := other
           END;
         END;
       WRITELN; WRITELN; WRITELN;
     END; (* getplotter *)

  BEGIN (* initialparms *)
    getplotter;
    plotrparms;
    xmargin := 0.08 * xsize;
    ymargin := 0.08 * ysize;
    xscale := xunitspercm;
    yscale := yunitspercm;
    style := cladogram;
    grows := vertical;
    labelrotation := 45.0;
    nodespace := 3.0;
    stemlength := 0.05;
    treedepth := 0.5/0.95;
    IF uselengths
    THEN nodeposition := intermediate
    ELSE nodeposition := vshaped;
  END; (* initialparms *)

  PROCEDURE showparms;
    VAR i : INTEGER;
      ch : CHAR;
  BEGIN (* showparms *)
    FOR i := 1 TO 24 DO WRITELN;
    WRITELN('Here are the settings: ');
    WRITELN;
    WRITE(' (1)               Tree grows:  ');
    IF grows = vertical
    THEN WRITELN('Vertically')
    ELSE WRITELN('Horizontally');
    WRITE(' (2)            Style of tree:  ');
    CASE style OF
      cladogram : WRITELN('Cladogram');
      phenogram : WRITELN('Phenogram');
      curvogram : WRITELN('Curvogram');
      eurogram :  WRITELN('Eurogram');
      swoopogram : WRITELN('Swoopogram')
      END;
    WRITE(' (3)       Use branch lengths:  ');
    IF haslengths
    THEN BEGIN
      IF uselengths
      THEN WRITELN('Yes')
      ELSE WRITELN('No');
      END
    ELSE WRITELN('(no branch lengths available)');
    WRITE(' (4)          Angle of labels:');
    IF labelrotation < 10.0
    THEN WRITELN(labelrotation:5:1)
    ELSE WRITELN(labelrotation:6:1);
    WRITELN(' (5)       Horizontal margins:', xmargin:6:2, ' cm');
    WRITELN(' (5)         Vertical margins:', ymargin:6:2, ' cm');
    WRITELN(' (6)    Depth/Breadth of tree:', treedepth:6:2);
    WRITELN(' (7)   Stem-length/tree-depth:', stemlength:6:2);
    WRITELN(' (8) Character ht / tip space:', (1.0/nodespace):8:4);
    WRITE(' (9)          Ancestral nodes:');
    CASE nodeposition OF
      weighted     : WRITELN('  Weighted');
      intermediate : WRITELN('  Intermediate');
      centered     : WRITELN('  Centered');
      inner        : WRITELN('  Inner');
      vshaped      : WRITELN('  so tree is V-shaped')
      END;
    WRITELN;
    WRITELN(' Do you want to accept these? (Yes or No)');
    REPEAT
      WRITELN(' Type Y or N or the number (1-9) of the one to change: ');
      READLN(ch);
      uppercase (ch);
    UNTIL (ch = 'Y') OR (ch = 'N') OR ((ch >= '0') AND (ch <= '9'));
    IF NOT ((ch = 'Y') OR (ch = 'N'))
    THEN numtochange := ORD(ch) - ORD('0')
    ELSE numtochange := 0;
    parmsready := (ch = 'Y');
  END; (* showparms *)

  PROCEDURE getparms;
    (* get from user the relevant parameters for the plotter and diagram *)
    VAR ch : CHAR;
      ok : BOOLEAN;
  BEGIN (* getparms *)
    IF numtochange = 0
    THEN REPEAT
      WRITELN(' Type the number of one that you want to change (1-9):');
      READLN(numtochange);
    UNTIL ( numtochange >= 1) AND (numtochange <= 9);
    CASE numtochange OF
      1: IF grows = vertical
        THEN grows := horizontal
        ELSE grows := vertical;
      2: BEGIN
        WRITELN;
        WRITELN('What style tree is this to be:');
        WRITE('   Cladogram, Phenogram, curVogram, Eurogram,');
        WRITELN('  or Swoopogram');
        WRITELN(' (C, P, V, E, or S)');
        REPEAT
          WRITELN(' Choose one: ');
          READLN(ch);
          uppercase (ch);
        UNTIL (ch = 'C') OR (ch = 'P') OR (ch = 'V')
            OR (ch = 'E') OR (ch = 'S');
        CASE ch OF
          'C' : style := cladogram;
          'P' : style := phenogram;
          'E' : style := eurogram;
          'S' : style := swoopogram;
          'V' : style := curvogram
          END;
        END;
      3: BEGIN
        IF haslengths
        THEN BEGIN
           uselengths := NOT uselengths;
           IF NOT uselengths
           THEN nodeposition := vshaped
           ELSE nodeposition := intermediate;
           END
        ELSE BEGIN
          WRITELN('Cannot use lengths since not all of them exist');
          uselengths := FALSE;
          END;
        END;
      4: BEGIN
        WRITELN;
        WRITELN('(Considering the tree as if it "grew" vertically:)');
        WRITELN('Are the labels to be plotted vertically (90),');
        WRITELN(' horizontally (0), or at a 45-degree angle?');
        REPEAT
          WRITELN(' Choose an angle in degrees from 90 to 0:');
          READLN(labelrotation);
          uppercase (ch);
        UNTIL (labelrotation >= 0.0) OR (labelrotation <= 90.0);
        END;
      5: BEGIN
        WRITELN;
        WRITELN('The tree will be drawn to fit in a rectangle which has ');
        WRITELN(' margins in the horizontal and vertical directions of:');
        WRITELN(xmargin:6:2, ' cm (horizontal margin) and',
             ymargin:6:2, ' cm (vertical margin)');
        WRITELN;
        REPEAT
          WRITELN(' New value (in cm) of horizontal margin?');
          READLN(xmargin);
          ok := (xmargin >= 0.0) AND (xmargin < xsize/2.0);
          IF NOT ok
          THEN WRITELN(' Impossible value.  Please retype it.');
        UNTIL ok;
        REPEAT
          WRITELN(' New value (in cm) of vertical margin?');
          READLN(ymargin);
          ok := (ymargin >= 0.0) AND (ymargin < ysize/2.0);
          IF NOT ok
          THEN WRITELN(' Impossible value.  Please retype it.');
        UNTIL ok;
        END;
      6: BEGIN
        WRITELN('New value of depth of tree as fraction of its breadth?');
        READLN(treedepth);
        END;
      7: BEGIN
        REPEAT
          WRITELN('New value of stem length as fraction of tree depth?');
          READLN(stemlength);
        UNTIL (stemlength >= 0.0) AND (stemlength < 0.9);
        END;
      8: BEGIN
        WRITELN('New value of character height as fraction of tip spacing?');
        READLN(nodespace);
        nodespace := 1.0/nodespace;
        END;
      9: BEGIN
        WRITELN('Should interior node positions:');
        WRITELN(' be Intermediate between their immediate descendants,');
        WRITELN('    Weighted average of tip positions');
        WRITELN('    Centered among their ultimate descendants');
        WRITELN('    iNnermost of immediate descendants');
        WRITELN(' or so that tree is V-shaped');
        REPEAT
          WRITELN(' (type I, W, C, N or V):');
          READLN(ch);
          uppercase(ch);
        UNTIL (ch = 'I') OR (ch = 'W')
           OR (ch = 'C') OR (ch = 'N') OR (ch = 'V');
        CASE ch OF
          'W' : nodeposition := weighted;
          'I' : nodeposition := intermediate;
          'C' : nodeposition := centered;
          'N' : nodeposition := inner;
          'V' : nodeposition := vshaped
          END;
        END
      END;
  END; (* getparms *)

  PROCEDURE calculate;
    (* compute coordinates for tree *)
    VAR sum, tipx, maxtextlength, textlength,
      firstlet, maxheight, fontheight, angle : REAL;
      i : INTEGER;

    PROCEDURE calctraverse (p : nodeptr; lengthsum : REAL);
      (* traverse to establish initial node coordinates *)
      VAR x1, y1, x2, y2, x3, w1, w2, sumwx, sumw, nodeheight : REAL;
        pp, plast : nodeptr;
    BEGIN (* calctraverse *)
      IF p = root
      THEN nodeheight := 0.0
      ELSE IF uselengths
        THEN nodeheight := lengthsum + p^.level
        ELSE nodeheight := 1.0;
      IF nodeheight > maxheight THEN maxheight := nodeheight;
      IF p^.tip
      THEN BEGIN
        p^.xcoord := tipx;
        IF uselengths
        THEN p^.ycoord := nodeheight
        ELSE p^.ycoord := 1.0;
        tipx := tipx + tipspacing;
        END
      ELSE BEGIN
        sumwx := 0.0;
        sumw := 0.0;
        pp := p^.next;
        x3 := 0.0;
        REPEAT
          calctraverse (pp^.back, nodeheight);
          sumw := sumw + pp^.back^.tipsabove;
          sumwx := sumwx + pp^.back^.tipsabove*pp^.back^.xcoord;
          IF ABS(pp^.back^.xcoord-0.5) < ABS(x3-0.5)
          THEN x3 := pp^.back^.xcoord;
          plast := pp;
          pp := pp^.next;
        UNTIL pp = p;
        x1 := p^.next^.back^.xcoord;
        x2 := plast^.back^.xcoord;
        y1 := p^.next^.back^.ycoord;
        y2 := plast^.back^.ycoord;
        CASE nodeposition OF
          weighted : BEGIN
            w1 := y1 - nodeheight;
            w2 := y2 - nodeheight;
            IF (w1 + w2) <= 0.0
            THEN p^.xcoord := (x1 + x2)/2.0
            ELSE p^.xcoord := (w2*x1 + w1*x2)/(w1+w2);
            END;
          intermediate : p^.xcoord := (x1 + x2)/2.0;
          centered : p^.xcoord := sumwx/sumw;
          inner : p^.xcoord := x3;
          vshaped : p^.xcoord := (x1 + x2 + (y1 - y2)/maxheight)/2.0
          END;
        IF uselengths
        THEN p^.ycoord := nodeheight
        ELSE BEGIN
          p^.ycoord := (x1 - x2 + y1 + y2)/2.0;
          IF nodeposition = inner
          THEN BEGIN
            IF ABS(x1-0.5) > ABS(x2 - 0.5)
            THEN BEGIN
              p^.ycoord := y1 + x1 - x2;
              w1 := y2 - p^.ycoord;
              END
            ELSE BEGIN
              p^.ycoord := y2 + x1 - x2;
              w1 := y1 - p^.ycoord;
              END;
            IF w1 < epsilon
            THEN p^.ycoord := p^.ycoord - ABS(x1-x2);
            END;
          END;
        END;
    END; (* traverse *)

    FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER): REAL;
      VAR i, j, code : INTEGER;
        sumlength : REAL;
    BEGIN (* lengthext *)
      sumlength := 0.0;
      FOR i := 1 TO nchars DO BEGIN
        code := ORD(pstring[i]);
        j := 1;
        WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
          j := font[j];
        IF font[j+1] = code
        THEN sumlength := sumlength + font[j+3];
        END;
      lengthtext := sumlength;
    END; (* lengthtext *)

  BEGIN (* calculate *)
    maxheight := 0.0;
    maxtextlength := 0.0;
    IF nodep[1]^.naymlength > 0
    THEN firstlet := lengthtext (nodep[1]^.nayme, 1)
    ELSE firstlet := 0.0;
    sum := 0.0;
    tipx := 0.0;
    FOR i := 1 TO nextnode DO
      IF nodep[i]^.tip
      THEN BEGIN
        textlength := lengthtext (nodep[i]^.nayme, nodep[i]^.naymlength);
        IF textlength > maxtextlength
        THEN maxtextlength := textlength;
        END;
    fontheight := font[3];
    angle := pi*labelrotation/180.0;
    maxtextlength := maxtextlength/fontheight;
    textlength := textlength/fontheight;
    firstlet := firstlet/fontheight;
    IF ntips > 1
    THEN labelheight := 1.0/(nodespace*(ntips-1))
    ELSE labelheight := 1.0/nodespace;
    IF angle < (pi/6.0)
    THEN tipspacing := (nodespace
                  + COS(angle)*(maxtextlength-0.5))*labelheight
    ELSE IF ntips > 1
      THEN tipspacing := 1.0/(ntips-1.0)
      ELSE tipspacing := 1.0;
    topoflabels := labelheight*(1.0 + SIN(angle)*(maxtextlength-0.5)
                          + COS(angle)*0.5);
    rightoflabels := labelheight*(COS(angle)*(textlength-0.5) +SIN(angle)*0.5);
    leftoflabels := labelheight*(COS(angle)*firstlet*0.5+SIN(angle)*0.5);
    calctraverse (root, sum);
    rooty := root^.ycoord;
    FOR i := 1 TO nextnode DO
      nodep[i]^.ycoord := stemlength*treedepth + (1.0-stemlength)*treedepth
         *(nodep[i]^.ycoord-rooty)/(maxheight-rooty);
    rooty := 0.0;
  END; (* calculate *)

  PROCEDURE rescale;
    (* compute coordinates of tree for plot or preview device *)
    VAR i : INTEGER;
      treeheight, treewidth, extrax, extray, temp : REAL;
  BEGIN (* rescale *)
    treeheight := 0.0;
    FOR i := 1 TO nextnode DO
      IF nodep[i]^.ycoord > treeheight
      THEN treeheight := nodep[i]^.ycoord;
    treeheight := treeheight + topoflabels;
    treewidth := (ntips-1)*tipspacing + rightoflabels + leftoflabels;
    IF grows = vertical
    THEN BEGIN
      expand := (xsize - 2*xmargin)/treewidth;
      IF (ysize - 2*ymargin)/treeheight < expand
      THEN expand := (ysize - 2*ymargin)/treeheight;
      extrax := (xsize - 2*xmargin - treewidth*expand)/2.0;
      extray := (ysize - 2*ymargin - treeheight*expand)/2.0;
      END
    ELSE BEGIN
      expand := (ysize - 2*ymargin)/treewidth;
      IF (xsize - 2*xmargin)/treeheight < expand
      THEN expand := (xsize - 2*xmargin)/treeheight;
      extrax := (xsize - 2*xmargin - treeheight*expand)/2.0;
      extray := (ysize - 2*ymargin - treewidth*expand)/2.0;
      END;
    FOR i := 1 TO nextnode DO BEGIN
      nodep[i]^.xcoord := expand*(nodep[i]^.xcoord + leftoflabels);
      nodep[i]^.ycoord := expand*(nodep[i]^.ycoord - rooty);
      IF grows = horizontal
      THEN BEGIN
        temp := nodep[i]^.ycoord;
        nodep[i]^.ycoord := expand*treewidth-nodep[i]^.xcoord;
        nodep[i]^.xcoord := temp;
        END;
      nodep[i]^.xcoord := nodep[i]^.xcoord + xmargin + extrax;
      nodep[i]^.ycoord := nodep[i]^.ycoord + ymargin + extray;
      END;
    IF grows = vertical
    THEN rooty := ymargin + extray
    ELSE rooty := xmargin + extrax;
  END; (* rescale *)

  PROCEDURE plot(pen : penstatustype; xabs, yabs : REAL);
    VAR xhigh, yhigh, xlow, ylow, newx, newy, ixnow, iynow,
      ixabs, iyabs, temp, xdiff, ydiff, err, x1, y1 : INTEGER;

    PROCEDURE pout(n : INTEGER);
    BEGIN (* pout *)
      IF previewing
      THEN WRITE(n:TRUNC(0.434295*LN(n)+0.0001))
      ELSE WRITE(plotfile, n:TRUNC(0.434295*LN(n)+0.0001))
    END; (* pout *)

    PROCEDURE plotdot (ix, iy : INTEGER);
      (* plot one dot at ix, iy *)
      VAR iy0, iy1, iy2 : INTEGER;
    BEGIN (* plotdot *)
      iy0 := strptop - iy;
      IF (iy0 >= 0) AND (iy0 < strpdeep) AND (ix > 0) AND (ix < strpwide)
      THEN BEGIN
        empty := FALSE;
        CASE plotter OF
          citoh : BEGIN
            iy1 := 1;
            iy2 := iy0;
            END;
          epson : BEGIN
            iy1 := 1;
            iy2 := 7 - iy0;
            END;
          oki : BEGIN
            iy1 := 1;
            iy2 := 7 - iy0;
            END;
          toshiba : BEGIN
            iy1 := iy0 DIV 6 + 1;
            iy2 := 5 - (iy0 MOD 6);
            END;
          other : BEGIN
            (* code for making dot array for a new printer
              goes here *)
            END
          END;
        IF ((ORD(stripe[iy1][ix]) DIV bit[iy2]) MOD 2) = 0
        THEN stripe[iy1][ix] := CHR(ORD(stripe[iy1][ix]) + bit[iy2]);
        END;
    END; (* plotdot *)

  BEGIN (* plot *)
    IF (NOT dotmatrix) OR previewing
    THEN CASE plotter OF
      tek : BEGIN
        IF pen = penup
        THEN BEGIN
          IF previewing
          THEN WRITE(CHR(29))
          ELSE WRITE(plotfile, CHR(29));
          END;
        ixnow := ROUND(xabs);
        iynow := ROUND(yabs);
        xhigh := ixnow DIV 32;
        yhigh := iynow DIV 32;
        xlow := ixnow MOD 32;
        ylow := iynow MOD 32;
        IF NOT ebcdic
        THEN BEGIN
          IF yhigh <> oldyhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(yhigh+32))
            ELSE WRITE(plotfile, CHR(yhigh+32));
            END;
          IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
            THEN BEGIN
              IF previewing
              THEN WRITE(CHR(ylow+96))
              ELSE WRITE(plotfile, CHR(ylow+96));
              END;
          IF xhigh <> oldxhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(xhigh+32))
            ELSE WRITE(plotfile, CHR(xhigh+32));
            END;
          IF previewing
          THEN WRITE(CHR(xlow+64))
          ELSE WRITE(plotfile, CHR(xlow+64));
          END
        ELSE BEGIN (* DLS/JMH -- for systems that use EBCDIC coding *)
          IF yhigh <> oldyhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(eb[yhigh+32]))
            ELSE WRITE(plotfile, CHR(eb[yhigh+32]));
            END;
          IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(eb[ylow+96]))
            ELSE WRITE(plotfile, CHR(eb[ylow+96]));
            END;
          IF xhigh <> oldxhigh
          THEN BEGIN
            IF previewing
            THEN WRITE(CHR(eb[xhigh+32]))
            ELSE WRITE(plotfile, CHR(eb[xhigh+32]));
            END;
          IF previewing
          THEN WRITE(CHR(eb[xlow+64]))
          ELSE WRITE(plotfile, CHR(eb[xlow+64]));
          END;

        (*  DLS/JMH -- Put CR-LFs in appropriate places.  This is needed if
            output is going to a file or if the terminal is treated as a file
            with a maximum record length.                     *)
       (*   Commented out because it won't work on Unix, but
          may be helpful if you have problems with lines in
          plotfile being too long with Tektronix option
        nmoves := nmoves + 1;
        IF nmoves = 12 THEN BEGIN
          nmoves := 0;
          IF NOT previewing
          THEN WRITELN(plotfile);
          END;
       end of section commented out *)
        oldxhigh := xhigh;
        oldxlow := xlow;
        oldyhigh := yhigh;
        oldylow := ylow;
        END;
      hp : BEGIN
        IF pen=pendown
        THEN WRITE(plotfile, 'PD')
        ELSE WRITE(plotfile, 'PU');
        pout(ROUND(xabs));
        WRITE(plotfile, ',');
        pout(ROUND(yabs));
        WRITELN(plotfile, ';');
        END;
      lw : BEGIN
        IF pen=pendown
        THEN WRITELN(plotfile, xabs:8:2, yabs:8:2, ' lineto ')
        ELSE WRITELN(plotfile, 'stroke ', xabs:8:2, yabs:8:2, ' moveto ');
        END;
      ibmpc : BEGIN
      (*  for most IBM PC monitors using TURBO 4.0 or 5.0 BGI interface
       Written by Buz Wilson, November, 1988 *)
(* Comment out Turbo specific code
        newx := ROUND(xabs);
        newy := ABS(ROUND(yabs)-GetMaxY);
        IF pen = pendown
        THEN Line(oldx,oldy,newx,newy);
        oldx := newx;
        oldy := newy
end commenting out of Turbo specific code *)
        END;
      houston : BEGIN
        IF pen = pendown
        THEN WRITE(plotfile, 'D ')
        ELSE WRITE(plotfile, 'U ');
        pout(ROUND(xabs));
        WRITE(plotfile, ',');
        pout(ROUND(yabs));
        WRITELN(plotfile);
        END;
      decregis : BEGIN
        newx := ROUND(xabs);
        newy := ABS(ROUND(yabs)-479);
        IF pen = pendown
        THEN BEGIN
          IF previewing
          THEN BEGIN
            WRITE('P['); pout(oldx);
            WRITE(','); pout(oldy);
            WRITE(']V['); pout(newx);
            WRITE(','); pout(newy);
            WRITE(']');
            END
          ELSE BEGIN
            WRITE(plotfile, 'P['); pout(oldx);
            WRITE(plotfile, ','); pout(oldy);
            WRITE(plotfile, ']V['); pout(newx);
            WRITE(plotfile, ','); pout(newy);
            WRITE(plotfile, ']');
            END;
          nmoves := nmoves + 1;
          IF nmoves = 3
          THEN BEGIN
            nmoves := 0;
            IF NOT previewing
            THEN WRITELN(plotfile);
            END
          END;
        oldx := newx;
        oldy := newy;
        END;
      other : BEGIN
        (* code for a pen move on a new plotter goes here *)
        END
      END
    ELSE BEGIN
      ixabs := ROUND(xabs);
      iyabs := ROUND(yabs);
      ixnow := ROUND(xnow);
      iynow := ROUND(ynow);
      IF (pen = pendown) AND
        (NOT (((iyabs < strpbottom) AND (iynow < strpbottom))
          OR ((iyabs > strptop) AND (iynow > strptop))))
      THEN BEGIN
        IF ixabs < ixnow
        THEN BEGIN
          temp := ixnow;
          ixnow := ixabs;
          ixabs := temp;
          temp := iynow;
          iynow := iyabs;
          iyabs := temp;
          END;
        xdiff := ixabs - ixnow;
        ydiff := iyabs - iynow;
        IF ydiff >= 0
        THEN IF xdiff >= ydiff
          THEN BEGIN
            err := -xdiff DIV 2;
            FOR x1 := ixnow TO ixabs DO BEGIN
              plotdot (x1, iynow);
              err := err + ydiff;
              IF err > 0
              THEN BEGIN
                iynow := iynow + 1;
                err := err - xdiff;
                END;
              END;
            END
          ELSE BEGIN
            err := -ydiff DIV 2;
            FOR y1 := iynow TO iyabs DO BEGIN
              plotdot (ixnow, y1);
              err := err + xdiff;
              IF err > 0
              THEN BEGIN
                ixnow := ixnow + 1;
                err := err - ydiff;
                END;
              END;
            END
        ELSE IF xdiff < -ydiff
          THEN BEGIN
            err := ydiff DIV 2;
            FOR y1 := iynow DOWNTO iyabs DO BEGIN
              plotdot (ixnow, y1);
              err := err + xdiff;
              IF err > 0
              THEN BEGIN
                ixnow := ixnow + 1;
                err := err + ydiff;
                END;
              END;
            END
          ELSE BEGIN
            err := -xdiff DIV 2;
            FOR x1 := ixnow TO ixabs DO BEGIN
              plotdot (x1, iynow);
              err := err - ydiff;
              IF err > 0
              THEN BEGIN
                iynow := iynow - 1;
                err := err - xdiff;
                END;
              END;
            END;
        END;
      xnow := xabs;
      ynow := yabs;
      END;
  END; (* plot *)

  PROCEDURE plottree (p, q : nodeptr);
    (* plot part or all of tree on the plotting device *)
    CONST segments = 40;
    VAR i : INTEGER;
      x1, y1, x2, y2, x3, y3, f, g, h, fract, minny, miny : REAL;
      pp : nodeptr;
  BEGIN (* plottree *)
    x2 := xscale*(xoffset + p^.xcoord);
    y2 := yscale*(yoffset + p^.ycoord);
    IF p <> root
    THEN BEGIN
      x1 := xscale*(xoffset + q^.xcoord);
      y1 := yscale*(yoffset + q^.ycoord);
      plot (penup, x1, y1);
      CASE style OF
        cladogram : plot (pendown, x2, y2);
        phenogram : BEGIN
          IF grows = vertical
          THEN plot (pendown, x2, y1)
          ELSE plot (pendown, x1, y2);
          plot (pendown, x2, y2);
          END;
        curvogram : FOR i := 1 TO segments DO BEGIN
            f := i/segments;
            g := i/segments;
            h := 1.0 - SQRT(1.0-g*g);
            IF grows = vertical
            THEN BEGIN
              x3 := x1*(1.0-f)+x2*f;
              y3 := y1 + (y2 - y1)*h;
              END
            ELSE BEGIN
              x3 := x1 + (x2 - x1)*h;
              y3 := y1*(1.0-f)+y2*f;
              END;
            plot (pendown, x3, y3);
            END;
        eurogram : BEGIN
          IF grows = vertical
          THEN plot (pendown, x2, (2*y1+y2)/3)
          ELSE plot (pendown, (2*x1+x2)/3, y2);
          plot (pendown, x2, y2);
          END;
        swoopogram : IF (grows = vertical) AND (NOT (ABS(y1-y2) < epsilon))
              OR ((grows = horizontal) AND (NOT (ABS(x1-x2) < epsilon)))
          THEN BEGIN
            IF grows = vertical
            THEN miny := p^.ycoord
            ELSE miny := p^.xcoord;
            pp := q^.next;
            WHILE pp <> q DO BEGIN
              IF grows = vertical
              THEN minny := pp^.back^.ycoord
              ELSE minny := pp^.back^.xcoord;
              IF minny < miny
              THEN miny := minny;
              pp := pp^.next;
              END;
            IF grows = vertical
            THEN miny := yscale*(yoffset+miny)
            ELSE miny := xscale*(xoffset+miny);
            IF grows = vertical
            THEN fract := 0.3333*(miny-y1)/(y2-y1)
            ELSE fract := 0.3333*(miny-x1)/(x2-x1);
            FOR i := 1 TO segments DO BEGIN
              f := i/segments;
              IF f < fract
              THEN g := f/fract
              ELSE g := (f-fract)/(1.0-fract);
              IF f < fract
              THEN h := fract*SQRT(1.0-(1.0-g)*(1.0-g))
              ELSE h := fract + (1.0-fract)*(1.000001 - SQRT(1.000001-g*g));
              IF grows = vertical
              THEN BEGIN
                x3 := x1*(1.0-f)+x2*f;
                y3 := y1 + (y2 - y1)*h;
                END
              ELSE BEGIN
                x3 := x1 + (x2 - x1)*h;
                y3 := y1*(1.0-f)+y2*f;
                END;
              plot (pendown, x3, y3);
              END;
            END
        END;
      END
    ELSE BEGIN
      IF grows = vertical
      THEN BEGIN
        x1 := xscale*(xoffset + p^.xcoord);
        y1 := yscale*(yoffset + rooty);
        END
      ELSE BEGIN
        x1 := xscale*(xoffset + rooty);
        y1 := yscale*(yoffset + p^.ycoord);
        END;
      plot (penup, x1, y1);
      plot (pendown, x2, y2);
      END;
    IF NOT p^.tip
    THEN BEGIN
      pp := p^.next;
      WHILE pp <> p DO BEGIN
        plottree (pp^.back, p);
        pp := pp^.next;
        END;
      END;
  END; (* plottree *)

  PROCEDURE plotlabels;
    TYPE pentype = (treepen, labelpen);
    VAR i : INTEGER;
      compr, dx, dy, angle : REAL;
      lp : nodeptr;

    PROCEDURE changepen(pen : pentype);
    BEGIN (* changepen *)
      CASE pen OF
        treepen: CASE plotter OF
          tek : ;  (* DLS/JMH -- avoids noisy warnings *)
          hp : WRITELN(plotfile, 'SP1;');
          lw : BEGIN
            WRITELN(plotfile, 'stroke ', treeline:8:2, ' setlinewidth ');
            WRITELN(plotfile, ' 1 setlinecap 1 setlinejoin ');
            END;
          ibmpc : BEGIN END;
          houston : BEGIN {??} END;
          other : BEGIN
            (* code for a new plotter goes here *)
            END
          END;
        labelpen: CASE plotter OF
          tek : ;  (* DLS/JMH -- avoids noisy warnings *)
          hp : WRITELN(plotfile, 'SP2;');
          lw : BEGIN
            WRITELN(plotfile, ' stroke', labelline:8:2, ' setlinewidth ');
            WRITELN(plotfile, '1 setlinecap 1 setlinejoin ');
            END;
          ibmpc : BEGIN END;
          houston : BEGIN END;
          other : BEGIN
            (* code for a new plotter goes here *)
            END
          END;
        END;
    END; (* changepen *)

    PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
      height, compress : REAL; x, y : REAL;
      slope : REAL; VAR font : fonttype);
      CONST xstart = 10;
        ystart = 35;
      VAR i, j, code, coord : INTEGER;
        heightfont, xfactor, yfactor, xfont, yfont,
        xplot, yplot, sinslope, cosslope, xx, yy : REAL;
        penstatus : penstatustype;

      PROCEDURE plotchar(VAR place : INTEGER);
      BEGIN (* plotchar *)
        heightfont := font[place+2];
        yfactor := height/heightfont;
        xfactor := yfactor;
        place := place+3;
        REPEAT
          place := place+1;
          coord := font[place];
          IF coord > 0
            THEN penstatus := pendown
            ELSE penstatus := penup;
          coord := ABS(coord);
          coord := coord MOD 10000;
          xfont := ((coord DIV 100) - xstart)*xfactor;
          yfont := ((coord MOD 100) - ystart)*yfactor;
          xplot := xx + (xfont*cosslope + yfont*sinslope)*compress;
          yplot := yy - xfont*sinslope + yfont*cosslope;
          plot(penstatus, xplot, yplot);
          UNTIL ABS(font[place]) >= 10000;
        xx := xplot;
        yy := yplot;
      END; (* plotchar *)

    BEGIN (* plottext *)
      xx := x;
      yy := y;
      sinslope := SIN(pi*slope/180.0);
      cosslope := COS(pi*slope/180.0);
      FOR i := 1 TO nchars DO BEGIN
        code := ORD(pstring[i]);
        j := 1;
        WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
          j := font[j];
        IF font[j+1] = code THEN plotchar(j);
        END;
    END; (* plottext *)

  BEGIN (* plotlabels *)
    compr := xunitspercm/yunitspercm;
    IF penchange = yes THEN changepen(labelpen);
    angle := labelrotation*pi/180.0;
    FOR i := 1 TO nextnode DO BEGIN
      IF nodep[i]^.tip
      THEN BEGIN
        lp := nodep[i];
        dx := -labelheight*expand*0.70710*COS(angle+pi/4.0);
        dy := labelheight*expand*(1.0-0.70710*SIN(angle+pi/4.0));
        IF grows = vertical
        THEN plottext(lp^.nayme,lp^.naymlength,
          labelheight*expand*xscale/compr, compr,
          xscale*(lp^.xcoord+dx+xoffset), yscale*(lp^.ycoord+dy+yoffset),
          -labelrotation,font)
        ELSE plottext(lp^.nayme,lp^.naymlength,
          labelheight*expand*yscale, compr, xscale*(lp^.xcoord+dy+xoffset),
          yscale*(lp^.ycoord-dx+yoffset), -labelrotation+90.0,font)
        END;
      END;
    IF penchange = yes THEN changepen(treepen);
  END; (* plotlabels *)

  PROCEDURE initplotter;
  BEGIN (* initplotter *)
    CASE plotter OF
      tek : BEGIN
        oldxhigh := -1;
        oldxlow := -1;
        oldyhigh := -1;
        oldylow := -1;
        nmoves := 0;  (* DLS/JMH -- See Procedure PLOT *)
        IF previewing
        THEN WRITE(escape,  CHR(12))   (* DLS/JMH *)
        ELSE WRITE(plotfile, escape,  CHR(12));   (* DLS/JMH *)
        END;
      hp : WRITELN(plotfile, 'IN;SP1;VS10.0;');
      lw : BEGIN
        WRITELN(plotfile, '%!PS-Adobe-1.0');
        treeline := 0.18*labelheight*yscale*expand;
        labelline := 0.06*labelheight*yscale*expand;
        WRITELN(plotfile, ' 1 setlinecap 1 setlinejoin ');
        WRITELN(plotfile, treeline:8:2, ' setlinewidth ', 'newpath ');
        END;
      ibmpc : BEGIN
        (* for most IBM PC monitors using TURBO 4 or 5 BGI interface *)
(* Comment out Turbo specific code
         InitGraph(GraphDriver,HiMode,'');
end commenting out of Turbo specific code *)
        END;
      houston : BEGIN END;
      decregis : BEGIN
        oldx := 300;
        oldy := 1;
        nmoves := 0;
        IF previewing
        THEN BEGIN
          WRITE(escape,'[2J');
          WRITE(escape,'Pp');
          WRITE('W(I3)');
          WRITELN('S(A[0,0][799,479])');
          END
        ELSE BEGIN
          WRITE(plotfile, escape,'[2J');
          WRITE(plotfile, escape,'Pp');
          WRITE(plotfile, 'W(I3)');
          WRITELN(plotfile, 'S(A[0,0][799,479])');
          END;
        END;
      epson : WRITE(plotfile, CHR(27), '3', CHR(24));
      oki : WRITE(plotfile, CHR(27), '%9', CHR(16));
      citoh : WRITE(plotfile, CHR(27), 'T16');
      toshiba : BEGIN
        WRITELN(plotfile, CHR(27), CHR(26), 'I');
        WRITELN(plotfile);
        WRITELN(plotfile, CHR(27), 'L06');
        END;
      other : BEGIN
        (* initialization code for a new plotter goes here *)
        END
      END;
  END; (* initplotter *)

  PROCEDURE finishplotter;
    VAR trash : CHAR;  (* DLS/JMH for ega, screen drivers *)
  BEGIN (* finishplotter *)
    CASE plotter OF
      tek : BEGIN
        IF previewing
        THEN BEGIN
          READLN(trash);
          WRITE(escape, CHR(12));
          END
        ELSE BEGIN
          WRITELN(plotfile);
          plot (penup, 1.0, 1.0);
          END;
        END;
      hp : BEGIN
        plot(penup, 1.0, 1.0);
        WRITELN(plotfile,'SP;');
        END;
      lw : BEGIN
        WRITELN(plotfile, 'stroke showpage ');
        WRITELN(plotfile);
        END;
      ibmpc : BEGIN
        (* termination code for IBM Monitor output, TURBO 4 or 5 specific *)
(* Comment out Turbo specific code
        trash := ReadKey;
        RestoreCRTMode;
end commenting out of Turbo specific code *)
        END;
      houston : BEGIN END;
      decregis : BEGIN
          plot (penup, 1.0, 1.0);
          IF previewing
          THEN WRITE(escape, ' ')
          ELSE WRITE(plotfile, escape, ' ');
          IF previewing
          THEN READ(trash);
          END;
      epson : WRITE(plotfile, CHR(27), '3', CHR(36));
      oki : BEGIN END;
      citoh : WRITE(plotfile, CHR(27), 'A');
      toshiba : WRITELN(plotfile, CHR(27), CHR(26), 'I');
      other : BEGIN
        (* termination code for a new plotter goes here *)
        END
      END;
  END; (* finishplotter *)

  PROCEDURE plotpreview;
    (* plot out a preview of the final plot *)
    VAR ch : CHAR;

    PROCEDURE makebox;
      (* draw the box on screen which represents plotting area *)
      VAR ch : CHAR;
    BEGIN (* makebox *)
      WRITELN;
      WRITELN('We now will preview the tree.  The box that will be');
      WRITELN('plotted on the screen represents the boundary of the');
      WRITELN('final plotting surface.  To see the preview, press on');
      WRITELN('the ENTER or RETURN key (you may need to do it twice).');
      WRITELN('When finished viewing it, press on that key again.');
      oldpenchange := penchange;
      oldxsize := xsize;
      oldysize := ysize;
      oldxunitspercm := xunitspercm;
      oldyunitspercm := yunitspercm;
      oldxcorner := xcorner;
      oldycorner := ycorner;
      oldplotter := plotter;
      plotter := previewer;
      READLN(ch);
      plotrparms;
      xcorner := xcorner + 0.05*xsize;
      ycorner := ycorner + 0.05*ysize;
      xsize := xsize*0.9;
      ysize := ysize*0.9;
      scale := ysize/oldysize;
      IF (xsize/oldxsize) < scale THEN scale := xsize/oldxsize;
      xoffset := (xcorner + (xsize - oldxsize*scale)/2.0)/scale;
      yoffset := (ycorner + (ysize - oldysize*scale)/2.0)/scale;
      xscale := scale*xunitspercm;
      yscale := scale*yunitspercm;
      initplotter;
      plot (penup, xscale*xoffset, yscale*yoffset);
      plot (pendown, xscale*xoffset, yscale*(yoffset+oldysize));
      plot (pendown, xscale*(xoffset+oldxsize), yscale*(yoffset+oldysize));
      plot (pendown, xscale*(xoffset+oldxsize), yscale*yoffset);
      plot (pendown, xscale*xoffset, yscale*yoffset);
    END; (* makebox *)

  BEGIN (* plotpreview *)
    previewing := TRUE;
    makebox;
    plottree (root, root);
    plotlabels;
    finishplotter;
    penchange := oldpenchange;
    xsize := oldxsize;
    ysize := oldysize;
    xunitspercm := oldxunitspercm;
    yunitspercm := oldyunitspercm;
    xscale := xunitspercm;
    yscale := yunitspercm;
    plotter := oldplotter;
    xcorner := oldxcorner;
    ycorner := oldycorner;
    WRITELN(' Is the tree ready to be plotted? (Answer Y or N)');
    REPEAT
      WRITELN(' Type Y or N:');
      READLN(ch);
      uppercase (ch);
    UNTIL (ch = 'Y') OR (ch = 'N');
    canbeplotted := (ch = 'Y');
  END; (* plotpreview *)

  PROCEDURE drawit;
    VAR i, j, line : INTEGER;
      null : CHAR;
      filesize : REAL;

    PROCEDURE striprint;
      VAR i, j, width : INTEGER;
        done : BOOLEAN;
    BEGIN (* striprint *)
      done := FALSE;
      width := strpwide;
      WHILE NOT done DO BEGIN
        FOR i := 1 TO strpdiv DO
          done := done OR (stripe[i][width] <> null);
        IF NOT done
        THEN width := width - 1;
        done := done OR (width = 0);
        END;
      CASE plotter OF
        epson: BEGIN
          IF NOT empty
          THEN BEGIN
            WRITE(plotfile, CHR(27), 'L', CHR(width MOD 256),
              CHR(width DIV 256));
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i]);
            filesize := filesize + width + 4;
            END;
          WRITELN(plotfile);
          END;
        oki : BEGIN
          IF NOT empty
          THEN BEGIN
            WRITE(plotfile, CHR(27), '%1', CHR(width DIV 128),
              CHR(width MOD 128));
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i]);
            filesize := filesize + width + 5;
            END;
          WRITELN(plotfile);
          END;
        citoh: BEGIN
          IF NOT empty
          THEN BEGIN
            WRITE(plotfile, CHR(27), 'S', (width DIV 1000):1,
              ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
              (width MOD 10):1);
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i]);
            filesize := filesize + width + 6;
            END;
          WRITELN(plotfile);
          END;
        toshiba: BEGIN
          IF NOT empty
          THEN BEGIN
            FOR i := 1 TO width DO
              FOR j := 1 TO 4 DO
                stripe[j][i] := CHR(ORD(stripe[j][i])+64);
            WRITE(plotfile, CHR(27), ';', (width DIV 1000):1,
              ((width DIV 100) MOD 10):1, ((width DIV 10) MOD 10):1,
              (width MOD 10):1);
            FOR i := 1 TO width DO
              WRITE(plotfile, stripe[1][i], stripe[2][i],
                   stripe[3][i], stripe[4][i]);
            filesize := filesize + 4*width + 6;
            END;
          WRITELN(plotfile);
          END;
        other : BEGIN
          (* graphics print code for a new printer goes here *)
          END
        END;
    END; (* striprint *)

  BEGIN (* drawit *)
    xoffset := 0.0;
    yoffset := 0.0;
    j := 1;
    FOR i := 0 TO 7 DO BEGIN
      bit[i] := j;
      j := j + j;
      END;
    null := CHR(0);
    strptop := numlines*strpdeep;
    strpbottom := strptop-strpdeep+1;
    IF dotmatrix
    THEN BEGIN
      WRITELN(' writing', numlines:3, ' lines ...');
      WRITELN('  Line     Output file size');
      WRITELN('  ----     ------ ---- ----');
      END;
    filesize := 0.0;
    FOR line := 1 TO numlines DO BEGIN
      FOR i := 1 TO strpdiv DO
        FOR j := 1 TO strpwide DO
          stripe[i][j] := null;
      empty := TRUE;
      xnow := strpwide / 2.0;
      ynow := 0.0;
      plottree (root, root);
      plotlabels;
      strptop := strpbottom - 1;
      strpbottom := strpbottom - strpdeep;
      IF dotmatrix
      THEN BEGIN
        striprint;
        IF (line MOD 5) = 0
        THEN WRITELN(line:5, TRUNC(filesize):16);
        END;
      END;
  END; (* drawit *)

BEGIN (* drawgram *)
(* Comment out Turbo specific code
  IF (RegisterBGIDriver(@CGADriver) < 0) OR
     (RegisterBGIDriver(@EGAVGADriver) < 0) OR
     (RegisterBGIDRiver(@HERCDriver) < 0) OR
     (RegisterBGIDRiver(@ATTDriver) < 0)
  THEN BEGIN
     WRITELN('Graphics Error:', GraphErrorMsg(GraphResult));
     Halt(1);
     END;
end commenting out of Turbo specific code *)
  (*  ASSIGN(treefile, 'TREEFILE');   Turbo Pascal only *)
  RESET(treefile);
  (*  ASSIGN(plotfile, 'PLOTFILE');   Turbo Pascal only *)
  REWRITE(plotfile);
  (*  ASSIGN(fontfile, 'FONTFILE');       Turbo Pascal only *)
  RESET(fontfile);
  WRITELN('DRAWGRAM from PHYLIP version 3.3');
  setup;
  WRITELN('Reading tree ... ');
  treeread;
  WRITELN('Tree has been read.');
  WRITELN('Loading the font ... ');
  loadfont(font);
  WRITELN('Font loaded.');
  previewing := FALSE;
  initialparms;
  canbeplotted := FALSE;
  WHILE NOT canbeplotted DO BEGIN
    REPEAT
      showparms;
      IF NOT parmsready THEN getparms;
    UNTIL parmsready;
    calculate;
    rescale;
    canbeplotted := TRUE;
    IF preview THEN plotpreview;
    END;
  previewing := FALSE;
  initplotter;
  IF NOT (plotter = ibmpc)
  THEN WRITELN('Writing plot file ...');
  drawit;
  finishplotter;
  WRITELN('Finished.');
  (* CLOSE(fontfile);   may need this here *)
  (* CLOSE(treefile);   may need this here *)
  (* CLOSE(plotfile);  Turbo Pascal and maybe others too *)
  WRITELN('End of run.');
END.

