PROGRAM Gendist (infile, OUTPUT);
  (* version 3.3. (c) Copyright 1988, 1990 by the University of Washington
     and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 40;   (* maximum number of species *)
   maxchr = 100;    (* maximum number of characters *)
   namelength = 10; (* number of characters max. in species name *)
   epsilon = 0.000001; (* a very small number *)

TYPE phenotype = ARRAY [1..maxchr] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;

VAR infile : TEXT;
   numsp, loci, totalleles, df : INTEGER;
   alleles : ARRAY[1..maxchr] OF INTEGER;
   x : ARRAY[1..maxsp] OF phenotype;
   d : ARRAY[1..maxsp] OF ARRAY[1..maxsp] OF REAL;
   names : ARRAY[1..maxsp] OF naym;
   all, cavalli, lower, nei, reynolds, anerror : BOOLEAN;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         --  works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert a character to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         READ(infile, numsp, loci);
         IF numsp > maxsp
         THEN WRITELN(OUTPUT, 'TOO MANY SPECIES: adjust CONSTants');
         IF loci > maxchr
         THEN WRITELN(OUTPUT, 'TOO MANY CHARACTERS: adjust CONSTants');
         anerror := (numsp > maxsp) OR (loci > maxchr);
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
      BEGIN (* getoptions *)
         all := FALSE;
         cavalli := FALSE;
         lower := FALSE;
         nei := TRUE;
         reynolds := FALSE;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'C') OR (ch = 'L')
               OR (ch = 'N') OR (ch = 'R')
            THEN BEGIN
               CASE ch OF
                  'A' : all := TRUE;
                  'C' : cavalli := TRUE;
                  'L' : lower := TRUE;
                  'N' : nei := TRUE;
                  'R' : reynolds := TRUE
                  END;
               IF reynolds OR cavalli
               THEN nei := FALSE;
               END
            ELSE IF ch <> ' '
               THEN BEGIN
                  WRITELN(OUTPUT, 'BAD OPTION CHARACTER: ', ch);
                  anerror := TRUE;
                  END;
            END;
      END; (* getoptions *)

      PROCEDURE getalleles;
         VAR i : INTEGER;
      BEGIN (* getalleles *)
         totalleles := 0;
         READLN(infile);
         FOR i := 1 TO loci DO BEGIN
            IF EOLN(infile)
            THEN READLN(infile);
            READ(infile, alleles[i]);
            totalleles := totalleles + alleles[i];
            END;
         anerror := anerror OR (totalleles > maxchr);
         df := totalleles - loci;
         IF totalleles > maxchr
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, 'TOO MANY LOCI OR ALLELES: adjust CONSTants');
            END;
      END; (* getalleles *)

      PROCEDURE getdata;
         VAR i, j, k, m, n, p : INTEGER;
            sum : REAL;
      BEGIN (* getdata *)
         FOR i := 1 TO numsp DO
            IF NOT anerror
            THEN BEGIN
               READLN(infile);
               FOR j := 1 TO namelength DO READ(infile, names[i][j]);
               m := 1;
               p := 1;
               FOR j := 1 TO loci DO
                  IF NOT anerror
                  THEN BEGIN
                     sum := 0.0;
                     IF all
                     THEN n := alleles[j]
                     ELSE n := alleles[j] - 1;
                     FOR k := 1 TO n DO BEGIN
                        IF EOLN(infile) THEN READLN(infile);
                        READ(infile, x[i][m]);
                        sum := sum + x[i][m];
                        IF x[i][m] < 0.0
                        THEN BEGIN
                           WRITELN(OUTPUT);
                           WRITE(OUTPUT, 'LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': AN ALLELE');
                           WRITELN(OUTPUT, ' FREQUENCY IS NEGATIVE');
                           anerror := TRUE;
                           END;
                        p := p + 1;
                        m := m + 1;
                        END;
                     IF all AND (ABS(sum - 1.0) > epsilon)
                     THEN BEGIN
                        WRITELN(OUTPUT);
                        WRITELN(OUTPUT, 'LOCUS', j:3, ' IN SPECIES',
                              i:3, ': FREQUENCIES DO NOT ADD UP TO 1');
                        anerror := TRUE;
                        END;
                     IF NOT all
                     THEN BEGIN
                        x[i][m] := 1.0 - sum;
                        IF x[i][m] < -epsilon
                        THEN BEGIN
                           WRITELN(OUTPUT);
                           WRITELN(OUTPUT, 'LOCUS', j:3, ' IN SPECIES',
                                 i:3, ': FREQUENCIES ADD UP TO MORE THAN 1');
                           anerror := TRUE;
                           END;
                        m := m + 1;
                        END;
                     END;
               END;
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      IF NOT anerror
      THEN getoptions;
      IF NOT anerror
      THEN getalleles;
      IF NOT anerror
      THEN getdata;
   END; (* getinput *)

   PROCEDURE makedists;
      VAR i, j, k : INTEGER;
         s, s1, s2, s3, f : REAL;
   BEGIN (* makedists *)
      FOR i := 1 TO numsp DO d[i][i] := 0.0;
      FOR i := 1 TO numsp DO
         FOR j := 1 TO i-1 DO BEGIN
            IF cavalli
            THEN BEGIN
               s := 0;
               FOR k := 1 TO totalleles DO BEGIN
                  f := x[i][k] * x[j][k];
                  IF f > 0.0
                  THEN s := s + SQRT(f);
                  END;
               d[i][j] := 4*(loci - s)/df;
               END;
            IF nei
            THEN BEGIN
               s1 := 0.0;
               s2 := 0.0;
               s3 := 0.0;
               FOR k := 1 TO totalleles DO BEGIN
                  s1 := s1 + x[i][k] * x[j][k];
                  s2 := s2 + SQR(x[i][k]);
                  s3 := s3 + SQR(x[j][k]);
                  END;
               IF s1 <= 1.0E-20
               THEN d[i][j] := -1.0
               ELSE d[i][j] := - LN(s1/SQRT(s2*s3));
               END;
            IF reynolds
            THEN BEGIN
               s1 := 0.0;
               s2 := 0.0;
               FOR k := 1 TO totalleles DO BEGIN
                  s1 := s1 + SQR(x[i][k] - x[j][k]);
                  s2 := s2 + x[i][k] * x[j][k];
                  END;
               d[i][j] := s1 / (2*loci - 2*s2);
               END;
            d[j][i] := d[i][j];
            END;
   END; (* makedists *)

   PROCEDURE writedists;
      VAR i, j, k : INTEGER;
   BEGIN (* writedists *)
     WRITELN(OUTPUT,  numsp:5);
     FOR i := 1 TO numsp DO BEGIN
        FOR j := 1 TO namelength DO WRITE(OUTPUT, names[i][j]);
        IF lower
        THEN k := i-1
        ELSE k := numsp;
        FOR j := 1 TO k DO BEGIN
           WRITE(OUTPUT, d[i][j]:8:4);
           IF (((j+1) MOD 7) = 0) AND (j < k)
           THEN WRITELN(OUTPUT);
           END;
        WRITELN(OUTPUT);
        END;
   END; (* writedists *)

BEGIN (* main program *)
   (* ASSIGN(infile, 'INFILE');  Turbo Pascal only *)
   RESET(infile);
   anerror := FALSE;
   getinput;
   IF NOT anerror
   THEN makedists;
   IF NOT anerror
   THEN writedists;
(* CLOSE(infile);  -- may need this here *)
END.

